/*--------------------------------------------------------------------
 * $Id: linear.c,v 1.4 2002/12/18 13:57:26 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <malloc.h>
#include "numeric.h"

/***********************************************************************************/
/* Function: linear_coeffc                                                @31_30i@ */
/* Description:                                                                    */
/*  Calculate coefficients for linear interpolation;                               */
/*  memory for coefficients will be allocated automatically!                       */
/*  These function has been created for compatibility with the spline              */
/*  interpolation functions; for this reason four coefficients are calculated,     */
/*  but a2[] and a3[] are set to zero. The interpolation may be done with          */
/*  calc_spline_values().                                                          */
/*                                                                                 */
/* Parameters:                                                                     */
/*  double *x:    x values of the data points, i=0...number-1                      */
/*  double *y:    y values of the data points, i=0...number-1                      */
/*  int number:   number of datapoints                                             */
/*  double **a0:  array of coefficients, i=0...number-1                            */ 
/*  double **a1:  array of coefficients, i=0...number-1                            */ 
/*  double **a2:  array of coefficients, i=0...number-1, set to zero               */ 
/*  double **a3:  array of coefficients, i=0...number-1, set to zero               */ 
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i31_30@ */
/***********************************************************************************/

int linear_coeffc (double *x, double *y, int number, 
		   double **a0, double **a1, double **a2, double **a3)
{
  int i=0;


  /* check if x ascending */
  for (i=0; i<number-1; i++)  {
    if (x[i]>=x[i+1]) {
      fprintf(stderr,"x not ascending %d %f %f\n", i, x[i], x[i+1]);
      return X_NOT_ASCENDING;
    }
  }

  
  /* allocate memory for coefficients */
  *a0 = (double *) calloc (number, sizeof(double));
  *a1 = (double *) calloc (number, sizeof(double));
  *a2 = (double *) calloc (number, sizeof(double));
  *a3 = (double *) calloc (number, sizeof(double));


  /* calculate coefficients */
  for (i=0; i<number-1; i++)  {
    (*a0)[i] = y[i];
    (*a1)[i] = (y[i+1]-y[i]) / (x[i+1]-x[i]);
  }

  if (number==1) {
    (*a0)[0] = y[0];
    (*a1)[0] = 0;
  }

  return 0;   /* if o.k. */
}



/***********************************************************************************/
/* Function: slinear_coeffc                                               @31_30i@ */
/* Description:                                                                    */
/*  Like linear_coeffc(), but sort data before calculating coefficients. Please    */
/*  note that the fields x and y themselves are sorted by slinear_coeffc()  which  */
/*  is an important pre-requisite when they are passed to calc_splined_value()     */
/*  later.                                                                         */
/*                                                                                 */
/* Parameters:                                                                     */
/*  double *x:    x values of the data points, i=0...number-1                      */
/*  double *y:    y values of the data points, i=0...number-1                      */
/*  int number:   number of datapoints                                             */
/*  double **a0:  array of coefficients, i=0...number-1                            */ 
/*  double **a1:  array of coefficients, i=0...number-1                            */ 
/*  double **a2:  array of coefficients, i=0...number-1, set to zero               */ 
/*  double **a3:  array of coefficients, i=0...number-1, set to zero               */ 
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i31_30@ */
/***********************************************************************************/

int slinear_coeffc (double *x, double *y, int number, 
		    double **a0, double **a1, double **a2, double **a3)
{
  int i=0, status=0;
  double **value=NULL;

  value = calloc (number, sizeof(double *));

  for (i=0; i<number; i++) {
    value[i] = calloc (2, sizeof(double));
    value[i][0] = x[i];
    value[i][1] = y[i];
  }

  status = ASCII_sortarray (value, number, 2, 0, 0);
  if (status!=0) {
    fprintf (stderr, "Error %d sorting data\n", status);
    return status;
  }

  for (i=0; i<number; i++) {
    x[i] = value[i][0];
    y[i] = value[i][1];
    free(value[i]);
  }

  free(value);
  
  /* allocate memory for coefficients */
  *a0 = (double *) calloc (number, sizeof(double));
  *a1 = (double *) calloc (number, sizeof(double));
  *a2 = (double *) calloc (number, sizeof(double));
  *a3 = (double *) calloc (number, sizeof(double));


  /* calculate coefficients */
  for (i=0; i<number-1; i++)  {
    (*a0)[i] = y[i];
    (*a1)[i] = (y[i+1]-y[i]) / (x[i+1]-x[i]);
  }

  if (number==1) {
    (*a0)[0] = y[0];
    (*a1)[0] = 0;
  }

  return 0;   /* if o.k. */
}


