/*--------------------------------------------------------------------
 * $Id: ancillary.c,v 1.59 2003/01/13 10:10:12 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <ckdfu.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/numeric.h"
#include "../src_f/solver.h"

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

/* prototypes of internal functions */
static int cnvlv (double *x_spec, float  *y_spec, int n_spec,
		  double *x_slit, double *y_slit, int n_slit);


static float radiance2bt (float rad, float *wvnmlo, float *wvnmhi, float *filter, int n);

static int scale_output (input_struct input, output_struct *output, 
			 double ffactor, double rfactor, int iv);

static int output2bt (input_struct input, output_struct *output);

static int read_photon_file (char *filename, float *lambda_t, int nlamdba_t, float **fraction);


/***************************************************************/
/* Setup the wavelength grid for the transmission calculation. */
/***************************************************************/

int setup_wlgrid (input_struct input, output_struct *output)
{     
  int iv=0, status=0;
  float *lambda_lower=NULL, *lambda_upper=NULL;

  
  if (input.wl.start>0 && input.wl.end>0) {
    output->wl.start = input.wl.start;
    output->wl.end   = input.wl.end;
  }

  if (input.ck_scheme == CK_NONE || input.ck_scheme == CK_SBDART) {   /* no correlated-k */

    if ( strlen(input.filename[FN_WLBANDS]) > 0 ) {
      
      /* read center wavelength and band limits [wavenumbers] from file */
      status = read_3c_file_float (input.filename[FN_WLBANDS],
				   &(output->wl.lambda_t),
				   &(lambda_lower), 
				   &(lambda_upper),
				   &output->wl.nlambda_t);

      if (status!=0) {
	fprintf (stderr, "Error %d opening %s\n", status, input.filename[FN_WLBANDS]);
	return status;
      }

      /* convert lower and upper wavelength bounds to wavenumbers */
      output->wl.wvnmlo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
      output->wl.wvnmhi_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

      for (iv=0; iv<output->wl.nlambda_t; iv++) {
	output->wl.wvnmlo_t[iv] = 1.0E7 / lambda_upper[iv];
	output->wl.wvnmhi_t[iv] = 1.0E7 / lambda_lower[iv];
      }

      /* free temporary wavelengths arrays */
      free (lambda_lower);
      free (lambda_upper);

      if (input.wl.start < 0 || input.wl.end < 0) {
	output->wl.start = output->wl.lambda_t[0];
	output->wl.end   = output->wl.lambda_t[output->wl.nlambda_t-1];
	if (!input.quiet) 
	    fprintf (stderr, " ... setting wavelength range to %f - %fnm\n",
		     output->wl.start, output->wl.end);
      }


      /* select the required wavelength range */
      output->wl.nlambda_rte_upper = 0;
      
      iv=0;
      while (output->wl.lambda_t[iv] <= output->wl.start)
	if (++iv==output->wl.nlambda_t)
	  break;
    
      if (iv>0)
        iv-=1;
  
      output->wl.nlambda_rte_lower = iv;
  
      iv=0;
      while (output->wl.lambda_t[iv] < output->wl.end) {
        if (++iv==output->wl.nlambda_t)
  	break;
      }    
  
      if (iv==output->wl.nlambda_t)
        iv-=1;
  
      output->wl.nlambda_rte_upper = iv;
    }
    else {

      if ( strlen(input.filename[FN_WLTRANS]) > 0 ) {
	
	/* read wavelengths from transmittance file */
	status = read_1c_file_float (input.filename[FN_WLTRANS],
				     &(output->wl.lambda_t),
				     &output->wl.nlambda_t);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d opening %s\n", status, input.filename[FN_WLTRANS]);
	  return status;
	}
	
	if (input.wl.start < 0 || input.wl.end < 0) {
	  output->wl.start = output->wl.lambda_t[0];
	  output->wl.end   = output->wl.lambda_t[output->wl.nlambda_t-1];
	  if (!input.quiet) 
	    fprintf (stderr, " ... setting wavelength range to %f - %fnm\n",
		     output->wl.start, output->wl.end);
	}
	
	
	/* select the required wavelength range */
	output->wl.nlambda_rte_upper = 0;
	
	iv=0;
	while (output->wl.lambda_t[iv] <= output->wl.start)
	  if (++iv==output->wl.nlambda_t)
	    break;
    
	if (iv>0)
	  iv-=1;
	
	output->wl.nlambda_rte_lower = iv;
	
	iv=0;
	while (output->wl.lambda_t[iv] < output->wl.end) {
	  if (++iv==output->wl.nlambda_t)
	    break;
	}    
	
	if (iv==output->wl.nlambda_t)
        iv-=1;
	
	output->wl.nlambda_rte_upper = iv;
      }
      else {
	
	/* set up a reasonable wavelength grid for the transmittance calculation */
	status = set_transmittance_wl_grid(&input.wl, &output->wl);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by set_transmittance_wl_grid()\n", status);
	  return status;
	}
	
	/* here we need to calculate the full wavelength range */
	output->wl.nlambda_rte_lower = 0;
	output->wl.nlambda_rte_upper = output->wl.nlambda_t-1;
      }
      
      /* allocate memory for the band limit arrays */
      output->wl.wvnmlo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
      output->wl.wvnmhi_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
      
      /* the default bandwidth is 1cm-1 to get the emittance per cm-1; */
      /* input.bandwidth can be set with thermal_bandwidth             */       
      for (iv=0; iv<output->wl.nlambda_t; iv++) {
	output->wl.wvnmlo_t[iv] = 1.0E7 / output->wl.lambda_t[iv] - input.bandwidth / 2.0;
	output->wl.wvnmhi_t[iv] = output->wl.wvnmlo_t[iv] + input.bandwidth;
      }
    }
      
    /* check if the transmittance grid covers the required wavelength range */
    if (output->wl.lambda_t[0] > output->wl.start || 
	output->wl.lambda_t[output->wl.nlambda_t-1] < output->wl.end) {
      fprintf (stderr, "Error, transmittance calculation grid (%f - %f nm)\n", 
	       output->wl.lambda_t[0], output->wl.lambda_t[output->wl.nlambda_t-1]);
      fprintf (stderr, "does not cover the user-defined wavelength range (%f - %f nm)\n",
	       output->wl.start, output->wl.end);
      return -1;
    }
    
    /* now setup the ck structure for the SBDART table */
    if (input.ck_scheme == CK_SBDART) {
      
      /* read SBDART tables */
      status = sbdart_readtables (&(output->ck), 
				  output->wl.nlambda_t,
				  input.filename[FN_PATH],
				  input.rte.mc.filename[FN_MC_PHOTONS],
				  input.quiet);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by sbdart_readtables()\n", status); 
	return status;
      }
    }

    /* need to read the photons file */
    if (input.ck_scheme == CK_NONE) {


      if (strlen(input.rte.mc.filename[FN_MC_PHOTONS])>0) {
	status = read_photon_file (input.rte.mc.filename[FN_MC_PHOTONS], 
				   output->wl.lambda_t, output->wl.nlambda_t, &(output->wl.pfraction));
	if (status!=0) {
	  fprintf (stderr, "Error %d reading photon file name %s\n", status, input.rte.mc.filename[FN_MC_PHOTONS]);
	  return status;
	}
      }
      else {
	output->wl.pfraction = calloc (output->wl.nlambda_t, sizeof (float));
	for (iv=0; iv<output->wl.nlambda_t; iv++)
	  output->wl.pfraction[iv] = 1.0 / (float) output->wl.nlambda_t;
      }
    }
  }
  else { /* correlated-k */

    /* read information about wavelength grid and quadrature points */
    switch(input.ck_scheme) {
    case CK_KATO:
    case CK_KATO2:
      
      /* read Kato et al. [1999] tables */
      status = kato_readtables (input.ck_scheme, &(output->ck), input.filename[FN_PATH],
				input.rte.mc.filename[FN_MC_PHOTONS]);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by kato_readtables()\n", status); 
	return status;
      }
      
      break;

    case CK_FU:

      /* read Fu and Liou [1992/93] tables */
      status = fu_readtables (&(output->ck), input.filename[FN_PATH],
			      input.rte.mc.filename[FN_MC_PHOTONS]);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by fu_readtables()\n", status); 
	return status;
      }

      break;

    case CK_AVHRR_KRATZ:

      /* read Kratz [1999] tables */
      status = avhrr_kratz_readtables (&(output->ck), input.filename[FN_PATH],
				       input.rte.mc.filename[FN_MC_PHOTONS]);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by avhrr_kratz_readtables()\n", status); 
	return status;
      }
      
      break;

    case CK_GENERIC:

      /* read generic tables in CDF format */
      status = ck_generic_readtables (&(output->ck), input.filename[FN_CK_GENERIC],
				      input.rte.mc.filename[FN_MC_PHOTONS]);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ck_generic_readtables()\n", status); 
	return status;
      }
      
      break;

    default:
      fprintf (stderr, "Error: unsupported correlated-k scheme\n");
      return -1;
    }


    /* copy center wavelengths to transmittance grid  */
    /* and set wavenumber intervals                   */

    output->wl.nlambda_t = output->ck.n_wvl;
    output->wl.lambda_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
    output->wl.wvnmlo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
    output->wl.wvnmhi_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
    
    for (iv=0; iv<output->wl.nlambda_t; iv++) {
      output->wl.lambda_t[iv] = output->ck.wvlc[iv+1];
      output->wl.wvnmlo_t[iv] = 1.0E7 / output->ck.wvl2[iv+1];
      output->wl.wvnmhi_t[iv] = 1.0E7 / output->ck.wvl1[iv+1];
    }

    if (input.wl.start < 0 || input.wl.end < 0) {
      output->wl.start = output->wl.lambda_t[0];
      output->wl.end   = output->wl.lambda_t[output->wl.nlambda_t-1];
      if (!input.quiet) 
	  fprintf (stderr, " ... setting wavelength range to %f - %fnm\n",
		   output->wl.start, output->wl.end);
    }

    /* select the required wavelength range */
    output->wl.nlambda_rte_upper = 0;

    iv=0;
    while (output->wl.lambda_t[iv] <= output->wl.start)
      if (++iv==output->wl.nlambda_t)
	break;
    
    if (iv>0)
      iv-=1;

    output->wl.nlambda_rte_lower = iv;


    iv=0;
    while (output->wl.lambda_t[iv] < output->wl.end) {
      if (++iv==output->wl.nlambda_t)
	break;
    }    

    if (iv==output->wl.nlambda_t)
      iv-=1;

    output->wl.nlambda_rte_upper = iv;
  }
    

  /* inform the user about wavelength selection */
  if (!input.quiet) {
    if (output->wl.nlambda_rte_upper != output->wl.nlambda_t-1 ||
	output->wl.nlambda_rte_lower != 0) {
      fprintf (stderr, " ... user wavelength range: %f - %f nm\n", 
	       input.wl.start, input.wl.end);
      fprintf (stderr, " ... selected wavelength indices %d - %d\n", 
	       output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper);
      fprintf (stderr, " ... %f - %f nm\n", 
	       output->wl.lambda_t[output->wl.nlambda_rte_lower], 
	       output->wl.lambda_t[output->wl.nlambda_rte_upper]);
    }
  }

  return 0;
}  






/***************************************************************/
/* Interpolate from one wavelength grid to another, either     */
/* with natural cubic splines (linear=0), linear (linear=1),   */ 
/* or logarithmic (linear=2) if the input is sorted in         */
/* descending order, descend needs to be set to 1.             */
/***************************************************************/

int arb_wvn (int n_old_x, float *old_x, float *old_y,  
	     int n_new_x, float *new_x, float *new_y, 
	     int linear, int descend)
{
  int i=0, j=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *x=NULL, *y=NULL;
  double *tmp_y=NULL;
  double ynew=0;
  double tst1=0, tst2=0;

  x = (double *) calloc (n_old_x, sizeof(double));
  y = (double *) calloc (n_old_x, sizeof(double));

  if (!descend) {
    for (i=0; i<n_old_x; i++) {
      x[i] = (double) old_x[i];
      y[i] = (double) old_y[i];
    }
  }
  else {
    for (i=0; i<n_old_x; i++) {
      x[i] = (double) old_x[n_old_x-1-i];
      y[i] = (double) old_y[n_old_x-1-i];
    }
  }

  /* check if the array is now sorted in ascending order */
  for (i=0; i<n_old_x-1; i++)
    if (x[i]>=x[i+1]) {
      fprintf (stderr, "Error, x not sorted in ascending order\n");
      return -1;
    }

  switch (linear) {
  case 0:
    status = spline_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do spline interpolation\n");
      fprintf (stderr, "spline_coeffc() returned status %d\n", status);
      return status;
    }
    break;

  case 1:
    status = linear_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do linear interpolation\n");
      fprintf (stderr, "linear_coeffc() returned status %d\n", status);
      return status;
    }
    break;

  case 2:

    tmp_y = (double *) calloc (n_new_x, sizeof(double));
    for (i=0; i<n_new_x; i++)  {
    
      if (new_x[i] < x[0] || new_x[i] > x[n_old_x-1]) {
	fprintf (stderr, "Error, %f out of range\n", new_x[i]);
	return -1;
      }
      
      j=0;
      while (new_x[i] > x[j])	
	j++;
    
      if (j>0) 
	j--;
    
      /* logarithmic interpolation (if reasonable) */
      tst1 = fabs(y[j+1]-y[j]);
      tst2 = (y[j+1]<y[j] ? y[j+1] : y[j]);
      if (tst1<=0.001*y[j] || tst2<=0) /* linear */
	tmp_y[i] = y[j] + (new_x[i]-x[j]) / (x[j+1]-x[j]) * (y[j+1]-y[j]);
      else /* logarithmic */
	tmp_y[i] = exp (log(y[j]) + (new_x[i]-x[j])/(x[j+1]-x[j]) * 
			(log(y[j+1])-log(y[j])));
    }
			  
    /* first interpolate, then copy because then source */
    /* and target may be one and the same               */
    for (i=0; i<n_new_x; i++)
      new_y[i] = (float) tmp_y[i];

    free (tmp_y);
    break;

  default:
    fprintf (stderr, "Error, unknown interpolation type %d\n", linear);
    return -1;
  }


  switch (linear) {
  case 0:
  case 1:
    
    for (i=0; i<n_new_x; i++)  {
      status = calc_splined_value ((double) new_x[i], &ynew, x, n_old_x, a0, a1, a2, a3);
      
      /* this check is new in 0.99-alpha-6; so far, no error */
      /* was reported if a value could not be interpolated   */
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by calc_splined_value (%g)\n",
		 status, new_x[i]);
	fprintf (stderr, " %d data points, x[0] = %g, x[%d] = %g\n", n_old_x, x[0], n_old_x-1, x[n_old_x-1]);
	return status;
      }

      new_y[i] = (float) ynew;
    }
    
    free(a0); free(a1); free(a2); free(a3);
    break;

  case 2:
    /* no need to do anything because interpolation has already been done above */
    break;

  default:
    fprintf (stderr, "Error, unknown interpolation type %d\n", linear);
    return -1;
  }

  free(x); free(y);
  
  return 0;
}




/***************************************************************/
/* Interpolate from one wavelength grid to another, either     */
/* with natural cubic splines (linear=0) or linear (linear=1); */ 
/* if the input is sorted in descending order, descend needs   */
/* to be set to 1. In contrast to arb_wvn(), values that       */
/* cannot be interpolated will be set to 0.                    */
/***************************************************************/

int arb_wvn_zero (int n_old_x, float *old_x, float *old_y,  
		  int n_new_x, float *new_x, float *new_y, 
		  int linear, int descend)
{
  int i=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *x=NULL, *y=NULL;
  double ynew=0;

  x = (double *) calloc (n_old_x, sizeof(double));
  y = (double *) calloc (n_old_x, sizeof(double));

  if (!descend) {
    for (i=0; i<n_old_x; i++) {
      x[i] = (double) old_x[i];
      y[i] = (double) old_y[i];
    }
  }
  else {
    for (i=0; i<n_old_x; i++) {
      x[i] = (double) old_x[n_old_x-1-i];
      y[i] = (double) old_y[n_old_x-1-i];
    }
  }

  switch (linear) {
  case 0:
    status = spline_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do spline interpolation\n");
      fprintf (stderr, "spline_coeffc() returned status %d\n", status);
      return status;
    }
    break;

  case 1:
    status = linear_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do linear interpolation\n");
      fprintf (stderr, "linear_coeffc() returned status %d\n", status);
      return status;
    }
    break;

  case 2:
  default:
    fprintf (stderr, "Error, unknown interpolation type %d\n", linear);
    return -1;
  }

  
  switch (linear) {
  case 0:
  case 1:
    for (i=0; i<n_new_x; i++)  {
      status = calc_splined_value ((double) new_x[i], &ynew, x, n_old_x, a0, a1, a2, a3);
      
      if (status==0)
	new_y[i] = (float) ynew;
      else 
	new_y[i] = 0;
    }

    free(a0); free(a1); free(a2); free(a3);

    break;

  case 2:
  default:
    fprintf (stderr, "Error, unknown interpolation type %d\n", linear);
    return -1;
  }

  free(x); free(y);
    
  return 0;
}






/***************************************************************/
/* Interpolate from one wavelength grid to another, either     */
/* with natural cubic splines (linear=0) or linear (linear=1); */ 
/* interpolate only to a selected range of the output grid     */
/* (including n_new_x_lower and n_new_x_upper                  */
/***************************************************************/

int arb_wvn2 (int n_old_x, float *old_x, float *old_y,  
	      int n_new_x_lower, int n_new_x_upper, float *new_x, float *new_y, int linear)
{
  int i=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *x=NULL, *y=NULL;
  double ynew=0;

  x = (double *) calloc (n_old_x, sizeof(double));
  y = (double *) calloc (n_old_x, sizeof(double));

  for (i=0;i<n_old_x;i++) {
    x[i] = (double) old_x[i];
    y[i] = (double) old_y[i];
  }

  if (linear) {
    status = linear_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do linear interpolation\n");
      fprintf (stderr, "linear_coeffc() returned status %d\n", status);
      return status;
    }
  }
  else {
    status = spline_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do spline interpolation\n");
      fprintf (stderr, "spline_coeffc() returned status %d\n", status);
      return status;
    }
  }


  for (i=n_new_x_lower; i<=n_new_x_upper; i++)  {
    status = calc_splined_value ((double) new_x[i], &ynew, x, n_old_x, a0, a1, a2, a3);

    /* this check is new in 0.99-alpha-6; so far, no error */
    /* was reported if a value could not be interpolated   */

    if (status!=0) {
      fprintf (stderr, "Error %d returned by calc_splined_value (%g)\n",
	       status, new_x[i]);
      fprintf (stderr, " %d data points, x[0] = %g, x[%d] = %g\n", n_old_x, x[0], n_old_x-1, x[n_old_x-1]);
      return status;
    }

    new_y[i] = (float) ynew;
  }

  free(a0); free(a1); free(a2); free(a3);
  free(x); free(y);

  return 0;
}



/**************************************************************/
/* Interpolate a given profile (x,y) to a new grid xnew;      */
/* data are written to original array y and memory is         */
/* automatically reallocated                                  */
/**************************************************************/

int interpolate_profile (float *x, float **y, int n,  
				float *xnew, int nnew, int linear)
{
  int status=0;
  float *tmp = calloc (nnew, sizeof(float));

  status = arb_wvn (n, x, *y, nnew, xnew, tmp, linear, 1);
  if (status!=0) {
    fprintf (stderr, "Error %d interpolating profile\n", status);
    return status;
  }

  free(*y);
  *y = tmp;

  return 0;
}



/**************************************************************/
/* Define the internal wavelength grid for the transmittance  */
/* calculation.                                               */
/**************************************************************/

int set_transmittance_wl_grid(wl_inp_struct *wl_inp, wl_out_struct *wl_out)
{
  int iv=0;
  float lambda=0.0, wvn_step_t=0.0;
  
  /* Determine number of wavelengths needed for transmittance calculation */
  lambda = wl_inp->start;
  iv = 1;
  while (lambda < wl_inp->end) {

    if (lambda < 350.0)
      wvn_step_t = 0.5;
    else
      wvn_step_t = 1.0;

    lambda += wvn_step_t;
    iv++;
  }

  wl_out->nlambda_t = iv;
  wl_out->lambda_t = (float *) calloc (wl_out->nlambda_t, sizeof(float));


  /* Set wavelengths for transmittance calculation */
  lambda = wl_inp->start;

  for (iv=0; iv<wl_out->nlambda_t; iv++) {
    wl_out->lambda_t[iv] = lambda;

    if ( lambda < 350.0 )
      wvn_step_t = 0.5;
    else
      wvn_step_t = 1.0;

    lambda += wvn_step_t;
  }

  return 0;
}



/**************************************************************/
/* Convolve the spectrum with a user-defined slit function    */
/**************************************************************/

int convolve (input_struct input, output_struct *output)
{
  int is=0, js=0, ks=0, iv=0, iu=0, j=0, lev=0;
  int status=0;
  double *x_data=NULL;

  int n_slit=0;
  double *x_slit=NULL, *y_slit=NULL;


  /* read slit function from file */
  status = read_2c_file (input.filename[FN_SLITFUNCTION], &x_slit, &y_slit, &n_slit); 

  if (status!=0)  { 
    fprintf (stderr, "Error %d reading file %s\n", status, input.filename[FN_SLITFUNCTION]); 
    return status; 
  } 

  x_data = calloc (output->wl.nlambda_h, sizeof(double)); 

  for (iv=0; iv<output->wl.nlambda_h; iv++)
    x_data[iv] = (double) output->wl.lambda_h[iv];
  
  if (input.rte.solver == POLRADTRAN ) {
    for (lev=0; lev<input.atm.nzout; lev++) {

      /* Convolve up_flux and down_flux*/
      for (is=0; is<input.rte.pol_nstokes; is++) {

	status = cnvlv (x_data, output->up_flux[lev][is], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);

	if (status!=0)  { 
	  fprintf (stderr, "Error %d convoluting up_flux\n", status); 
	  return status; 
	} 

	status = cnvlv (x_data, output->down_flux[lev][is], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);

	if (status!=0)  { 
	  fprintf (stderr, "Error %d convoluting down_flux\n", status); 
	  return status; 
	} 

	/* Convolve up_rad and down_rad*/
	for (j=0; j<input.rte.nphi; j++) {
	  for (iu=0; iu<input.rte.nstr/2; iu++) {
	    status = cnvlv (x_data, output->down_rad[lev][j][iu][is], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);

	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting down_rad\n", status); 
	      return status; 
	    } 

	    status = cnvlv (x_data, output->up_rad[lev][j][iu][is], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting up_rad\n", status); 
	      return status; 
	    } 
	  }
	}
      }
    }
  }
  else {

    for (lev=0; lev<input.atm.nzout; lev++) {

      /* Convolve rfldir, rfldn, flup and uavg for all cases */

      /* rfldir */
      status = cnvlv (x_data, output->rfldir[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);

      if (status!=0)  { 
	fprintf (stderr, "Error %d convoluting rfldir\n", status); 
	return status; 
      } 
      
      /* rfldn */
      status = cnvlv (x_data, output->rfldn[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, "Error %d convoluting rfldn\n", status); 
	return status; 
      } 
      
      /* flup */
      status = cnvlv (x_data, output->flup[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, "Error %d convoluting flup\n", status); 
	return status; 
      } 
      
      /* uavg */
      status = cnvlv (x_data, output->uavg[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, "Error %d convoluting uavg\n", status); 
	return status; 
      } 
      
      
      /* 3D fields */
      if (input.rte.mc.passback3D)
	for (is=0; is<input.rte.mc.Nx_sample; is++)
	  for (js=0; js<input.rte.mc.Ny_sample; js++) {

	    /* rfldir3d */
	    status = cnvlv (x_data, output->rfldir3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting rfldir3d\n", status); 
	      return status; 
	    } 

	    /* rfldn3d */
	    status = cnvlv (x_data, output->rfldn3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting rfldn3d\n", status); 
	      return status; 
	    } 

	    /* flup3d */
	    status = cnvlv (x_data, output->flup3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting flup3d\n", status); 
	      return status; 
	    } 

	    /* radiance3d */
	    status = cnvlv (x_data, output->radiance3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting radiance3d\n", status); 
	      return status; 
	    } 
	  }
	

      if (input.rte.solver != TWOSTR) {
	
	/* Convolve uavgso, uavgdn, uavgup */
	
	/* uavgso */
	status = cnvlv (x_data, output->uavgso[lev], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, "Error %d convoluting uavgso\n", status); 
	  return status; 
	} 
	
	/* uavgdn */
	status = cnvlv (x_data, output->uavgdn[lev], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, "Error %d convoluting uavgdn\n", status); 
	  return status; 
	} 
	
	/* uavgup */
	status = cnvlv (x_data, output->uavgup[lev], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, "Error %d convoluting uavgup\n", status); 
	  return status; 
	} 

	
	if ( input.rte.numu > 0 ) {
	  
	  /* Convolve u0u */
	  for (iu=0; iu<input.rte.maxumu; iu++) {
	    status = cnvlv (x_data, output->u0u[lev][iu], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting u0u\n", status); 
	      return status; 
	    } 
	  }
	  
	  if (output->print_phi > 0) {
	    
	    /* Convolve uu */
	    for (j=0; j<input.rte.nphi; j++) {
	      for (iu=0; iu<input.rte.maxumu; iu++) {
		status = cnvlv (x_data, output->uu[lev][j][iu], output->wl.nlambda_h,
				x_slit, y_slit, n_slit);
		
		if (status!=0)  { 
		  fprintf (stderr, "Error %d convoluting u0u\n", status); 
		  return status; 
		} 
	      }
	    }
	  }
	}
      }
    }

    /* 3D absorption */
    if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)

      for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
	for (is=0; is<input.rte.mc.Nx_cloud; is++)
	  for (js=0; js<input.rte.mc.Ny_cloud; js++) {
	    
	    status = cnvlv (x_data, output->abs3d[ks][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting abs3d\n", status); 
	      return status; 
	    } 
	  }
  }


  free(x_data);  

  return 0;
}



/**************************************************************/
/* Convolve a spectrum with a slit function.                  */
/* Internal function, used by convolve().                     */
/**************************************************************/

static int cnvlv (double *x_spec, float  *y_spec, int n_spec,
		  double *x_slit, double *y_slit, int n_slit) 
{
  int iv=0, status=0;
  double *y_data = calloc (n_spec, sizeof(double));
  double *tmp=NULL;

  for (iv=0; iv<n_spec; iv++) 
    y_data[iv] = (double) y_spec[iv];
  
  status = int_convolute (x_spec, y_data, n_spec, x_slit, y_slit, n_slit, &tmp);
  if (status!=0) 
    return status; 
  
  for (iv=0; iv<n_spec; iv++) 
    y_spec[iv] = (float) tmp[iv];
  
  free(y_data);
  free(tmp);

  return 0;
}


/**************************************************************/
/* Interpolate the hires spectrum to a user-defined           */
/* wavelength grid.                                           */
/**************************************************************/

int spline_interpolate(input_struct input, output_struct *output)
{
  int is=0, js=0, ks=0, iu=0, iv=0, j=0, lev=0;
  int status=0;
  double *x_user=NULL;
  int linear=0;

  if (strlen(input.filename[FN_SPLINE]) > 0) {

    /* read file with user x values */
    status = read_1c_file (input.filename[FN_SPLINE], &x_user, &output->wl.nlambda_s);
    if (status!=0)  {
      fprintf (stderr, "spline_interpolate: error %d reading file %s\n", 
	       status, input.filename[FN_SPLINE]);
      return (status);
    }

    output->wl.lambda_s = (float *) calloc (output->wl.nlambda_s, sizeof(float));

    for (iv=0; iv<output->wl.nlambda_s; iv++)
      output->wl.lambda_s[iv] = (float) x_user[iv];

    /* Check wavelength ranges */
    status=0;
    if (output->wl.start > output->wl.lambda_s[0])  {
      fprintf (stderr, "Error, spline wavelength %f is smaller than wvn %f\n",
	       output->wl.lambda_s[0], output->wl.start);
      status--;
    }

    if (output->wl.end < output->wl.lambda_s[output->wl.nlambda_s-1])  {
      fprintf (stderr, "Error, spline wavelength %f is greater than wvn %f\n",
	       output->wl.lambda_s[output->wl.nlambda_s-1], output->wl.end);
      status--;
    }

    if (status!=0) 
      return status;
  }
  else {
    output->wl.nlambda_s = 
      (int) ((input.spline_lambda_1-input.spline_lambda_0)/input.spline_lambda_step+1);
    
    output->wl.lambda_s = (float *) calloc (output->wl.nlambda_s, sizeof(float));
    
    for (iv=0; iv<output->wl.nlambda_s; iv++)
      output->wl.lambda_s[iv] = input.spline_lambda_0 + (float) iv * input.spline_lambda_step;
  }
  

  /* Spline interpolate to given wavelengths */
  if (input.rte.solver == POLRADTRAN ) {

    for (lev=0; lev<input.atm.nzout; lev++) {

      /* Spline interpolate up_flux and down_flux*/
      for (is=0; is<input.rte.pol_nstokes; is++) {

	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->up_flux[lev][is],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->up_flux[lev][is], 
			 linear, 0);

	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating up_flux\n", status);
	  return status;
	}

	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->down_flux[lev][is],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->down_flux[lev][is], 
			 linear, 0);

	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating down_flux\n", status);
	  return status;
	}

	/* Spline interpolate up_rad and down_rad */
	for (j=0; j<input.rte.nphi; j++) {

	  for (iu=0; iu<input.rte.nstr/2; iu++) {

	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->down_rad[lev][j][iu][is],  
			     output->wl.nlambda_s, output->wl.lambda_s, output->down_rad[lev][j][iu][is],
			     linear, 0);

	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating down_rad\n", status);
	      return status;
	    }

	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->up_rad[lev][j][iu][is],  
			     output->wl.nlambda_s, output->wl.lambda_s, output->up_rad[lev][j][iu][is],
			     linear, 0);

	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating up_rad\n", status);
	      return status;
	    }
	  }
	}
      }
    }
  }
  else {
    for (lev=0; lev<input.atm.nzout; lev++) {
      
      /* Spline interpolate rfldir, rfldn, flup and uavg for all cases */
      
      /* rfldir */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldir[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->rfldir[lev], 
		       linear, 0);

      if (status!=0) {
	fprintf (stderr, "Error %d interpolating rfldir\n", status);
	return status;
      }
      
      /* rfldn */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldn[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->rfldn[lev], 
		       linear, 0);

      if (status!=0) {
	fprintf (stderr, "Error %d interpolating rfldn\n", status);
	return status;
      }
      
      /* flup */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->flup[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->flup[lev], 
		       linear, 0);
      if (status!=0) {
	fprintf (stderr, "Error %d interpolating flup\n", status);
	return status;
      }
      
      /* uavg */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavg[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->uavg[lev], 
		       linear, 0);

      if (status!=0) {
	fprintf (stderr, "Error %d interpolating uavg\n", status);
	return status;
      }
      

      /* 3D fields */
      if (input.rte.mc.passback3D)
	for (is=0; is<input.rte.mc.Nx_sample; is++)
	  for (js=0; js<input.rte.mc.Ny_sample; js++) {

	    /* rfldir3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldir3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->rfldir3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating rfldir3d\n", status);
	      return status;
	    }
	    
	    /* rfldn3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldn3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->rfldn3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating rfldn3d\n", status);
	      return status;
	    }
	    
	    /* rflup3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->flup3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->flup3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating flup3d\n", status);
	      return status;
	    }

	    /* radiance3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->radiance3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->radiance3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating radiance3d\n", status);
	      return status;
	    }
	  }
      
      if (input.rte.solver != TWOSTR) {
	
	/* Spline interpolate uavgso, uavgdn, uavgup */
	
	/* uavgso */
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgso[lev],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->uavgso[lev], 
			 linear, 0);

	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating uavgso\n", status);
	  return status;
	}
	
	/* uavgdn */
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgdn[lev],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->uavgdn[lev], 
			 linear, 0);

	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating uavgdn\n", status);
	  return status;
	}
	
	/* uavgup */
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgup[lev],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->uavgup[lev], 
			 linear, 0);

	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating uavgup\n", status);
	  return status;
	}
	
	if (input.rte.numu > 0) {
	  
	  /* Spline interpolate u0u */
	  for (iu=0; iu<input.rte.maxumu; iu++) {
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->u0u[lev][iu],  
			     output->wl.nlambda_s, output->wl.lambda_s, output->u0u[lev][iu], 
			     linear, 0);

	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating u0u\n", status);
	      return status;
	    }
	  }
	  
	  if (output->print_phi > 0) {
	    
	    /* Spline interpolate uu */
	    for (j=0; j<input.rte.nphi; j++) {
	      for (iu=0; iu<input.rte.maxumu; iu++) {
		status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uu[lev][j][iu],  
				 output->wl.nlambda_s, output->wl.lambda_s, output->uu[lev][j][iu], 
				 linear, 0);

		if (status!=0) {
		  fprintf (stderr, "Error %d interpolating uu\n", status);
		  return status;
		}
	      }
	    }
	  }
	}
      }
    }


    if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
      for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
	for (is=0; is<input.rte.mc.Nx_cloud; is++)
	  for (js=0; js<input.rte.mc.Ny_cloud; js++) {

	    
	    status = arb_wvn(output->wl.nlambda_t, output->wl.lambda_h, output->abs3d[ks][is][js], 
			     output->wl.nlambda_h, output->wl.lambda_s, output->abs3d[ks][is][js], 
			     linear, 0);

	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating abs3d\n", status);
	      return status;
	    }
	  }

  }
  

  free(output->wl.lambda_h);
  
  output->wl.nlambda_h = output->wl.nlambda_s;
  output->wl.lambda_h = (float *) calloc (output->wl.nlambda_s, sizeof(float));

  for (iv=0; iv<output->wl.nlambda_h; iv++)
    output->wl.lambda_h[iv] = (float) output->wl.lambda_s[iv];

  return status;
  
}





/**************************************************************/
/* Interpolate the transmittance from the internal grid to    */
/* high resolution grid.                                      */
/**************************************************************/

int interpolate_transmittance (input_struct input, output_struct *output)
{
  int lu=0, is=0, js=0, ks=0, iv=0, ivh=0, iu=0, j=0;
  int linear=0, status=0;


  /* If only one wavelength is required, or in correlated-k mode, */
  /* arrays are not interpolated but copied                       */
  
  if (output->wl.nlambda_t == 1 || (input.ck_scheme!=CK_NONE && input.ck_scheme!=CK_SBDART)) {
    
    for (ivh=0; ivh<output->wl.nlambda_h; ivh++) {
      
      if (output->wl.nlambda_t == 1)
	iv = ivh;
      else 
	iv = output->wl.map_e2h[ivh];
      
      for (lu=0; lu<input.atm.nzout; lu++) {
	if (input.rte.solver == POLRADTRAN) {

	  for (is=0;is<input.rte.pol_nstokes;is++) {
	    output->up_flux   [lu][is][ivh] = output->up_flux_t   [lu][is][iv];
	    output->down_flux [lu][is][ivh] = output->down_flux_t [lu][is][iv];
	  }

	  for (j=0; j<input.rte.nphi; j++)
	    for (iu=0; iu<input.rte.nstr/2; iu++)
	      for (is=0; is<input.rte.pol_nstokes; is++) {
		output->down_rad [lu][j][iu][is][ivh] = output->down_rad_t [lu][j][iu][is][iv];
		output->up_rad   [lu][j][iu][is][ivh] = output->up_rad_t   [lu][j][iu][is][iv];
	      } 
	}
	else {
	  output->flup   [lu][ivh] = output->flup_t   [lu][iv];
	  output->rfldir [lu][ivh] = output->rfldir_t [lu][iv];
	  output->rfldn  [lu][ivh] = output->rfldn_t  [lu][iv];
	  output->uavg   [lu][ivh] = output->uavg_t   [lu][iv];
	  output->uavgdn [lu][ivh] = output->uavgdn_t [lu][iv];
	  output->uavgso [lu][ivh] = output->uavgso_t [lu][iv];
	  output->uavgup [lu][ivh] = output->uavgup_t [lu][iv];


	  /* radiances */
	  for (iu=0; iu<input.rte.maxumu; iu++) {
	    output->u0u[lu][iu][ivh] = output->u0u_t[lu][iu][iv];

	    for (j=0; j<input.rte.nphi; j++)
	      output->uu[lu][j][iu][ivh] = output->uu_t[lu][j][iu][iv];
	  }

	  /* 3D fields */
	  if (input.rte.mc.passback3D)
	    for (is=0; is<input.rte.mc.Nx_sample; is++)
	      for (js=0; js<input.rte.mc.Ny_sample; js++) {
		output->rfldir3d   [lu][is][js][ivh] = output->rfldir3d_t   [lu][is][js][iv];
		output->rfldn3d    [lu][is][js][ivh] = output->rfldn3d_t    [lu][is][js][iv];
		output->flup3d     [lu][is][js][ivh] = output->flup3d_t     [lu][is][js][iv];
		output->radiance3d [lu][is][js][ivh] = output->radiance3d_t [lu][is][js][iv];
	      }
	}
      }
      
      /* absorption is defined on the 3D cloud grid, not on the user-defined grid */
      if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
	for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
	  for (is=0; is<input.rte.mc.Nx_cloud; is++)
	    for (js=0; js<input.rte.mc.Ny_cloud; js++)
	      output->abs3d [ks][is][js][ivh] = output->abs3d_t [ks][is][js][iv];

      output->sza [ivh] = output->atm.sza_t [iv];
    }
  } 
  else {   /* interpolation to different wavelengths required */
    for (lu=0; lu<input.atm.nzout; lu++) {

      if (input.rte.solver == POLRADTRAN) {
	for (is=0; is<input.rte.pol_nstokes; is++) {
	  status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->down_flux_t[lu][is], 
			    output->wl.nlambda_h, output->wl.lambda_h, output->down_flux[lu][is], linear, 0);
	  status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->up_flux_t[lu][is], 
			    output->wl.nlambda_h, output->wl.lambda_h, output->up_flux[lu][is], linear, 0);
	}
      }
      else {
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->flup_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->flup[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldir_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->rfldir[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldn_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->rfldn[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavg_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavg[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgdn_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavgdn[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgso_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavgso[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgup_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavgup[lu], linear, 0);
	
	/* 3D fields */
	if (input.rte.mc.passback3D)
	  for (is=0; is<input.rte.mc.Nx_sample; is++)
	    for (js=0; js<input.rte.mc.Ny_sample; js++) {
	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldir3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->rfldir3d[lu][is][js], linear, 0);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldn3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->rfldn3d[lu][is][js], linear, 0);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->flup3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->flup3d[lu][is][js], linear, 0);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->radiance3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->radiance3d[lu][is][js], linear, 0);

	    }
	
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->atm.sza_t, 
			  output->wl.nlambda_h, output->wl.lambda_h, output->sza, linear, 0);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	  return status;
	}
      }
      
      if (input.rte.solver == POLRADTRAN) {

	for (j=0; j<input.rte.nphi; j++) {
	  for (iu=0; iu<input.rte.nstr/2; iu++) {
	    for (is=0; is<input.rte.pol_nstokes; is++) {
	      
	      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->down_rad_t[lu][j][iu][is], 
				output->wl.nlambda_h, output->wl.lambda_h, output->down_rad[lu][j][iu][is], 
				linear, 0);

	      if (status!=0) {
		fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
		return status;
	      }

	      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->up_rad_t[lu][j][iu][is], 
				output->wl.nlambda_h, output->wl.lambda_h, output->up_rad[lu][j][iu][is], 
				linear, 0);

	      if (status!=0) {
		fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
		return status;
	      }
	    }
	  }
	}
      }
      else {
	for (iu=0; iu<input.rte.maxumu; iu++) {
	  status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->u0u_t[lu][iu], 
			    output->wl.nlambda_h, output->wl.lambda_h, output->u0u[lu][iu], 
			    linear, 0);
	  
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	    return status;
	  }
	}
	
	for (j=0; j<input.rte.nphi; j++) {
	  for (iu=0; iu<input.rte.maxumu; iu++) {
	    
	    status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->uu_t[lu][j][iu], 
			      output->wl.nlambda_h, output->wl.lambda_h, output->uu[lu][j][iu], 
			      linear, 0);

	    if (status!=0) {
	      fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	      return status;
	    }
	  }
	}
      }
    }

    if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
      for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
	for (is=0; is<input.rte.mc.Nx_cloud; is++)
	  for (js=0; js<input.rte.mc.Ny_cloud; js++) {
	    status = arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->abs3d_t[ks][is][js], 
			     output->wl.nlambda_h, output->wl.lambda_h, output->abs3d[ks][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	      return status;
	    }
	  }
  }

  return 0;
}





/**************************************************************/
/* Multiply the transmittance with the extraterrestrial       */
/* irradiance.                                                */
/**************************************************************/

int multiply_extraterrestrial (input_struct input, output_struct *output)
{
  int iv=0, status=0;
  double ffactor=0, rfactor=0;
  
  /* multiply with extraterrestrial irradiance */
  for (iv=0; iv<output->wl.nlambda_h; iv++) {
    
    switch (input.source) {
    case SRC_THERMAL:
      ffactor = output->wl.filter[iv];
      rfactor = output->wl.filter[iv];
      break;

    case SRC_SOLAR:
      switch (input.processing) {
      case PROCESS_INT:
      case PROCESS_SUM:
	ffactor = output->wl.fbeam[iv] * output->wl.filter[iv];
	rfactor = output->wl.fbeam[iv] * output->wl.filter[iv];
	break;

      case PROCESS_NONE:
	switch (input.calibration) {
	case OUTCAL_ABSOLUTE:
	  ffactor = output->wl.fbeam[iv] * output->wl.filter[iv];
	  rfactor = output->wl.fbeam[iv] * output->wl.filter[iv];
	  break;
	  
	case OUTCAL_TRANSMITTANCE:
	  ffactor = 1.0 * output->wl.filter[iv];;
	  rfactor = 1.0 * output->wl.filter[iv];;
	  break;
	  
	case  OUTCAL_REFLECTIVITY:
	  ffactor = 1.0 / cos(output->atm.sza_t[iv]*PI/180.0) * output->wl.filter[iv];
	  rfactor = PI  / cos(output->atm.sza_t[iv]*PI/180.0) * output->wl.filter[iv];
	  break;

	default:
	  fprintf (stderr, "Error, unknown output calibration %d\n", input.calibration);
	  return -1;
	}

	break;

      default:
	fprintf (stderr, "Error, unknown output processing %d\n", input.processing);
	return -1;
      }

      break;

    default:
      fprintf (stderr, "Error, unknown source %d\n", input.source);
      return -1;
    }

    /**************************************************************/
    /* now scale irradiances with ffactor, radiances with rfactor */
    /**************************************************************/

    status = scale_output (input, output, ffactor, rfactor, iv);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by scale_output()\n", status);
      return status;
    }
  }

  return 0;
}


/***********************************************************************************/
/* Function: optical_properties                                           @61_30i@ */
/* Description:                                                                    */
/*  Calculates optical depth, single scattering albedo and                         */
/*  phase function from absorption and scattering cross sections                   */
/*  corresponding gas and particulate matter concentrations for                    */
/*  one wavelength.                                                                */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i61_30@ */
/***********************************************************************************/

int optical_properties (input_struct input, output_struct *output, int iv, int iq)
{
  static int first=1;
  int k=0, lc=0;
  int status=0;
  double babsa=0, babsd=0, babsi=0, babso=0, babs=0;
  double bscaa=0, bscad=0, bscai=0, bscar=0, bsca=0;
  double babso_md=0, babs_md=0; 
  double bext_md=0; /* The absorption of gases minus the one specified with a 
                        matrix using the dens_file command (for airmass factor calculations). */

  double babsatot=0, babsdtot=0, babsitot=0, babsotot=0;
  double bscaatot=0, bscadtot=0, bscaitot=0, bscartot=0;

  double moma=0, momd=0, momi=0;
  double bext=0, deltaz=0, g1a=0, g2a=0, fa=0, g1d=0, g2d=0, fd=0, g1i=0, g2i=0, fi=0;
  double sumsca=0;
  double p1=0, p2=0, p3=0, p4=0, p5=0, p6=0;
  FILE *f=NULL;

  int nlev = output->atm.nlev_common;
  int nlyr = nlev-1;

  double momag1=0, momag2=0;
  double momdg1=0, momdg2=0;
  double momig1=0, momig2=0;

  if (input.verbose) {
    fprintf (stderr, "\n*** wavelength: iv = %d, %f nm\n", iv, output->wl.lambda_t[iv]);
    fprintf (stderr, "*** optical_properties()\n");
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");
    fprintf (stderr, "   lc |  z[km] |  Rayleigh |       Aerosol              |       Water cloud             |        Ice cloud              |   Molecular \n");
    fprintf (stderr, "      |        |    dtau   |   scatter.      abs.  asy. |    scatter.        abs.  asy. |    scatter.        abs.  asy. |  absorption \n");
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");
  }

  if (first!=0) {
    first = 0;
    
    /* allocate memory for profiles */

    output->dtauc        = (float *) calloc (nlyr, sizeof(float));
    output->dtauc_md     = (float *) calloc (nlyr, sizeof(float));
    output->ssalb        = (float *) calloc (nlyr, sizeof(float));


    output->mc.z       = (float *) calloc (nlev, sizeof(float));
    output->mc.temper  = (float *) calloc (nlev, sizeof(float));

    output->mc.dtair   = (float *) calloc (nlev, sizeof(float));
    output->mc.dtmol   = (float *) calloc (nlev, sizeof(float));
    output->mc.dtaer   = (float *) calloc (nlev, sizeof(float));
    output->mc.dtcld   = (float *) calloc (nlev, sizeof(float));
    output->mc.dtice   = (float *) calloc (nlev, sizeof(float));
    output->mc.omaer   = (float *) calloc (nlev, sizeof(float));
    output->mc.omcld   = (float *) calloc (nlev, sizeof(float));
    output->mc.omice   = (float *) calloc (nlev, sizeof(float));

    output->mc.g1aer   = (float *) calloc (nlev, sizeof(float));
    output->mc.g2aer   = (float *) calloc (nlev, sizeof(float));
    output->mc.ffaer   = (float *) calloc (nlev, sizeof(float));

    output->mc.g1cld   = (float *) calloc (nlev, sizeof(float));
    output->mc.g2cld   = (float *) calloc (nlev, sizeof(float));
    output->mc.ffcld   = (float *) calloc (nlev, sizeof(float));

    output->mc.g1ice   = (float *) calloc (nlev, sizeof(float));
    output->mc.g2ice   = (float *) calloc (nlev, sizeof(float));
    output->mc.ffice   = (float *) calloc (nlev, sizeof(float));

    output->mc.lwccld  = (float *) calloc (nlev, sizeof(float));
    output->mc.recld   = (float *) calloc (nlev, sizeof(float));

    output->mc.momaer  = (float **) calloc (nlev, sizeof(float *));
    output->mc.nmomaer = (int *)    calloc (nlev, sizeof(int));

    for (lc=0; lc<nlev; lc++) {
      output->mc.ffcld[lc] = 1.0;
      output->mc.ffice[lc] = 1.0;
    }
    
    /* output->atm.nmom+1 stores the maximum number of moments */
    /* for all wavelengths and layers, minus 1                 */

    status = ASCII_calloc_float(&output->pmom, nlyr, output->atm.nmom+1);
    
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for output->pmom\n", status);
      return status;
    }

  }  

  babsatot=0;
  babsdtot=0;
  babsitot=0;
  babsotot=0;

  bscaatot=0;
  bscadtot=0;
  bscaitot=0;
  bscartot=0;

  for (lc=0; lc<nlyr; lc++) {

    /* The factor 1.e+5 converts z from km to cm.*/
    deltaz= (output->atm.zd[lc] - output->atm.zd[lc+1])*1e5;

    /* only in the Fu and Liou case, the Rayleigh scattering */
    /* cross section depends on the subband                  */
    switch(input.ck_scheme) {
    case CK_FU:
      bscar = output->atm.optprop.tau_rayleigh_t[lc][iv][iq];
      break;

    case CK_KATO:
    case CK_KATO2:
    case CK_AVHRR_KRATZ:
    case CK_GENERIC:
    case CK_SBDART:
    case CK_NONE:
      bscar = output->atm.optprop.tau_rayleigh_t[lc][iv][0];
      break;
      
    default:
      fprintf (stderr, "Error: unsupported correlated-k scheme %d\n", input.ck_scheme);
      return -1;

      break;
    }

    switch (output->atm.molabs) {
    case MOLABS_CALC:
      babso    = output->atm.optprop.tau_molabs_t[lc][iv][iq];
      babso_md = output->atm.optprop.tau_molabs_md_t[lc][iv][iq];
      break;
      
    case MOLABS_FILE_MONO:
      babso = output->atm.optprop.tau_molabs_user[lc];
      babso_md=0;
      break;

    case MOLABS_FILE_SPEC:
      babso = output->atm.optprop.tau_molabs_user_t[lc][iv];
      babso_md=0;
      break;

    case MOLABS_NONE:
      babso = 0;
      babso_md=0;
      break;
      
    default:
      fprintf (stderr, "Error, unknown molecular absorption option %d\n", 
	       output->atm.molabs);
      return -1;
    }

    /* scattering by aerosols, clouds, and ice clouds */
    bscaa = output->aer.optprop.ssa[iv][lc] * output->aer.optprop.dtau[iv][lc];
    bscad = output->wc.optprop.ssa [iv][lc] * output->wc.optprop.dtau [iv][lc];
    bscai = output->ic.optprop.ssa [iv][lc] * output->ic.optprop.dtau [iv][lc];
	  
    g1a   = output->aer.optprop.g1[iv][lc];
    g2a   = output->aer.optprop.g2[iv][lc];
    fa    = output->aer.optprop.ff[iv][lc];

    g1d   = output->wc.optprop.g1 [iv][lc];
    g2d   = output->wc.optprop.g2 [iv][lc];
    fd    = output->wc.optprop.ff [iv][lc];

    g1i   = output->ic.optprop.g1 [iv][lc];
    g2i   = output->ic.optprop.g2 [iv][lc];
    fi    = output->ic.optprop.ff [iv][lc];

    /*  Phase function */
    sumsca =  bscar + bscad +  bscaa + bscai;
    if (sumsca < FLT_MIN)
      sumsca = FLT_MIN;


    /* moments of the phase functions */

    momag1=bscaa;
    momag2=bscaa;

    momdg1=bscad;
    momdg2=bscad;

    momig1=bscai;
    momig2=bscai;

    output->pmom[lc][0] = 1.0;

    for (k=1; k<=output->atm.nmom; k++) {

      /* aerosol */
      if (output->aer.optprop.nmom[iv][lc] > 0) {
	if (k<output->aer.optprop.nmom[iv][lc])
	  moma = bscaa * output->aer.optprop.moment[iv][lc][k] / 
	    output->aer.optprop.moment[iv][lc][0];
	else 
	  moma=0;
      }
      else {  /* double Henyey-Greenstein */
	momag1 *= g1a;
	momag2 *= g2a;

	moma = momag1*fa + (1.0-fa)*momag2;
      }
      
      /* water clouds */
      if (output->wc.optprop.nmom[iv][lc] > 0) {
	if (k<output->wc.optprop.nmom[iv][lc])
	  momd = bscad * output->wc.optprop.moment[iv][lc][k] / output->wc.optprop.moment[iv][lc][0];
	else 
	  momd=0;
      }
      else {  /* double Henyey-Greenstein */
	momdg1 *= g1d;
	momdg2 *= g2d;

	momd = momdg1*fd + (1.0-fd)*momdg2;
      }

      /* ice clouds */
      if (output->ic.optprop.nmom[iv][lc] > 0) {
	if (k<output->ic.optprop.nmom[iv][lc])
	  momi = bscai * output->ic.optprop.moment[iv][lc][k] / output->ic.optprop.moment[iv][lc][0];
	else 
	  momi=0;
      }
      else {  /* double Henyey-Greenstein */
	momig1 *= g1i;
	momig2 *= g2i;

	momi = momig1*fi + (1.0-fi)*momig2;
      }

      output->pmom[lc][k] = (momd + moma + momi) / sumsca;

      if (k == 2)   /* Rayleigh scattering */
	output->pmom[lc][k] += bscar/(sumsca*10.0);
    }

    /* absorption by aerosols, clouds, and ice clouds */
    babsa  = (1.0 - output->aer.optprop.ssa[iv][lc]) * output->aer.optprop.dtau[iv][lc];
    babsd  = (1.0 - output->wc.optprop.ssa [iv][lc]) * output->wc.optprop.dtau [iv][lc];
    babsi  = (1.0 - output->ic.optprop.ssa [iv][lc]) * output->ic.optprop.dtau [iv][lc];
    
    /* absorption */
    if (input.absorption) {
      babs    = babsa + babso    + babsd + babsi;
      babs_md = babsa + babso_md + babsd + babsi;
      if (babs > FLT_MAX) 
	babs = FLT_MAX;
      if (babs_md > FLT_MAX) 
	babs_md = FLT_MAX;
    }
    else {
      babs  = 0;
      babso = 0;
      babsa = 0;
      babsd = 0;
      babsi = 0;
      babs_md  = 0;
      babso_md = 0;
    }

    /* Scattering coefficient */
    if (input.scattering) {
      bsca = bscar + bscad + bscaa + bscai;
      if (bsca > FLT_MAX) 
	bsca = FLT_MAX;
    }
    else {
      bsca  = 0;
      bscar = 0;
      bscaa = 0;
      bscad = 0;
      bscai = 0;
    }

    /* Extinction coefficient */
    bext = babs + bsca;
    if (bext > FLT_MAX) 
      bext = FLT_MAX;
    bext_md = babs_md + bsca;
    if (bext_md > FLT_MAX) 
      bext_md = FLT_MAX;

    if (input.verbose) {
      fprintf (stderr, "%5d | %6.2f | %9.6f | %9.6f %9.6f %5.3f | %11.6f %11.6f %5.3f | %11.6f %11.6f %5.3f | %11.6f\n", 
	       lc, output->atm.zd[lc+1], 
	       bscar,
	       bscaa, babsa, g1a*fa + (1.0-fa)*g2a,
	       bscad, babsd, g1d*fd + (1.0-fd)*g2d,
	       bscai, babsi, g1i*fi + (1.0-fi)*g2i,
	       babso
	       );
    }

    /* total column for verbose output */
    babsatot+=babsa;
    babsdtot+=babsd;
    babsitot+=babsi;
    babsotot+=babso;

    bscaatot+=bscaa;
    bscadtot+=bscad;
    bscaitot+=bscai;
    bscartot+=bscar;

    /* Optical properties for the mc model */
    output->mc.dtair[nlyr-1-lc] = bscar;
    output->mc.dtmol[nlyr-1-lc] = babso;
    output->mc.dtaer[nlyr-1-lc] = babsa + bscaa;
    output->mc.dtcld[nlyr-1-lc] = babsd + bscad;
    output->mc.dtice[nlyr-1-lc] = babsi + bscai;
		  	      
    output->mc.omaer[nlyr-1-lc] = output->aer.optprop.ssa[iv][lc];
    output->mc.omcld[nlyr-1-lc] = output->wc.optprop.ssa [iv][lc];
    output->mc.omice[nlyr-1-lc] = output->ic.optprop.ssa [iv][lc];
    		  	      
    output->mc.g1aer[nlyr-1-lc] = output->aer.optprop.g1[iv][lc];
    output->mc.g2aer[nlyr-1-lc] = output->aer.optprop.g2[iv][lc];
    output->mc.ffaer[nlyr-1-lc] = output->aer.optprop.ff[iv][lc];

    output->mc.g1cld[nlyr-1-lc] = output->wc.optprop.g1 [iv][lc];
    output->mc.g2cld[nlyr-1-lc] = output->wc.optprop.g2 [iv][lc];
    output->mc.ffcld[nlyr-1-lc] = output->wc.optprop.ff [iv][lc];

    output->mc.g1ice[nlyr-1-lc] = output->ic.optprop.g1 [iv][lc];
    output->mc.g2ice[nlyr-1-lc] = output->ic.optprop.g2 [iv][lc];
    output->mc.ffice[nlyr-1-lc] = output->ic.optprop.ff [iv][lc];

    /* effective droplet radius             */
    /* ??? this is a temporary solution ??? */
    output->mc.recld [nlyr-1-lc] = output->wc.microphys.effr_layer[lc];
    output->mc.lwccld[nlyr-1-lc] = output->wc.microphys.lwc_layer [lc];

    /* Set optical depth and single scattering albedo */
    if (bext < FLT_MIN) 
      bext = FLT_MIN;
    if (bext_md < FLT_MIN) 
      bext_md = FLT_MIN;

    output->dtauc[lc]    = bext;
    output->dtauc_md[lc] = bext_md;
    output->ssalb[lc] = bsca/bext;

    if ( 1.0-output->ssalb[lc] < FLT_MIN ) 
      output->ssalb[lc] = 1.0 - FLT_MIN;
    
    if (input.rte.solver == POLRADTRAN ) {
      if ( (f = fopen(&output->atm.pol_scat_files[lc*64], "w")) == NULL) return 1;
      fprintf(f,"%e\n",bext);
      fprintf(f,"%e\n",bsca);
      fprintf(f,"%e\n",bsca/bext);

      fprintf(f,"%d\n", input.rte.nstr); 

      momag1=bscaa;
      momag2=bscaa;

      momig1=bscai;
      momig2=bscai;

      momdg1=bscad;
      momdg2=bscad;

      for (k=0; k<=output->atm.nmom; k++) { 
	if (output->aer.optprop.nmom[iv] > 0) {
	  if (k<output->aer.optprop.nmom[iv][lc])
	    moma = bscaa * output->aer.optprop.moment[iv][lc][k] / 
	      output->aer.optprop.moment[iv][lc][0];
	  else 
	    moma=0;
	}
	else {
	  if (k>0) {
	    momag1 *= g1a;
	    momag2 *= g2a;
	  }
	  momd = momag1*fa + (1.0-fa)*momag2;
	}

	if (k>0) {
	  momdg1 *= g1d;
	  momdg2 *= g2d;
	}
	momd = momdg1*fd + (1.0-fd)*momdg2;

	if (k>0) {
	  momig1 *= g1i;
	  momig2 *= g2i;
	}
	momi = momig1*fi + (1.0-fi)*momig2;
	
	/* The first numbers in the columns are the Legendre coefficients
	   for the Rayleigh scattering phase matrix. Note that disort wants
	   phase function moments g_l while polradtran wants Legendre 
	   coefficients C_l = (2*l+1)g_l. p1 is the scalar story of everything. */
	if (k==0 && input.rayleigh) {     
	  p1 =  1.0;
	  p2 = -0.5;
	  p3 =  0.0;
	  p4 =  0.0;
	  p5 =  1.0;
	  p6 =  0.0;
	} 
	else if (k==1 && input.rayleigh) { 
	  p1 =  (0.0 + (2*k+1)*(momd + moma + momi) )/sumsca; 
	  p2 =  0.0; 
	  p3 =  1.5; 
	  p4 =  0.0; 
	  p5 =  0.0; 
	  p6 =  1.5; 
	} 
	else if (k==2 && input.rayleigh) { 
	  p1 =  0.5*bscar/sumsca + (2*k+1)*(momd + moma + momi)/ sumsca; 
	  p2 =  0.5; 
	  p3 =  0.0; 
	  p4 =  0.0; 
	  p5 =  0.5; 
	  p6 =  0.0; 
	}
	else {
	  if (k==0) { p1 = 1.0; } 
	  else {      p1 =  (0.0+ (2*k+1)*(momd + moma + momi)/sumsca );  }
	  p2 =  0.0; 
	  p3 =  0.0; 
	  p4 =  0.0; 
	  p5 =  0.0; 
	  p6 =  0.0; 
	} 
	fprintf(f,"%d %f %f %f %f %f %f\n", k, p1, p2, p3, p4, p5, p6); 
      }
      fclose(f);
    }
}


  output->mc.dtair[nlyr] = 0.0;
  output->mc.dtmol[nlyr] = 0.0;
  output->mc.dtaer[nlyr] = 0.0; 
  output->mc.dtcld[nlyr] = 0.0;
  output->mc.dtice[nlyr] = 0.0;
                     
  output->mc.omaer[nlyr] = 0.0;
  output->mc.omcld[nlyr] = 0.0;
  output->mc.omice[nlyr] = 0.0;

  output->mc.omaer[nlyr] = 0.0;
  output->mc.omcld[nlyr] = 0.0;
  output->mc.omice[nlyr] = 0.0;

  /* loop includes nlyr because altitude and temperature */
  /* are defined per level                               */
  for (lc=0; lc<=nlyr; lc++) {
    output->mc.z     [nlyr-lc] = output->atm.zd[lc];
    output->mc.temper[nlyr-lc] = output->atm.microphys.temper[lc];
  }

  /* loop stops at nlyr-1 because aerosol properties */
  /* are defined per layer                           */
  for (lc=0; lc<nlyr; lc++) {

    output->mc.nmomaer[nlyr-lc-1] = output->aer.optprop.nmom[iv][lc];
    output->mc.momaer [nlyr-lc-1] = calloc (output->mc.nmomaer[nlyr-lc-1], sizeof(float));
    
    /* ??? should free this memory later ??? */

    for (k=0; k<output->mc.nmomaer[nlyr-lc-1]; k++)
      output->mc.momaer[nlyr-lc-1][k] = 
	output->aer.optprop.moment[iv][lc][k] / output->aer.optprop.moment[iv][lc][0];
  }

  
  if (input.verbose) {
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");
    fprintf (stderr, "%5s | %6.2f | %9.6f | %9.6f %9.6f %5.3f | %11.6f %11.6f %5.3f | %11.6f %11.6f %5.3f | %11.6f\n", 
	     "sum", 0.0/0.0,
	     bscartot,
	     bscaatot, babsatot, 0.0/0.0,
	     bscadtot, babsdtot, 0.0/0.0,
	     bscaitot, babsitot, 0.0/0.0,
	     babsotot);
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");
  }
	   

  return 0;
}



/***********************************************************************************/
/* Function: crs_rayleigh                                                          */
/* Description:                                                                    */
/*    Calculates the Rayleigh scattering cross section according to the formula    */
/*    by Nicolet, M., `On the molecular scattering in the terrestrial atmosphere:  */
/*    an empirical formula for its calculation in the homosphere', Planet. Space   */
/*    Sci., 32, 1467-1468, 1984.                                                   */
/*                                                                                 */
/* Parameters:                                                                     */
/*  float* lambda:         Pointer to ary holding the wavelength in nanometers.    */
/*  int n_lambda:          Number of wavelength points.                            */
/*                                                                                 */
/* Return value:                                                                   */
/*  double pointer to ary holding the Rayleigh scattering cross section as a       */
/*  function of wavelength.                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

int crs_rayleigh (float *lambda, int n_lambda, float ***crs) 
{
  float ray_const  = 4.02e-28;
  float from_nm_um = 1.0e-3;
  float lambda_um=0.0, x=0.0;
  int iv=0;
  
  for (iv=0; iv<n_lambda; iv++) {

    lambda_um = lambda[iv] * from_nm_um;

    if (lambda_um > 0.55)
      x = 0.04;
    else
      x = 0.389*lambda_um + 0.09426/lambda_um - 0.3228;
    
    crs[0][MOL_AIR][iv] = ray_const / pow(lambda_um,4+x);
  }

  return 0;
}



/***********************************************************************************/
/* Sum the 3D results over wavelength and print data to mc.sum                     */ 
/***********************************************************************************/

int integrate3D (input_struct input, output_struct *output)
{
  int is=0, js=0, ks=0, iv=0;
  double rfldir=0, rfldn=0, flup=0, absorption=0;

  char sumfilename[FILENAME_MAX] = "";
  char absfilename[FILENAME_MAX] = "";
  char radfilename[FILENAME_MAX] = "";

  FILE *fsum=NULL, *fabs=NULL, *frad=NULL;
  
  /* generate output file names */
  strcpy (sumfilename, input.rte.mc.filename[FN_MC_BASENAME]);
  strcpy (absfilename, input.rte.mc.filename[FN_MC_BASENAME]);
  strcpy (radfilename, input.rte.mc.filename[FN_MC_BASENAME]);

  strcat (sumfilename, ".sum");
  strcat (radfilename, "rad.spc");


  /* extension for absorption/heating/actinic etc. file */
  switch (input.rte.mc.absorption) {
  case MCABS_ACTINIC:
    strcat (absfilename, "act.sum");
    break;

  case MCABS_ABSORPTION:
  case MCABS_EMISSION:
  case MCABS_HEATING:
  case MCABS_NONE:
    strcat (absfilename, "abs.sum");
    break;

  default:
    fprintf (stderr, "Error, unknown absorption type %d\n", 
	     input.rte.mc.absorption);
    break;
  }



  if ((fsum = fopen (sumfilename, "w"))==NULL) {
    fprintf (stderr, "Error opening %s for writing\n", sumfilename);
    return -1;
  }

  if ((fabs = fopen (absfilename, "w"))==NULL) {
    fprintf (stderr, "Error opening %s for writing\n", absfilename);
    return -1;
  }

  if ((frad = fopen (radfilename, "w"))==NULL) {
    fprintf (stderr, "Error opening %s for writing\n", radfilename);
    return -1;
  }

  for (ks=0; ks<input.atm.nzout; ks++) {
    for (is=0; is<input.rte.mc.Nx_sample; is++) {
      for (js=0; js<input.rte.mc.Ny_sample; js++) {
	
	rfldir=0;
	rfldn=0;
	flup=0;
	
	for (iv=0;iv<output->wl.nlambda_h;iv++) {
	  rfldir   += output->rfldir3d   [ks][is][js][iv];
	  rfldn    += output->rfldn3d    [ks][is][js][iv];
	  flup     += output->flup3d     [ks][is][js][iv];

	  fprintf (frad, "%d %d %d %d %g\n", iv, is, js, ks, output->radiance3d[ks][is][js][iv]);
	}
	
	fprintf (fsum, "%d %d %d %g %g %g\n", is, js, ks, rfldir, rfldn, flup);
      }
    }
  }
  

  if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
    for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
      for (is=0; is<input.rte.mc.Nx_cloud; is++)
	for (js=0; js<input.rte.mc.Ny_cloud; js++) {
	  absorption = 0;
	  for (iv=0;iv<output->wl.nlambda_h;iv++)
	    absorption += output->abs3d[ks][is][js][iv];
	  
	  fprintf (fabs, "%d %d %d %g\n", is, js, ks, absorption);
	}

  
  
  fclose(fsum); fclose(fabs); fclose(frad);

  return 0;
}
	  

/***********************************************************************************/
/* Sum the 1D results over wavelength and write data to the first array element    */ 
/***********************************************************************************/

int sum1D (input_struct input, output_struct *output)
{ 
  int lev=0, iv=0, iu=0, j=0, is=0, status=0;
  double ffactor=0, rfactor=0;
  double incident=0;

  /* calculate incident flux */
  for (iv=0; iv<output->wl.nlambda_h; iv++)
    incident += output->wl.filter[iv] * output->wl.fbeam[iv] * cos(output->atm.sza_t[iv]*PI/180.0);

  /* sum extraterrestrial irradiance */
  for (iv=1; iv<output->wl.nlambda_h; iv++)
    output->wl.fbeam[0] += output->wl.filter[iv] * output->wl.fbeam[iv];
      
  /* fluxes and radiances */
  for (iv=1; iv<output->wl.nlambda_h; iv++) {

    for (lev=0; lev<input.atm.nzout; lev++) {
      output->rfldir[lev][0] += output->rfldir[lev][iv];
      output->rfldn [lev][0] += output->rfldn [lev][iv];
      output->flup  [lev][0] += output->flup  [lev][iv];
      output->uavg  [lev][0] += output->uavg  [lev][iv];
      output->uavgso[lev][0] += output->uavgso[lev][iv];
      output->uavgdn[lev][0] += output->uavgdn[lev][iv];
      output->uavgup[lev][0] += output->uavgup[lev][iv];

      for (iu=0; iu<input.rte.maxumu; iu++) {
        output->u0u[lev][iu][0] += output->u0u[lev][iu][iv];

	for (j=0; j<input.rte.nphi; j++)
	  output->uu[lev][j][iu][0] += output->uu[lev][j][iu][iv];
      }
    }
  }

  /* polarized fluxes and radiances */
  if (input.rte.solver == POLRADTRAN) 
    for (iv=0; iv<output->wl.nlambda_h; iv++) {

      for (lev=0; lev<input.atm.nzout; lev++) 
	for (is=0; is<input.rte.pol_nstokes; is++) {
	  output->down_flux [lev][is][0] += output->down_flux [lev][is][iv];
	  output->up_flux   [lev][is][0] += output->up_flux   [lev][is][iv];

	  for (iu=0; iu<input.rte.nstr/2; iu++)
	    for (j=0; j<input.rte.nphi; j++) {
	      output->down_rad [lev][j][iu][is][0] += output->down_rad [lev][j][iu][is][iv]; 
	      output->up_rad   [lev][j][iu][is][0] += output->up_rad   [lev][j][iu][is][iv];
	    }
	}
    }
    
  /* if transmittance, then divide by the extraterrestrial flux */
  if (input.source != SRC_THERMAL) {
    switch (input.calibration) {
    case OUTCAL_ABSOLUTE:
      ffactor = 1.0;   /* irradiance multiplicator */
      rfactor = 1.0;   /* radiance multiplicator   */
      break;

    case OUTCAL_TRANSMITTANCE:
    case OUTCAL_BRIGHTNESS:
      ffactor = 1.0 / output->wl.fbeam[0];   /* irradiance multiplicator */
      rfactor = 1.0 / output->wl.fbeam[0];   /* radiance multiplicator   */
      break;

    case OUTCAL_REFLECTIVITY:

      ffactor = 1.0 / incident;
      rfactor = PI  / incident;
      break;
      
    default:
      fprintf (stderr, "Error, unknown output calibration %d\n", input.calibration);
      return -1;
    }

    /**************************************************************/
    /* now scale irradiances with ffactor, radiances with rfactor */
    /**************************************************************/

    status = scale_output (input, output, ffactor, rfactor, 0);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by scale_output()\n", status);
      return status;
    }
  }
  else {
    if (input.calibration==OUTCAL_BRIGHTNESS) {

      fprintf (stderr, " ... converting radiances to brightness temperatures\n");
      fflush (stderr);

      /* convert irradiances / radiances to brightness temperatures */
      status = output2bt (input, output);
      if (status!=0) {
	fprintf (stderr, "Error %d returned by output2bt()\n", status);
	return status;
      }
    }
  }
  
  output->wl.nlambda_h = 1;

  return 0;
}


/***********************************************************************************/
/* Convert radiance to brightness temperature for a given filter function.         */ 
/***********************************************************************************/

float radiance2bt (float rad, float *wvnmlo, float *wvnmhi, float *filter, int n)
{
  int iv=0;
  float t1=0, t2=0, plkrad1=0, plkrad2=0, r=0;

  float accur = 1e-5;
  float dt=0.001;
  
  t1=273;
  t2 = t1+dt;

  if (rad<=0)
    return 0;
    

  plkrad1=0;
  plkrad2=0;
  for (iv=0; iv<n; iv++) {
    F77_FCN (cplkavg) (&(wvnmlo[iv]), &(wvnmhi[iv]), &t1, &r);
    plkrad1 += filter[iv] * r;

    F77_FCN (cplkavg) (&(wvnmlo[iv]), &(wvnmhi[iv]), &t2, &r);
    plkrad2 += filter[iv] * r;
  }


  while (fabs((plkrad1 - rad)/rad) > accur) {
    t1 = t1+(rad-plkrad1)/(plkrad2-plkrad1)*dt;
    t2 = t1+dt;

    plkrad1=0;
    plkrad2=0;
    for (iv=0; iv<n; iv++) {
      F77_FCN (cplkavg) (&(wvnmlo[iv]), &(wvnmhi[iv]), &t1, &r);
      plkrad1 += filter[iv] * r;

      F77_FCN (cplkavg) (&(wvnmlo[iv]), &(wvnmhi[iv]), &t2, &r);
      plkrad2 += filter[iv] * r;
    }
  }

  return t1;
}




/***********************************************************************************/
/* Scale uvspec output fields with a given factor; ffactor is the scaling factor   */ 
/* for irradiances/actinic fluxes and rfactor is the scaling factor for radiances  */
/***********************************************************************************/

static int scale_output (input_struct input, output_struct *output, 
			 double ffactor, double rfactor, int iv)
{
  int lev=0, iu=0, j=0, is=0, js=0, ks=0;
  

  if (input.rte.solver == POLRADTRAN) {  /* polRadtran output */
    for (lev=0; lev<input.atm.nzout; lev++) 
      for (is=0; is<input.rte.pol_nstokes; is++) {

	/* irradiances */
	output->down_flux [lev][is][iv] *= ffactor;
	output->up_flux   [lev][is][iv] *= ffactor;
	
	/* radiances   */
	for (iu=0; iu<input.rte.nstr/2; iu++)
	  for (j=0; j<input.rte.nphi; j++) {
	    output->down_rad [lev][j][iu][is][iv] *= rfactor;
	    output->up_rad   [lev][j][iu][is][iv] *= rfactor;
	  }
      }
  }
  else { /* unpolarized output */

    for (lev=0; lev<input.atm.nzout; lev++) {

      /* irradiance / actinic flux */
      output->rfldir[lev][iv] *= ffactor;
      output->rfldn [lev][iv] *= ffactor;
      output->flup  [lev][iv] *= ffactor;
      output->uavg  [lev][iv] *= ffactor;
      output->uavgso[lev][iv] *= ffactor;
      output->uavgdn[lev][iv] *= ffactor;
      output->uavgup[lev][iv] *= ffactor;
      
      /* radiances   */
      for (iu=0; iu<input.rte.maxumu; iu++) {
	output->u0u[lev][iu][iv] *= rfactor;
	
	for (j=0; j<input.rte.nphi; j++)
	  output->uu[lev][j][iu][iv] *= rfactor;
      }

      /* 3D irradiances and radiances */
      if (input.rte.mc.passback3D)
	for (is=0; is<input.rte.mc.Nx_sample; is++)
	  for (js=0; js<input.rte.mc.Ny_sample; js++) {
	    output->rfldir3d   [lev][is][js][iv] *= ffactor;
	    output->rfldn3d    [lev][is][js][iv] *= ffactor;
	    output->flup3d     [lev][is][js][iv] *= ffactor;
	    output->radiance3d [lev][is][js][iv] *= rfactor;
	  }
    }
  
    /* 3D absorption fields; here the loop goes over all 3D boxes */
    if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
      for (ks=0; ks<input.rte.mc.Nz_cloud; ks++) 
	for (is=0; is<input.rte.mc.Nx_cloud; is++) 
	  for (js=0; js<input.rte.mc.Ny_cloud; js++) 
	    output->abs3d [ks][is][js][iv] *= ffactor;
    
  }
    

  return 0;
}




/***********************************************************************************/
/* Convert uvspec output to brightness temperatures; currently it is assumed that  */
/* the first element (iv==0) stores the spectrally integrated quantities;          */
/* only these will be considered.                                                  */
/***********************************************************************************/

static int output2bt (input_struct input, output_struct *output)
{
  int lev=0, iu=0, j=0, is=0, js=0;


  if (input.rte.solver == POLRADTRAN) { /* polRadtran output */
    for (lev=0; lev<input.atm.nzout; lev++) 
      for (is=0; is<input.rte.pol_nstokes; is++) {
	
	output->down_flux[lev][is][0] = 
	  radiance2bt (output->down_flux[lev][is][0] / PI,
		       output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
	
	output->up_flux[lev][is][0] = 
	  radiance2bt (output->up_flux[lev][is][0] / PI,
		       output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
	
	/* radiances */
	for (iu=0; iu<input.rte.nstr/2; iu++)
	  for (j=0; j<input.rte.nphi; j++) {
	    
	    output->down_rad [lev][j][iu][is][0] = 
	      radiance2bt (output->down_rad [lev][j][iu][is][0],
			   output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
	    
	    output->up_rad [lev][j][iu][is][0] = 
	      radiance2bt (output->up_rad [lev][j][iu][is][0],
			   output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
	  }
      }
  }
  else { /* unpolarized output */
    
    for (lev=0; lev<input.atm.nzout; lev++) {
      
      /* irradiance / actinic flux */
      output->rfldir[lev][0] = 
	radiance2bt (output->rfldir[lev][0] / PI,
		     output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
      
      output->rfldn [lev][0] = 
	radiance2bt (output->rfldn [lev][0] / PI,
		     output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
      
      output->flup [lev][0] = 
	radiance2bt (output->flup [lev][0] / PI,
		     output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
      
      output->uavg [lev][0] = 
	radiance2bt (output->uavg [lev][0] / PI,
		     output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
      
      output->uavgso[lev][0] = 
	radiance2bt (output->uavgso[lev][0] / PI,
		     output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
      
      output->uavgdn[lev][0] = 
	radiance2bt (output->uavgdn[lev][0] / PI,
		     output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
      
      output->uavgup[lev][0] = 
	radiance2bt (output->uavgup[lev][0] / PI,
		     output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);      

      /* radiances */
      for (iu=0; iu<input.rte.maxumu; iu++) {
	output->u0u[lev][iu][0] = 
	  radiance2bt (output->u0u[lev][iu][0],
		       output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
	
	for (j=0; j<input.rte.nphi; j++)
	  output->uu[lev][j][iu][0] = 
	    radiance2bt (output->uu[lev][j][iu][0],
			 output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
      }
      
      /* 3D fields */
      if (input.rte.mc.passback3D)
	for (is=0; is<input.rte.mc.Nx_sample; is++)
	  for (js=0; js<input.rte.mc.Ny_sample; js++) {

	    output->rfldir3d[lev][is][js][0] = 
	      radiance2bt (output->rfldir3d[lev][is][js][0] / PI,
			   output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);

	    output->rfldn3d[lev][is][js][0] = 
	      radiance2bt (output->rfldn3d[lev][is][js][0] / PI,
			   output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);

	    output->flup3d[lev][is][js][0] = 
	      radiance2bt (output->flup3d[lev][is][js][0] / PI,
			   output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);

	    output->radiance3d [lev][is][js][0] = 
	      radiance2bt (output->radiance3d [lev][is][js][0],
			   output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
	  }
    }
    
    /* 3D absorption fields; here the loop goes over all 3D boxes */

    /* conversion of 3D absorption to BT is probably useless, hence we don't do it
    if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
      for (ks=0; ks<input.rte.mc.Nz_cloud; ks++) 
        for (is=0; is<input.rte.mc.Nx_cloud; is++) 
	  for (js=0; js<input.rte.mc.Ny_cloud; js++) 
	    output->abs3d [ks][is][js][0] = 
	      radiance2bt (output->abs3d [ks][is][js][0] /PI, 
	       output->wl.wvnmlo_t, output->wl.wvnmhi_t, output->wl.filter, output->wl.nlambda_t);
    */
			   
  }
    

  return 0;
}





static int read_photon_file (char *filename, float *lambda_t, int nlambda_t, float **fraction)
{
  float *wvl=NULL;
  int iv=0, n=0, status=0;
  
  status = read_2c_file_float (filename, &wvl, fraction, &n);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* compare wavelength grids */
  if (nlambda_t != n) {
    fprintf (stderr, "Error, number of wavelengths in %s differing from\n", filename);
    fprintf (stderr, "internal wavelength grid\n");
    return -1;
  }
  
  for (iv=0; iv<n; iv++)
    if (lambda_t[iv] != wvl[iv]) {
      fprintf (stderr, "Error, wavelengths in %s differ from internal wavelength grid: %f vs. %f\n", 
	       filename, lambda_t[iv], wvl[iv]);
      return -1;
    }
  
  return 0;
}
