/*--------------------------------------------------------------------
 * $Id: atmosphere.c,v 1.42 2003/01/21 08:00:31 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <ckdfu.h>
#include "uvspec.h"
#include "cloud.h"
#include "../src_c/ascii.h"
#include "../src_c/radiative.h"
#include "../src_f/f77-uscore.h"

#define EPSILON 1E-6

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif


#define M_H2O 18.015
#define N_MOL_CKDFU 6.0235e23



/************************************/
/* prototypes of internal functions */
/************************************/

static int read_atmosphere (char *filename,
			    atm_out_struct *atm_out, 
			    int cdf, int quiet, int verbose);

static int read_density (char *filename,
			 atm_out_struct *atm_out, int dens_id, 
			 int quiet, int verbose); 

static int read_tau (char *filename, int nlyr,
		     float* zd, float** dtau);

static int read_molecular_absorption (char *filename, int nlyr,
				      float *zd, int quiet,
				      float **dtau_mono, 
				      float ***dtau_spec, float **lambda, int *nlambda,
				      int *monochromatic);

static int column (float *dens, float *zd, int nlev, float altitude, float *column);
static int scale_profile (float *dens, float *zd, int nlev, float newcol, float altitude, 
			  float *factor);

static int precip_water (float *dens, float *zd, int nlev, float altitude, float *precip);
static int scale_precip (float *dens, float *zd, int nlev, float precip, float altitude, float *factor);

static double vapor_pressure (double t);

static int read_and_convert_rh (char *filename,
				float *zd, float *t, int nlev,
				float *rh, float *h2o);

static int read_nc_atmosphere_grid (float **zd, int *nlev, char *filename);


/**************************************************************/
/* Setup model atmosphere.                                    */
/**************************************************************/

int setup_atmosphere (input_struct input, output_struct *output)
{
  int i=0, status=0, lc=0, iz=0, ialt=0, cdf=0;
  float scale_factor=0;
  float *zuser=NULL;
  int nuser=0;

  char filename[FILENAME_MAX];

  float *tmp_zd=NULL;
  int tmp_nlev=0;

  float *tmp_lambda=NULL, **tmp_dtau=NULL;
  int tmp_nlambda=0;

  float n_0=0, t_0=0, p_0=0, ar=0;

  int monochromatic=0;


  /* where to get the atmosphere from */
  if (input.ck_scheme == CK_GENERIC) {
    strcpy (filename, input.filename[FN_CK_GENERIC]);
    cdf=1;

    if (!input.quiet)
      if (strlen(input.filename[FN_ATMOSPHERE]) > 0)
	fprintf (stderr, " ... ignoring %s, reading atmosphere from %s\n",
		 input.filename[FN_ATMOSPHERE], input.filename[FN_CK_GENERIC]);
    
    /* determine altitude grid */
    status = read_nc_atmosphere_grid (&(output->atm.zd), &(output->atm.nlev), filename);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_nc_atmosphere_grid()\n", status);
      return status;
    }
  }
  else {
    strcpy (filename, input.filename[FN_ATMOSPHERE]);
    cdf=0;

    /* determine altitude grid */
    status = read_1c_file_float (filename, &(output->atm.zd), &(output->atm.nlev));
    if (status!=0) {
      fprintf (stderr, "Error %d determining altitude profile from %s\n", status, filename);
      return status;
    }
  }


  /* first, determine the complete set of altitudes           */
  /* for the atmosphere; to do that, combine the z-grids      */
  /* of the atmospheric profiles (atmosphere_file, dens_file) */
  /* and the relative humidity profile                        */
  
  status=0;

  /* relative humidity */
  if (strlen(input.filename[FN_RH]) > 0)
    status += add_file2grid (&(output->atm.zd), &(output->atm.nlev),
			     input.filename[FN_RH]);


  /* refractive index */
  if (strlen(input.filename[FN_REFIND]) > 0)
    status += add_file2grid (&output->atm.zd, &output->atm.nlev, 
			     input.filename[FN_REFIND]);

  /* densities */
  for (i=0; i<MOL_NN; i++)
    if (strlen(input.atm.filename[i]) > 0)
      status += add_file2grid (&(output->atm.zd), &(output->atm.nlev),
			       input.atm.filename[i]);

  /* add user-defined altitude as level */
  if (input.atm.altitude > 0) {
    zuser = (float *) calloc (1, sizeof(float));
    zuser[0] = input.atm.altitude;
    
    set_common_z_grid (zuser, 1, &output->atm.zd, &output->atm.nlev);
    free (zuser);
  }

  /* ??? user-defined equidistant altitude grid ??? */
  /* ??? may be added here; in this case, the   ??? */
  /* ??? the atmospheric grid will be           ??? */
  /* ??? interpolated to the user-define grid   ??? */

  if (status!=0) {
    fprintf (stderr, "Error creating combined atmospheric grid\n");
    return status;
  }


  /* number of layers */
  output->atm.nlyr = output->atm.nlev-1;


  /****************************/
  /* read atmospheric profile */
  /****************************/

  status = read_atmosphere (filename,
			    &output->atm, cdf,
			    input.quiet, input.verbose);

  if (status!=0) {
    fprintf (stderr, "Error %d reading atmosphere from %s\n",
	     status, filename);
    return status;
  }
  

  /*********************/
  /* density profiles  */
  /*********************/

  for (i=0; i<MOL_NN; i++) {
    if (strlen(input.atm.filename[i]) > 0) {
      if (input.verbose)
	fprintf (stderr,"Reading atmosphere file %s\n", input.atm.filename[i]);
      
      status = read_density (input.atm.filename[i],
			     &output->atm, i,
			     input.quiet, input.verbose);
      if (status !=0) {
	fprintf (stderr, "Error %d reading density profile\n", status);
	return status;
      }
    }
  }


  /*********************/
  /* relative humidity */
  /*********************/

  /* allocate memory in any case */
  output->atm.microphys.rh = (float *) calloc (output->atm.nlev, sizeof(float));
  if (output->atm.microphys.rh == NULL) {
    fprintf (stderr, "Error allocating memory for output->atm.microphys.rh\n");
    return -1;
  }
  
  /* read_and_convert_rh() needs to be called in any case */
  /* because it calculates the otherwise empty rh profile */
  /* even if it doesn't read rh from the file             */
  status = read_and_convert_rh (input.filename[FN_RH], 
				output->atm.zd, 
				output->atm.microphys.temper, output->atm.nlev,
				output->atm.microphys.rh,
				output->atm.microphys.dens[MOL_H2O]);

  if (status!=0) {
    fprintf (stderr, "Error %d returned by read_and_convert_rh()\n", status);
    return status;
  }


  /*********************/
  /* refractive index  */
  /*********************/

  /* allocate memory for refractive index array */
  output->atm.microphys.refind = (float *) calloc (output->atm.nlev, sizeof(float));
  if (output->atm.microphys.refind == NULL) {
    fprintf (stderr, "Error allocating memory for output->atm.microphys.refind\n");
    return -1;
  }
    
  /* Calculate refractive index of air */
  if (input.rte.nrefrac > 0) {
    if (strlen(input.filename[FN_REFIND])>0) {
      fprintf (stderr, "Sorry, refractive_index_file not yet implemented\n");
      return -1;
    }
    else {
      n_0 = 1.000293; /* refraction under normal conditions */      
      t_0 = 15.0;     /* normal temperature in celsius      */
      p_0 = 1013.25;  /* normal pressure in mbar            */
      ar = 0.00366;
      for (lc=0; lc<output->atm.nlev; lc++) {
	output->atm.microphys.refind[lc] = 1.0 + (n_0-1.0) *
	  (1.0 + ar*t_0) / (1.0 + ar * (output->atm.microphys.temper[lc]-273.15)) *
	  output->atm.microphys.press[lc] / p_0;    
      }
    }
  } 
  else {
    for (lc=0; lc<output->atm.nlev; lc++)
      output->atm.microphys.refind[lc] = 1.0;
  }


  /**************************************************/
  /* calculate the common atmospheric profile which */
  /* contains all possible inputs; this profile     */
  /* will form the basis for the atmospheric        */
  /* optical properties                             */
  /**************************************************/

  tmp_nlev = 0; 
  set_common_z_grid (output->atm.zd, output->atm.nlev, 
		     &tmp_zd, &tmp_nlev);
  
  
  status=0;
  if (strlen(input.filename[FN_RAYLEIGH]) > 0)
    status += add_file2grid (&tmp_zd, &tmp_nlev,
			     input.filename[FN_RAYLEIGH]);

  if (strlen(input.filename[FN_MOLABS]) > 0)
    status += add_file2grid (&tmp_zd, &tmp_nlev,
			     input.filename[FN_MOLABS]);

  if (status!=0) {
    fprintf (stderr, "Error creating common atmospheric grid\n");
    return status;
  }

  /*
  fprintf (stderr, "common atmospheric profile before interpolation:\n");
  for (lc=0; lc<tmp_nlev; lc++)
    fprintf (stderr, "%2d %6.2f %6.2f %6.2f\n", 
	     lc, tmp_zd[lc], output->atm.microphys.press[lc], output->atm.microphys.temper[lc]);
  fflush (stderr);
  */


  /*****************************************************************************/
  /* interpolate microphysics profile to common grid tmp_zd [0 ... tmp_nlev-1] */
  /*****************************************************************************/

  /* pressure, logarithmic */
  status=0;
  status += interpolate_profile (output->atm.zd, &(output->atm.microphys.press), output->atm.nlev,
				 tmp_zd, tmp_nlev, 2);


  /* temperature, linear */
  status += interpolate_profile (output->atm.zd, &(output->atm.microphys.temper), output->atm.nlev,
				 tmp_zd, tmp_nlev, 1);


  /* refind, linear */
  status += interpolate_profile (output->atm.zd, &(output->atm.microphys.refind), output->atm.nlev,
				 tmp_zd, tmp_nlev, 1);

  /* relative humidity, linear */
  status += interpolate_profile (output->atm.zd, &(output->atm.microphys.rh), output->atm.nlev,
				 tmp_zd, tmp_nlev, 1);


  /* densities, logarithmic */
  for (i=0; i<MOL_NN; i++)
    status += interpolate_profile (output->atm.zd, &(output->atm.microphys.dens[i]), output->atm.nlev,
				   tmp_zd, tmp_nlev, 2);

  /* denstab and denstab_amf, logarithmic */
  if (output->atm.microphys.denstab_id > 0) {
    for (i=0; i<output->atm.microphys.nsza_denstab; i++) {
      status += interpolate_profile (output->atm.zd, 
				     &(output->atm.microphys.denstab[output->atm.microphys.denstab_id][i]), 
				     output->atm.nlev, tmp_zd, tmp_nlev, 2);
      status += interpolate_profile (output->atm.zd, 
				     &(output->atm.microphys.denstab_amf[i]), 
				     output->atm.nlev, tmp_zd, tmp_nlev, 2);
    }
  }


  if (status!=0) {
    fprintf (stderr, "Error interpolating profiles to common atmospheric grid\n");
    return status;
  }

  /* copy altitude grid to final destination */
  free (output->atm.zd);
  output->atm.zd = calloc (tmp_nlev, sizeof(float));
  for (lc=0; lc<tmp_nlev; lc++)
    output->atm.zd[lc] = tmp_zd[lc];

  output->atm.nlev = tmp_nlev;
  output->atm.nlyr = tmp_nlev-1;

  free (tmp_zd);



  for (ialt=0; ialt<output->atm.nlev; ialt++)
    if (input.atm.altitude == output->atm.zd[ialt])
      break;
    
  if (ialt == output->atm.nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid.\n", 
	     input.atm.altitude);
    return -1;
  }

  /* Scale profiles of pressure and air density */
  if (input.pressure >= 0.0) {

    scale_factor = input.pressure / output->atm.microphys.press[ialt];
    for (lc=0; lc<output->atm.nlev; lc++) {
      output->atm.microphys.press[lc]         *= scale_factor;
      output->atm.microphys.dens[MOL_AIR][lc] *= scale_factor;
    } 
    
    if (input.verbose) {
      fprintf (stderr, " ... scaling pressure at %f km to %7.2f mbar\n", 
	       output->atm.zd[ialt], output->atm.microphys.press[ialt]);
      fprintf (stderr, " ... scaling density at %f km to %.6e\n", 
	       output->atm.zd[ialt], output->atm.microphys.dens[MOL_AIR][ialt]);
    }
  }

  /* Scale ozone column */
  status = column (output->atm.microphys.dens[MOL_O3], 
		   output->atm.zd, output->atm.nlev, input.atm.altitude, &(output->ozone_column));
  if (status!=0) {
    fprintf (stderr, "Error %d calculating ozone column\n", status);
    return status;
  }

  if ( input.atm.column[MOL_O3] >= 0.0 ) {
    if ( input.atm.unit[MOL_O3] != DU) {
      fprintf (stderr, " ... Unknown unit for ozone");
      status =-1;
      return status;
    }
    scale_profile (output->atm.microphys.dens[MOL_O3], 
		   output->atm.zd, output->atm.nlev, 
		   input.atm.column[MOL_O3], input.atm.altitude,
		   &(output->ozone_scale_factor));
    
    status = column (output->atm.microphys.dens[MOL_O3], 
		     output->atm.zd, output->atm.nlev, input.atm.altitude, &(output->ozone_column)); 
    if (status!=0) {
      fprintf (stderr, "Error %d calculating column\n", status);
      return status;
    }
  }

  /* Scale NO2 column */
  status = column (output->atm.microphys.dens[MOL_NO2], 
		   output->atm.zd, output->atm.nlev, input.atm.altitude, &(output->no2_column));
  if (status!=0) {
    fprintf (stderr, "Error %d calculating column\n", status);
    return status;
  }
  
  if (input.atm.column[MOL_NO2] >= 0.0) {
    if ( input.atm.unit[MOL_NO2] == DU ) {
      /* Everything ok, do nothing */
    }
    else if (input.atm.unit[MOL_NO2] == CM_2) {
      input.atm.column[MOL_NO2] /= 2.687E16;
    }
    else {
      fprintf (stderr, " ... Unknown unit for nitrogen dioxide");
      status =-1;
      return status;
    }
    scale_profile (output->atm.microphys.dens[MOL_NO2], 
		   output->atm.zd, output->atm.nlev, 
		   input.atm.column[MOL_NO2], input.atm.altitude,
		   &(output->no2_scale_factor));
    
    status = column (output->atm.microphys.dens[MOL_NO2], 
		     output->atm.zd, output->atm.nlev, input.atm.altitude, &(output->no2_column)); 
    if (status!=0) {
      fprintf (stderr, "Error %d calculating column\n", status);
      return status;
    }

    /* If density file multiply with column value, hence it is a scale factor      */
    /* and the name is misleading, however, this is only useful for airmass        */
    /* factor calculations for which the column is 0.0 when no absorber is present */
    if (output->atm.microphys.denstab_id == MOL_NO2) {
      for (i=1; i<=output->atm.microphys.nsza_denstab; i++) {
	for (lc=0; lc<output->atm.nlev; lc++) {
	  output->atm.microphys.denstab[output->atm.microphys.denstab_id][i-1][lc] *= input.atm.column[MOL_NO2];
	  output->atm.microphys.denstab_amf[i-1][lc]                     *= input.atm.column[MOL_NO2];
	}
      }
    }
  }

  /* Scale BrO column */
  status = column (output->atm.microphys.dens[MOL_BRO], 
		   output->atm.zd, output->atm.nlev, input.atm.altitude, &(output->bro_column));
  if (status!=0) {
    fprintf (stderr, "Error %d calculating column\n", status);
    return status;
  }
  
  if ( input.atm.column[MOL_BRO] >= 0.0 ) {
    if ( input.atm.unit[MOL_BRO] == CM_2 ) {
      input.atm.column[MOL_BRO] /= 2.687E16;
    }
    else {
      fprintf (stderr, " ... Unknown unit for bromine oxide");
      status =-1;
      return status;
    }
    scale_profile (output->atm.microphys.dens[MOL_BRO], 
		   output->atm.zd, output->atm.nlev, 
		   input.atm.column[MOL_BRO], input.atm.altitude,
		   &(output->bro_scale_factor));
    
    status = column (output->atm.microphys.dens[MOL_BRO], 
		     output->atm.zd, output->atm.nlev, input.atm.altitude, &(output->bro_column)); 
    if (status!=0) {
      fprintf (stderr, "Error %d calculating column\n", status);
      return status;
    }

    /* If density file multiply with column value, hence it is a scale factor      */
    /* and the name is misleading, however, this is only useful for airmass        */
    /* factor calculations for which the column is 0.0 when no absorber is present */
    if (output->atm.microphys.denstab_id == MOL_BRO) {
      for (i=1; i<=output->atm.microphys.nsza_denstab; i++) {
	for (lc=0; lc<output->atm.nlev; lc++) {
	  output->atm.microphys.denstab[output->atm.microphys.denstab_id][i-1][lc] *= input.atm.column[MOL_BRO];
	  output->atm.microphys.denstab_amf[i-1][lc]                     *= input.atm.column[MOL_BRO];
	}
      }
    }
  }

  /* Scale OClO column */
  status = column (output->atm.microphys.dens[MOL_OCLO], 
		   output->atm.zd, output->atm.nlev, input.atm.altitude, &(output->oclo_column));
  if (status!=0) {
    fprintf (stderr, "Error %d calculating column\n", status);
    return status;
  }
  
  if ( input.atm.column[MOL_OCLO] >= 0.0 ) {
    if ( input.atm.unit[MOL_OCLO] == CM_2 ) {
      input.atm.column[MOL_OCLO] /= 2.687E16;
    }
    else {
      fprintf (stderr, " ... Unknown unit for oclo");
      status =-1;
      return status;
    }
    scale_profile (output->atm.microphys.dens[MOL_OCLO], 
		   output->atm.zd, output->atm.nlev, 
		   input.atm.column[MOL_OCLO], input.atm.altitude,
		   &(output->oclo_scale_factor));
    
    status = column (output->atm.microphys.dens[MOL_OCLO], 
		     output->atm.zd, output->atm.nlev, input.atm.altitude, &(output->oclo_column)); 
    if (status!=0) {
      fprintf (stderr, "Error %d calculating column\n", status);
      return status;
    }

    /* If density file multiply with column value, hence it is a scale factor      */
    /* and the name is misleading, however, this is only useful for airmass        */
    /* factor calculations for which the column is 0.0 when no absorber is present */
    if (output->atm.microphys.denstab_id == MOL_OCLO) {
      for (i=1; i<=output->atm.microphys.nsza_denstab; i++) {
	for (lc=0; lc<output->atm.nlev; lc++) {
	  output->atm.microphys.denstab[output->atm.microphys.denstab_id][i-1][lc] *= input.atm.column[MOL_OCLO];
	  output->atm.microphys.denstab_amf[i-1][lc]                     *= input.atm.column[MOL_OCLO];
	}
      }
    }
  }

  /* Scale HCHO column */
  status = column (output->atm.microphys.dens[MOL_HCHO], 
		   output->atm.zd, output->atm.nlev, input.atm.altitude, &(output->hcho_column));
  if (status!=0) {
    fprintf (stderr, "Error %d calculating column\n", status);
    return status;
  }
  
  if ( input.atm.column[MOL_HCHO] >= 0.0 ) {
    if ( input.atm.unit[MOL_HCHO] == CM_2 ) {
      input.atm.column[MOL_HCHO] /= 2.687E16;
    }
    else {
      fprintf (stderr, " ... Unknown unit for hcho");
      status =-1;
      return status;
    }
    scale_profile (output->atm.microphys.dens[MOL_HCHO], 
		   output->atm.zd, output->atm.nlev, 
		   input.atm.column[MOL_HCHO], input.atm.altitude,
		   &(output->hcho_scale_factor));
    
    status = column (output->atm.microphys.dens[MOL_HCHO], 
		     output->atm.zd, output->atm.nlev, input.atm.altitude, &(output->hcho_column)); 
    if (status!=0) {
      fprintf (stderr, "Error %d calculating column\n", status);
      return status;
    }

    /* If density file multiply with column value, hence it is a scale factor      */
    /* and the name is misleading, however, this is only useful for airmass        */
    /* factor calculations for which the column is 0.0 when no absorber is present */
    if (output->atm.microphys.denstab_id == MOL_HCHO) {
      for (i=1; i<=output->atm.microphys.nsza_denstab; i++) {
	for (lc=0; lc<output->atm.nlev; lc++) {
	  output->atm.microphys.denstab[output->atm.microphys.denstab_id][i-1][lc] *= input.atm.column[MOL_HCHO];
	  output->atm.microphys.denstab_amf[i-1][lc]                     *= input.atm.column[MOL_HCHO];
	}
      }
    }
  }

  /* Scale precipitable water */
  status = precip_water (output->atm.microphys.dens[MOL_H2O], 
			 output->atm.zd, output->atm.nlev,
			 input.atm.altitude, &(output->precip_water));
  
  if (status!=0) {
    fprintf (stderr, "Error %d calculating precipitable water\n", status);
    return status;
  }
  
  if (input.precip_water >= 0.0) {

    if (!input.quiet)
      fprintf (stderr, " ... scaling precipitable water from %e kg/m2 to ", output->precip_water);
    
    status = scale_precip (output->atm.microphys.dens[MOL_H2O], 
			   output->atm.zd, output->atm.nlev, 
			   input.precip_water, input.atm.altitude,
			   &(output->water_scale_factor));
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by scale_precip()\n", status);
      return status;
    }

    status = precip_water (output->atm.microphys.dens[MOL_H2O], 
			   output->atm.zd, output->atm.nlev, 
			   input.atm.altitude, &(output->precip_water)); 

    if (status!=0) {
      fprintf (stderr, "Error %d returned by precip_water()\n", status);
      return status;
    }

    if (!input.quiet)
      fprintf (stderr, "%e kg/m2\n", output->precip_water);
  }



  /* Determine mixing ratios of trace gases at the */
  /* user-defined altitude and scale profiles      */

  for (ialt=0; ialt<output->atm.nlev; ialt++)
    if (input.atm.altitude == output->atm.zd[ialt])
      break;
    
  if (ialt == output->atm.nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid.\n", 
	     input.atm.altitude);
    return -1;
  }

  output->mixing_ratio = calloc (MX_NN, sizeof(float));

  /* O2 */
  if (input.mixing_ratio[MX_O2]<0) {
    /* If not specified, use the mixing ratio at the user-defined altitude */
    output->mixing_ratio[MX_O2] = output->atm.microphys.dens[MOL_O2][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_O2] = input.mixing_ratio[MX_O2];
    
    scale_factor = (input.mixing_ratio[MX_O2] / 1e6) / 
      (output->atm.microphys.dens[MOL_O2] [ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    for (i=0; i<output->atm.nlev; i++)
      output->atm.microphys.dens[MOL_O2][i] *= scale_factor;
  }
  

  /* CO2 */
  if (input.mixing_ratio[MX_CO2]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_CO2] = output->atm.microphys.dens[MOL_CO2][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_CO2] = input.mixing_ratio[MX_CO2];
    
    scale_factor = (input.mixing_ratio[MX_CO2] / 1e6) / 
      (output->atm.microphys.dens[MOL_CO2][ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    for (i=0; i<output->atm.nlev; i++)
      output->atm.microphys.dens[MOL_CO2][i] *= scale_factor;
  }
  

  /* H2O */
  if (input.mixing_ratio[MX_H2O]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_H2O] = output->atm.microphys.dens[MOL_H2O][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_H2O] = input.mixing_ratio[MX_H2O];
    
    scale_factor = (input.mixing_ratio[MX_H2O] / 1e6) / 
      (output->atm.microphys.dens[MOL_H2O][ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    for (i=0; i<output->atm.nlev; i++)
      output->atm.microphys.dens[MOL_H2O][i] *= scale_factor;
  }
  
  /* NO2 */
  if (input.mixing_ratio[MX_NO2]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_NO2] = output->atm.microphys.dens[MOL_NO2][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_NO2] = input.mixing_ratio[MX_NO2];
    
    scale_factor = (input.mixing_ratio[MX_NO2] / 1e6) / 
      (output->atm.microphys.dens[MOL_NO2][ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    for (i=0; i<output->atm.nlev; i++)
      output->atm.microphys.dens[MOL_NO2][i] *= scale_factor;
  }
  
  /* default CH4, N2O, F11, F12, and F22 are 'hardwired'; */
  /* these gases are assumed to be well-mixed throughout  */
  /* the whole atmosphere                                 */

  if (input.mixing_ratio[MX_CH4] < 0)
    output->mixing_ratio[MX_CH4] = 1.6; 
  else 
    output->mixing_ratio[MX_CH4] = input.mixing_ratio[MX_CH4];
  
  if (input.mixing_ratio[MX_N2O] < 0)
    output->mixing_ratio[MX_N2O] = 0.28; 
  else 
    output->mixing_ratio[MX_N2O] = input.mixing_ratio[MX_N2O];
  
  if (input.mixing_ratio[MX_F11] < 0)
    output->mixing_ratio[MX_F11] = 0.268E-3; 
  else 
    output->mixing_ratio[MX_F11] = input.mixing_ratio[MX_F11];
  
  if (input.mixing_ratio[MX_F12] < 0)
    output->mixing_ratio[MX_F12] = 0.503E-3; 
  else 
    output->mixing_ratio[MX_F12] = input.mixing_ratio[MX_F12];
  
  if (input.mixing_ratio[MX_F22] < 0)
    output->mixing_ratio[MX_F22] = 0.105E-3; 
  else 
    output->mixing_ratio[MX_F22] = input.mixing_ratio[MX_F22];
  

  /* Read Rayleigh cross section if specified */
  if (input.rayleigh == RAYLEIGH_FILE) {

    status = read_tau (input.filename[FN_RAYLEIGH], output->atm.nlyr,
		       output->atm.zd, &(output->atm.optprop.tau_rayleigh));
    if (status!=0) {
      fprintf (stderr, "Error, status %d returned by read_tau (\"%s\")\n", 
	       status, input.filename[FN_RAYLEIGH]);
      return status;
    }
  }			

  /* Read molecular absorption cross section if specified */
  if (strlen(input.filename[FN_MOLABS]) > 0) {
    status = read_molecular_absorption (input.filename[FN_MOLABS], output->atm.nlyr,
					output->atm.zd, input.quiet,
					&(output->atm.optprop.tau_molabs_user), 
					&tmp_dtau, &tmp_lambda, &tmp_nlambda,
					&monochromatic);

    if (status!=0) {
      fprintf (stderr, "Error, status %d returned by read_molecular_absorption (\"%s\")\n", 
	       status, input.filename[FN_MOLABS]);
      return status;
    }

    if (monochromatic) {
      output->atm.molabs = MOLABS_FILE_MONO;

      if (!input.quiet)
	fprintf (stderr, " ... read monochromatic molecular absorption from %s\n", 
		 input.filename[FN_MOLABS]);
    }
    else {
      output->atm.molabs = MOLABS_FILE_SPEC;
      
      if (!input.quiet)
	fprintf (stderr, " ... read spectral molecular absorption from %s\n", 
		 input.filename[FN_MOLABS]);
      
      /* in the first version, we force the user to define his own transmittance grid */
      if (strlen(input.filename[FN_WLTRANS])==0) {
	fprintf (stderr, "Error. If a wavelength-dependent molecular absorption cross section\n");
	fprintf (stderr, "is specified in molecular_tau_file, a transmittance_wl_file needs\n");
	fprintf (stderr, "to be specified as well to define the computational wavelength grid;\n");
	fprintf (stderr, "e.g. for a line-by-line calculation, it is reasonable to use identical\n");
	fprintf (stderr, "grids for both molecular_tau_file and transmittance_wl_file.\n");

	return -1;
      }	

      
      /* interpolate optical thickness to transmittance grid, linear */

      /* allocate memory for optical thickness on internal transmittance grid */
      output->atm.optprop.tau_molabs_user_t = calloc (output->atm.nlyr, sizeof(float));
      for (lc=0; lc<output->atm.nlyr; lc++)
	output->atm.optprop.tau_molabs_user_t[lc] = calloc (output->wl.nlambda_t, sizeof(float));
      
      for (lc=0; lc<output->atm.nlyr; lc++) {
	status = arb_wvn (tmp_nlambda, tmp_lambda, tmp_dtau[lc], 
			  output->wl.nlambda_t, output->wl.lambda_t, 
			  output->atm.optprop.tau_molabs_user_t[lc],
			  1, 0);
	
	if (status != 0) {
	  fprintf (stderr, "Error %d interpolating molecular absorption grid to internal\n", status);
	  fprintf (stderr, "wavelength grid\n");
	  return status;
	}
      }


      free(tmp_lambda);
      for (lc=0; lc<output->atm.nlyr; lc++)
	free (tmp_dtau[lc]);

      free (tmp_dtau);
    }
  }			


  /* Set common vertical resolution */
  output->atm.nlev_common = 0; 
  set_common_z_grid (output->atm.zd, output->atm.nlev, 
		     &output->atm.zd_common, &output->atm.nlev_common);


  if (input.verbose) {
    fprintf (stderr, "*** setup_atmosphere()\n");
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");
    fprintf (stderr, "   lc |  z[km] |  Pressure  | Temp.  |   Ozone     | Water vap. \n");
    fprintf (stderr, "      |        |   [mbar]   |  [K]   |   [cm-3]    |  [cm-3]    \n");
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");

    for (lc=0; lc<output->atm.nlev; lc++)
      fprintf (stderr, "%5d | %6.2f | %10.5f | %6.2f | %.5e | %.5e\n", 
	       lc, output->atm.zd[lc], 
	       output->atm.microphys.press[lc], output->atm.microphys.temper[lc],
	       output->atm.microphys.dens[MOL_O3][lc], output->atm.microphys.dens[MOL_H2O][lc]);

    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");
    fprintf (stderr, "%5s | %6.2f | %10.5f | %6.2f |  %7.3f DU | %6.3f kg/m2\n", 
	     "sum", 0.0/0.0,
	     0.0/0.0, 0.0/0.0,
	     output->ozone_column, output->precip_water);
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");
    
  }


  /* ??? user-defined equidistant altitude grid ??? */
  /* ??? may be added here; in this case, the   ??? */
  /* ??? the atmospheric grid will be           ??? */
  /* ??? interpolated to the user-define grid   ??? */
  if (input.atm.altitude_dz > 0) {
    nuser = (int) ((output->atm.zd[0] - output->atm.zd[output->atm.nlev-1]) / 
		   input.atm.altitude_dz + 0.5) + 1;
    
    zuser = (float *) calloc (nuser, sizeof(float));

    for (i=0; i<nuser; i++) 
      zuser[i] = output->atm.zd[output->atm.nlev-1] + input.atm.altitude_dz * (double) i;
    
    while (zuser[nuser-1] > output->atm.zd[0])
      nuser--;

    /* add user-defined grid to atmospheric grid */
    set_common_z_grid (zuser, nuser, &output->atm.zd_common, &output->atm.nlev_common);

    free (zuser);
  }


  /* calculate output levels relative to sea level   */
  /* (user-defined data are relative to the surface) */
  output->atm.nzout    = input.atm.nzout;
  output->atm.zout_sur = calloc (output->atm.nzout, sizeof (float));
  output->atm.zout_sea = calloc (output->atm.nzout, sizeof (float));

  for (iz=0; iz<output->atm.nzout; iz++) {
    output->atm.zout_sur[iz] = input.atm.zout_sur[iz];

    if (input.atm.zout_sur[iz] == -999)
      output->atm.zout_sea[iz] = -999;
    else
      output->atm.zout_sea[iz] = input.atm.zout_sur[iz] + input.atm.altitude;
  }

  return 0;
}


/**************************************************************/
/* Setup optical properties of trace gases.                   */
/**************************************************************/

int setup_gases (input_struct input, output_struct *output)
{
  static int first=1;
  int status=0;

  double bscar=0;
  double babso=0, babso_o3=0, babso_no2=0, babso_bro=0, babso_oclo=0,
    babso_hcho=0, babso_md=0;
  double babso_amf=0, babso_amftot=0, vertical_column=0, crs_amf=0;
  double bscartot=0;
  double babsotot=0, babsotot_o3=0, babsotot_no2=0, babsotot_bro=0, 
    babsotot_oclo=0, babsotot_hcho=0;
  double vcol_o3=0, vcol_no2=0, vcol_bro=0, vcol_oclo=0, vcol_hcho=0;
  double deltaz=0;
  int iq=0, is=0, isthis=0, iv=0, lc=0;
  int nq_rayleigh=0;
  int nlev = output->atm.nlev;
  int nlyr = output->atm.nlyr;

  int iq1=0, iq2=0, iq3=0, iq4=0;
  int nq=0, nq1=0, nq2=0, nq3=0, nq4=0;

  double factor=0;
  double weight[KATO_COMPONENTS];

  float *fu_rayleigh = calloc (nlyr, sizeof(float));


  /* number of quadrature points for Rayleigh scattering */
  switch (input.ck_scheme) {
  case CK_FU:
    nq_rayleigh=FU_MAXINT;
    break;
    
  case CK_KATO:
  case CK_KATO2:
  case CK_AVHRR_KRATZ:
  case CK_GENERIC:
  case CK_SBDART:
  case CK_NONE:
    nq_rayleigh=1;
    break;

  default:
    fprintf (stderr, "Error: unimplemented correlated-k scheme\n");
    return -1;
    break;
  }

  if (first!=0) {
    first = 0;

    /* allocate memory for profiles */
    output->atm.cdento   = (float *) calloc (nlyr, sizeof(float));
  
    output->atm.wght_t   = calloc (output->wl.nlambda_t, sizeof (double *));
    output->atm.nq_t     = calloc (output->wl.nlambda_t, sizeof (int));

    output->mc_photons_t = calloc (output->wl.nlambda_t, sizeof (double));


    /* copy ck weights */
    for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
      
      switch (input.ck_scheme) {
      case CK_FU:
	output->atm.nq_t[iv] = output->ck.wght[FU_WGHT][iv+1];
	break;
	
      case CK_KATO:
      case CK_KATO2:
	output->atm.nq_t[iv] = output->ck.wght[KATO_CO2][iv+1] *
                               output->ck.wght[KATO_O3] [iv+1] *
                               output->ck.wght[KATO_O2] [iv+1] *
                               output->ck.wght[KATO_H2O][iv+1];
	break;

      case CK_AVHRR_KRATZ:
	output->atm.nq_t[iv] = output->ck.wght[AVHRR_KRATZ_WGHT][iv+1];
	break;

      case CK_GENERIC:
	output->atm.nq_t[iv] = output->ck.wght[GENERIC_WGHT][iv+1];
	break;

      case CK_SBDART:
	output->atm.nq_t[iv] = SBDART_MAXINT;
	break;

      case CK_NONE:
	output->atm.nq_t[iv] = 1;
	break;
	
      default:
	fprintf (stderr, "Error: unimplemented correlated-k scheme\n");
	return -1;
	break;
      }

      output->atm.wght_t [iv] = calloc (output->atm.nq_t[iv], sizeof (double ));

      /* allocate memory for photon fractions */
      output->mc_photons_t [iv] = calloc (output->atm.nq_t[iv], sizeof (double ));


      /* copy photon fractions */
      switch(input.ck_scheme) {
      case CK_NONE:
	
	iq=0; /* no loop over quadrature points required */
	output->mc_photons_t [iv][iq] = output->wl.pfraction[iv];
	
	break;
	
      case CK_KATO:
      case CK_KATO2:
	nq1 = output->crs_ck.profile[KATO_CO2-1][iv].ngauss;
	nq2 = output->crs_ck.profile[KATO_O3-1 ][iv].ngauss;
	nq3 = output->crs_ck.profile[KATO_O2-1 ][iv].ngauss;
	nq4 = output->crs_ck.profile[KATO_H2O-1][iv].ngauss;
      
	iq=0;  /* reset quadrature counter */
	/* loop over quadrature points */
	for (iq1=0; iq1<nq1; iq1++)  /* CO2 */
	  for (iq2=0; iq2<nq2; iq2++)  /* O3 */
	    for (iq3=0; iq3<nq3; iq3++)  /* O2 */
	      for (iq4=0; iq4<nq4; iq4++)  /* H2O */
		switch (input.source) {
		case SRC_SOLAR: /* solar source */
		  output->mc_photons_t [iv][iq] = ((double *****) output->ck.x_solar)  [iv+1][iq1+1][iq2+1][iq3+1][iq4+1];
		  break;
		case SRC_THERMAL: /* thermal source */
		  output->mc_photons_t [iv][iq] = ((double *****) output->ck.x_thermal)[iv+1][iq1+1][iq2+1][iq3+1][iq4+1];
		  break;
		default:
		  fprintf (stderr, "Error, unknown source %d\n", input.source);
		  return -1;
		}
	break;
      
      case CK_FU:
      case CK_AVHRR_KRATZ:
      case CK_GENERIC:
      case CK_SBDART:
	
	nq = output->crs_ck.profile[0][iv].ngauss;
	
	for (iq=0; iq<nq; iq++)
	  switch (input.source) {
	  case SRC_SOLAR: /* solar source */
	    output->mc_photons_t [iv][iq] = ((double **) output->ck.x_solar)  [iv+1][iq+1];
	    break;
	  case SRC_THERMAL: /* thermal source */
	    output->mc_photons_t [iv][iq] = ((double **) output->ck.x_thermal)[iv+1][iq+1];
	    break;
	  default:
	    fprintf (stderr, "Error, unknown source %d\n", input.source);
	    return -1;
	  }
	
	
	break;
	
      default:
	fprintf (stderr, "Error: unimplemented correlated-k scheme\n");
	return -1;
	break;
      }
    }

    /* Rayleigh scattering optical thickness */
    if ((status = ASCII_calloc_double_3D_arylen (&output->atm.optprop.tau_rayleigh_t, 
						 nlyr, output->wl.nlambda_t, 
						 output->atm.nq_t)) != 0)
      return status;

    /* Molecular absorption optical thickness */
    if ((status = ASCII_calloc_double_3D_arylen (&output->atm.optprop.tau_molabs_t, 
						 nlyr, output->wl.nlambda_t, 
						 output->atm.nq_t)) != 0)
      return status;

    /* Molecular absorption optical thickness for AMF calculations */
    if ((status = ASCII_calloc_double_3D_arylen (&output->atm.optprop.tau_molabs_md_t, 
						 nlyr, output->wl.nlambda_t, 
						 output->atm.nq_t)) != 0)
      return status;
    
    /* Calculate column density of air (total density of the atmosphere). */
    /* Required to calculate Rayleigh optical depth.                      */

    if (input.rayleigh==RAYLEIGH_CALC) {
      for (lc=0; lc<nlyr; lc++) {
	
	/* The factor 1e5 converts z from km to cm.*/
	deltaz = (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1e5;
      
	/* average air density over a layer assuming  dens = exp (-c * z) */
	if (output->atm.microphys.dens[MOL_AIR][lc] > 0 && output->atm.microphys.dens[MOL_AIR][lc+1] > 0 &&
	    output->atm.microphys.dens[MOL_AIR][lc] != output->atm.microphys.dens[MOL_AIR][lc+1]) {
	  if (fabs(1.0 - output->atm.microphys.dens[MOL_AIR][lc+1] / output->atm.microphys.dens[MOL_AIR][lc]) > EPSILON ) 
	    output->atm.cdento[lc] = 1.0 / (log(output->atm.microphys.dens[MOL_AIR][lc+1]/output->atm.microphys.dens[MOL_AIR][lc])) * 
	      (output->atm.microphys.dens[MOL_AIR][lc+1] - output->atm.microphys.dens[MOL_AIR][lc]) * deltaz ;
	  else
	    output->atm.cdento[lc] = 0.5*fabs((output->atm.microphys.dens[MOL_AIR][lc]+
					       output->atm.microphys.dens[MOL_AIR][lc+1]) * deltaz);
	}
	else
	  output->atm.cdento[lc] = 0.5*fabs((output->atm.microphys.dens[MOL_AIR][lc]+
					     output->atm.microphys.dens[MOL_AIR][lc+1]) * deltaz);

      }
    }
  } /* End if first block */

  if ( output->atm.microphys.denstab_id ) {
    /* Find matrix profile corresponding to the sza this rte calculation is done for */
    /* This is airmass factor stuff */
    iv=0;  /* No solar zenith variations for a single amf calculation */
    for (is=0; is<output->atm.microphys.nsza_denstab; is++)
      if (fabs (output->atm.microphys.sza_denstab[is]-output->atm.sza_t[iv]) < 1.e-04)
	isthis = is;

    for (lc=0; lc<nlyr; lc++)
      output->atm.microphys.dens[output->atm.microphys.denstab_id][lc] = 
	output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc];
  }



  /* loop over wavelength */
  for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
    /***********************/
    /* Rayleigh scattering */
    /***********************/


    for (iq=0; iq<output->atm.nq_t[iv]; iq++) {

    /* FU special: The Rayleigh optical depth depends on the subband */

      for (iq=0; iq<output->atm.nq_t[iv]; iq++) {

	if (input.ck_scheme == CK_FU) {
#if HAVE_FULIOU
	  status = ckdfuray (iv+1, iq+1, cos(output->atm.sza_t[iv]*PI/180.0), nlev,
			     fu_rayleigh);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error, status %d returned by ckdfuray()\n", status);
	    return status;
	  }    
	}
#else
	fprintf (stderr, "Error, Fu and Liou not supported!\n");
	return -1;
#endif  
	
	
	/* loop over layers */
	for (lc=0; lc<nlyr; lc++) {
	  
	  /* The factor 1.e+5 converts z from km to cm */
	  deltaz= (output->atm.zd[lc] - output->atm.zd[lc+1])*1e5;
	  
	  switch (input.rayleigh) {  /* if Rayleigh scattering is not switched off */
	  case RAYLEIGH_CALC:
	    if (input.ck_scheme != CK_FU)
	      bscar = output->atm.cdento[lc] * output->crs.crs[0][MOL_AIR][iv];
	    else 
	      bscar = fu_rayleigh[lc];
	    break;
	    
	  case RAYLEIGH_FILE:
	    bscar = output->atm.optprop.tau_rayleigh[lc];
	    break;
	  
	  case RAYLEIGH_NONE:
	    bscar = 0;
	    break;
	    
	  default:
	    fprintf (stderr, "Error, unknown Rayleigh option %d\n", input.rayleigh);
	    return -1;
	  }
	  
	  output->atm.optprop.tau_rayleigh_t[lc][iv][iq] = bscar;
	}
      }
    }    



    switch(input.ck_scheme) {
    case CK_NONE:

      iq=0; /* no loop over quadrature points required */
      
      bscartot=0;
      babsotot=0; 
      babsotot_o3=0; babsotot_no2=0; babsotot_bro=0; babsotot_oclo=0; babsotot_hcho=0; babso_amftot=0;
      vcol_o3=0; vcol_no2=0; vcol_bro=0; vcol_oclo=0; vcol_hcho=0; 
      
      vertical_column=0;
      
      /* loop over layers */
      for (lc=0; lc<nlyr; lc++) {
	
	/* The factor 1.e+5 converts z from km to cm.*/
	deltaz= (output->atm.zd[lc] - output->atm.zd[lc+1])*1e5;
	
	babso=0;
	babso_o3=0;
	babso_no2=0;
	babso_bro=0;
	babso_oclo=0;
	babso_hcho=0;
	babso_amf=0;
	babso_md=0;
	
	switch (output->atm.molabs) {
	case MOLABS_CALC:
	  /* O3 absorption */
	  vcol_o3  += 0.5*deltaz * (output->atm.microphys.dens[MOL_O3][lc] + output->atm.microphys.dens[MOL_O3][lc+1]);
	  babso_o3 += 0.5*deltaz * (output->atm.microphys.dens[MOL_O3][lc]*output->crs.crs[lc][MOL_O3][iv]
				    + output->atm.microphys.dens[MOL_O3][lc+1]*output->crs.crs[lc+1][MOL_O3][iv]);
	  babso += babso_o3;
	  if ( output->atm.microphys.denstab_id != MOL_O3 ) babso_md += babso_o3;
	  
	  /* NO2 absorption */
	  vcol_no2  += 0.5*deltaz * (output->atm.microphys.dens[MOL_NO2][lc] + output->atm.microphys.dens[MOL_NO2][lc+1]);
	  babso_no2 += 0.5*deltaz * (output->atm.microphys.dens[MOL_NO2][lc]*output->crs.crs[lc][MOL_NO2][iv]
				       + output->atm.microphys.dens[MOL_NO2][lc+1]*output->crs.crs[lc+1][MOL_NO2][iv]);
	  babso += babso_no2;
	  if ( output->atm.microphys.denstab_id != MOL_NO2 ) babso_md += babso_no2;
	  
	  /* BRO absorption */
	  vcol_bro  += 0.5*deltaz * (output->atm.microphys.dens[MOL_BRO][lc] + output->atm.microphys.dens[MOL_BRO][lc+1]);
	  babso_bro += 0.5*deltaz * (output->atm.microphys.dens[MOL_BRO][lc]*output->crs.crs[lc][MOL_BRO][iv]
				     + output->atm.microphys.dens[MOL_BRO][lc+1]*output->crs.crs[lc+1][MOL_BRO][iv]);
	  babso += babso_bro;
	  if ( output->atm.microphys.denstab_id != MOL_BRO ) babso_md += babso_bro;
	  
	  /* OCLO absorption */
	  vcol_oclo  += 0.5*deltaz * (output->atm.microphys.dens[MOL_OCLO][lc] + output->atm.microphys.dens[MOL_OCLO][lc+1]);
	  babso_oclo += 0.5*deltaz * (output->atm.microphys.dens[MOL_OCLO][lc]*output->crs.crs[lc][MOL_OCLO][iv]
				      + output->atm.microphys.dens[MOL_OCLO][lc+1]*output->crs.crs[lc+1][MOL_OCLO][iv]);
	  babso += babso_oclo;
	  if ( output->atm.microphys.denstab_id != MOL_OCLO ) babso_md += babso_oclo;
	  
	  /* HCHO absorption */
	  vcol_hcho  += 0.5*deltaz * (output->atm.microphys.dens[MOL_HCHO][lc] + output->atm.microphys.dens[MOL_HCHO][lc+1]);
	  babso_hcho += 0.5*deltaz * (output->atm.microphys.dens[MOL_HCHO][lc]*output->crs.crs[lc][MOL_HCHO][iv]
				      + output->atm.microphys.dens[MOL_HCHO][lc+1]*output->crs.crs[lc+1][MOL_HCHO][iv]);
	  babso += babso_hcho;
	  if ( output->atm.microphys.denstab_id != MOL_HCHO )  babso_md += babso_hcho;
	  
	  if ( output->atm.microphys.denstab_id ) {
	    /* Use column for matrix profile corresponding to the sza this rte calculation is done for */
	    /* This is airmass factor stuff */
	    babso_amf += 0.5*deltaz * (output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc]*
				       output->crs.crs[lc][output->atm.microphys.denstab_id][iv]
				       + output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc+1]*
				       output->crs.crs[lc+1][output->atm.microphys.denstab_id][iv]);
	    vertical_column += 0.5*deltaz * (output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc]
					     + output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc+1]);
	    crs_amf = output->crs.crs[lc+1][output->atm.microphys.denstab_id][iv];
	  }
	  output->atm.optprop.tau_molabs_t[lc][iv][iq]    = babso;
	  output->atm.optprop.tau_molabs_md_t[lc][iv][iq] = babso_md;
	  break;
	  
	case MOLABS_FILE_MONO:
	  babso = output->atm.optprop.tau_molabs_user[lc];
	  output->atm.optprop.tau_molabs_t[lc][iv][iq]   = babso;
	  break;
	  
	case MOLABS_FILE_SPEC:
	  babso = output->atm.optprop.tau_molabs_user_t[lc][iv];
	  output->atm.optprop.tau_molabs_t[lc][iv][iq]   = babso;
	  break;
	  
	case MOLABS_NONE:
	  babso = 0;
	  output->atm.optprop.tau_molabs_t[lc][iv][iq]   = babso;
	  break;
	  
	default:
	  fprintf (stderr, "Error, unknown molecular absorption option %d\n", output->atm.molabs);
	  return -1;
	}
	
	output->atm.wght_t[iv][iq] = 1;  /* default weight */
      

	if (input.verbose) {
	  bscar=output->atm.optprop.tau_rayleigh_t[lc][iv][iq];
	  bscartot+=bscar;
	  
	  babsotot+=babso;
	  babsotot_o3+=babso_o3;
	  babsotot_no2+=babso_no2;
	  babsotot_bro+=babso_bro;
	  babsotot_oclo+=babso_oclo;
	  babsotot_hcho+=babso_hcho;
	  babso_amftot+=babso_amf;
	  
	  if ( lc ==0 ) {
	    fprintf (stderr, "\n*** wavelength: iv = %d, %f nm\n", iv, output->wl.lambda_t[iv]);
	    fprintf (stderr, "*** Rayleigh scattering cross section: %e\n", output->crs.crs[0][MOL_AIR][iv]);
	    fprintf (stderr, "*** setup_gases()\n");
	    fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------\n");
	    fprintf (stderr, "   lc |    z[km] |  Rayleigh   |   Molecular \n");
	    fprintf (stderr, "      |          |    dtau     |   absorption        o3            no2           bro           oclo          hcho\n");
	    fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------\n");
	  }		  
	  fprintf (stderr, "%5d | %8.2f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f\n", 
		   lc, output->atm.zd[lc+1],
		   bscar,
		   babso, babso_o3, babso_no2, babso_bro, babso_oclo, babso_hcho
		   );
	}
      } /* End layer loop */
      if (input.verbose) {
	fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------\n");
	fprintf (stderr, "%5s | %8.2f | %11.6f | %11.6f | %12.6e| %12.6e| %12.6e| %12.6e| %12.6e\n", 
		 "sum", 0.0/0.0,
		 bscartot,
		 babsotot, babsotot_o3, babsotot_no2, babsotot_bro, babsotot_oclo, babsotot_hcho);
	fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------\n");
	fprintf (stderr, "Crs at bottom level. For AMF calculations    | %12.6e| %12.6e| %12.6e| %12.6e| %12.6e\n", 
		 output->crs.crs[nlyr-1][MOL_O3][iv], output->crs.crs[nlyr-1][MOL_NO2][iv], 
		 output->crs.crs[nlyr-1][MOL_BRO][iv], output->crs.crs[nlyr-1][MOL_OCLO][iv],
		 output->crs.crs[nlyr-1][MOL_HCHO][iv]);
	fprintf (stderr, "The vertical column. For AMF calculations    | %12.6e| %12.6e| %12.6e| %12.6e| %12.6e\n", 
		 vcol_o3, vcol_no2, vcol_bro, vcol_oclo, vcol_hcho);
	fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------\n");
	if ( output->atm.microphys.denstab_id ) {
	  fprintf(stderr,"Density matrix absorption column for airmass calculations: %16.10e %16.10e %16.10e\n",
		  babso_amftot, vertical_column, crs_amf);
	fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------\n");
	}
      }
      
      break;
	
    case CK_KATO:
    case CK_KATO2:
      nq1 = output->crs_ck.profile[KATO_CO2-1][iv].ngauss;
      nq2 = output->crs_ck.profile[KATO_O3-1 ][iv].ngauss;
      nq3 = output->crs_ck.profile[KATO_O2-1 ][iv].ngauss;
      nq4 = output->crs_ck.profile[KATO_H2O-1][iv].ngauss;
      
      if (input.verbose) {
	fprintf (stderr, "*** subbands:\n");
	fprintf (stderr, "***  CO2: %2d\n", nq1);
	fprintf (stderr, "***  O3:  %2d\n", nq2);
	fprintf (stderr, "***  O2:  %2d\n", nq3);
	fprintf (stderr, "***  H2O: %2d\n", nq4);
      }
      
      for (lc=0; lc<output->atm.nlyr; lc++) {
	
	/* The factor 1e5 converts z from km to cm.*/
	deltaz = (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1e5;
      
	iq=0;  /* reset quadrature counter */
	/* loop over components (4D correlated-k) */
	for (iq1=0; iq1<nq1; iq1++) {  /* CO2 */
	  
	  /* set CO2 cross section and weights */
	  output->crs.crs[lc][MOL_CO2][iv] = 
	    output->crs_ck.profile[KATO_CO2-1][iv].crs[lc][iq1];
	  
	  
	  weight[KATO_CO2-1] = output->crs_ck.profile[KATO_CO2-1][iv].weight[iq1];
	  
	  for (iq2=0; iq2<nq2; iq2++) { /* O3 */
	    
	    /* set O3 cross section and weights */
	    output->crs.crs[lc][MOL_O3][iv]  = 
	      output->crs_ck.profile[KATO_O3-1][iv].crs[lc][iq2];
	    
	    weight[KATO_O3-1] = output->crs_ck.profile[KATO_O3-1][iv].weight[iq2];
	    
	    for (iq3=0; iq3<nq3; iq3++) { /* O2 */
	      
	      /* set O2 cross section and weights */
	      output->crs.crs[lc][MOL_O2][iv]  =
		output->crs_ck.profile[KATO_O2-1][iv].crs[lc][iq3];
	      
	      weight[KATO_O2-1] = output->crs_ck.profile[MOL_O2][iv].weight[iq3];
	      
	      for (iq4=0; iq4<nq4; iq4++) { /* H2O */
		
		/* set H2O cross section and weights */
		output->crs.crs[lc][MOL_H2O][iv]  = 
		  output->crs_ck.profile[KATO_H2O-1][iv].crs[lc][iq4];
		
		weight[KATO_H2O-1] = output->crs_ck.profile[MOL_H2O][iv].weight[iq4];
		
		/* quadrature weight */
		factor = weight[KATO_CO2-1]*weight[KATO_O3-1]*weight[KATO_O2-1]*weight[KATO_H2O-1];
		

		/* sum absorption */
		babso = 0;
		
		/* CO2 absorption */
		babso += deltaz * output->crs.crs[lc][MOL_CO2][iv] * 
		  0.5 * (output->atm.microphys.dens[MOL_CO2][lc] + output->atm.microphys.dens[MOL_CO2][lc+1]);
		
		/* O3 absorption */
		babso += deltaz * output->crs.crs[lc][MOL_O3][iv]  *
		  0.5 * (output->atm.microphys.dens[MOL_O3][lc]  + output->atm.microphys.dens[MOL_O3][lc+1]);
		
		/* O2 absorption */
		babso += deltaz * output->crs.crs[lc][MOL_O2][iv]  *
		  0.5 * (output->atm.microphys.dens[MOL_O2][lc]  + output->atm.microphys.dens[MOL_O2][lc+1]);
		
		/* H2O absorption */
		babso += deltaz * output->crs.crs[lc][MOL_H2O][iv] *
		  0.5 * (output->atm.microphys.dens[MOL_H2O][lc] + output->atm.microphys.dens[MOL_H2O][lc+1]);
		
		if (input.verbose)
		  fprintf (stderr, "*** ck-weight: %.6f\n", factor);
		
		/* copy absorption cross section and weight */
		/* to final destination                     */
		output->atm.wght_t[iv][iq] = factor;
		
		output->atm.optprop.tau_molabs_t[lc][iv][iq]   = babso;
		iq++;
		
	      }  /* end of correlated-k (iq4) loop */
	    }  /* end of correlated-k (iq3) loop */
	  }  /* end of correlated-k (iq2) loop */
	}  /* end of correlated-k (iq1) loop */
      }
  
      break;
	
    case CK_FU:
    case CK_AVHRR_KRATZ:
    case CK_GENERIC:
    case CK_SBDART:

      nq = output->crs_ck.profile[0][iv].ngauss;

      /* loop over quadrature points */
      
      for (iq=0; iq<nq; iq++) {
	
	/* set absorption coefficient */
	for (lc=0; lc<output->atm.nlyr; lc++) {

	  /* The factor 1e5 converts z from km to cm.*/
	  deltaz = (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1e5;

	  output->atm.optprop.tau_molabs_t[lc][iv][iq] =
	    deltaz/1e5 * output->crs_ck.profile[0][iv].crs[lc][iq];
	}
	  
	/* quadrature weight */
	factor = output->crs_ck.profile[0][iv].weight[iq];
	output->atm.wght_t[iv][iq] = factor;
      }
      
      break;
      
    default:
      fprintf (stderr, "Error: unimplemented correlated-k scheme\n");
      return -1;
      break;
    }
  } /* End wavelength loop */

  free(fu_rayleigh);


  return 0;  /* ok */
}



/**************************************************************/
/* Setup temperature at output levels.                        */
/**************************************************************/

int setup_temperature (input_struct input, output_struct *output)
{
  int status=0, lev=0, iz=0;
  float *tmp_temper=NULL, *tmp_zd=NULL, *zout=NULL;

  /* Linearly interpolate output temperature to zout */
  /* sort profile in ascending order, assuming that */
  /* the original profile was descending            */
  
  tmp_zd     = (float *) calloc (output->atm.nlev, sizeof(float));
  tmp_temper = (float *) calloc (output->atm.nlev, sizeof(float));
  
  /* sort temperature profile in ascending order */
  for (lev=0; lev<output->atm.nlev; lev++) {
    tmp_zd[lev]     = output->atm.zd     [output->atm.nlev-1-lev];
    tmp_temper[lev] = output->atm.microphys.temper [output->atm.nlev-1-lev];
  }

  output->atm.temper_out  = (float *) calloc (input.atm.nzout, sizeof(float));
 
  /* Now we need to define a temporary zout grid;        */
  /* for rte_solver montecarlo it is possible to         */
  /* define an output altitude of -999 which stands      */
  /* for the surface altitude which may be a function    */
  /* of (x,y). This value is replaced by 0 which means   */
  /* that the surface temperature at the surface output  */
  /* level is set to temperature(z=0) which is not quite */
  /* correct if a 2D altitude grid has been defined.     */
  /* However, the only place where this information is   */
  /* currently used is the flexstor header which is      */
  /* required for photolysis calculations.               */

  zout = calloc (output->atm.nzout, sizeof(float));
  for (iz=0; iz<output->atm.nzout; iz++) {
    if (output->atm.zout_sea[iz] != -999)
      zout[iz] = output->atm.zout_sea[iz];
    else
      zout[iz] = 0;
  }
  
  /* interpolate temperature to zout grid */
  status = arb_wvn (output->atm.nlev, tmp_zd, tmp_temper, 
		    input.atm.nzout, zout, output->atm.temper_out, 1, 0);

  if (status != 0) {
    fprintf (stderr, "Error interpolating temperature to output grid\n");
    return status;
  }


  free(tmp_temper); 
  free(tmp_zd);
  free(zout);


  /* Finally, copy the temperature of the lowest level to */
  /* output->temperature if the latter is not already set */

  if (output->surface_temperature < 0) {
    output->surface_temperature = output->atm.microphys.temper[output->atm.nlev-1];
    
    /* fprintf (stderr, " ... setting surface temperature to %f K\n", output->surface_temperature); */
  }

  return 0;
}




/**************************************************************/
/* Read atmosphere file and write to atm_out_struct           */
/**************************************************************/

static int read_atmosphere (char *filename,
			    atm_out_struct *out, int cdf, 
			    int quiet, int verbose) 
{
  int i=0, rows=0, min_columns=0, max_columns=0;
  int lev=0, status = 0;
  float **atm_data=NULL;
  double boltzmann = 1.38065e-23;

  long dimlen=0;
  
  int nlev=0, nlyr=0;

  float *zd=NULL, *press=NULL, *temper=NULL;
  float **dens=NULL;



#if HAVE_LIBNETCDF
  int ncid=0;
  int idd_nlev=0;
  int id_z=0, id_press=0, id_temper=0;
  int id_air=0, id_o3=0, id_o2=0, id_h2o=0, id_co2=0;
  
  long start[1] = {0};
  long count[1] = {0};
#endif

  if (cdf==0) {   /* classic ASCII atmospheric model */
    status = ASCII_file2float (filename, 
			       &rows, &max_columns, &min_columns, 
			       &atm_data);

    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    if (max_columns!=min_columns) {
      fprintf (stderr, "Warning: Inconsistent number of columns in %s\n", filename);
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
      return -1;
    }  
    
    if (min_columns<2) {
      fprintf (stderr, "Error, too few columns in %s! Need at least %d.\n", filename, 2);
      return -1;
    }
        
    nlev = rows;
    nlyr = rows-1;
  
    /* allocate memory for density array */
    dens = calloc(MOL_NN, sizeof(float *));
    if (dens == NULL) {
      fprintf (stderr, "Error allocating memory for dens\n");
      return -1;
    }
    
    press  = ASCII_column_float (atm_data, nlev, ATMFILE_PRESS );
    temper = ASCII_column_float (atm_data, nlev, ATMFILE_TEMPER );
    zd     = ASCII_column_float (atm_data, nlev, ATMFILE_ZD );

    if (min_columns>3) { /* air and o3   */
      dens[MOL_AIR] = ASCII_column_float(atm_data, nlev, ATMFILE_AIR );
      dens[MOL_O3]  = ASCII_column_float(atm_data, nlev, ATMFILE_O3 );
    }
    else {
      dens[MOL_AIR]  = (float *) calloc (nlev, sizeof(float));
      dens[MOL_O3]   = (float *) calloc (nlev, sizeof(float));

      /* Set air density */
      for (lev=0; lev<nlev; lev++)
	dens[MOL_AIR][lev] = press[lev]*1E-04/(boltzmann*temper[lev]);
    }
    
    /* the following columns are optional */
    
    if (min_columns>5)  /* oxygen           */
      dens[MOL_O2]  = ASCII_column_float(atm_data, nlev, ATMFILE_O2 );
    else 
      dens[MOL_O2]  = (float *) calloc (nlev, sizeof(float));
    
    
    if (min_columns>6)  /* water vapour     */
      dens[MOL_H2O] = ASCII_column_float(atm_data, nlev, ATMFILE_H2O);
    else 
      dens[MOL_H2O]  = (float *) calloc (nlev, sizeof(float));
    
    
    if (min_columns>7)  /* carbon dioxide   */
      dens[MOL_CO2] = ASCII_column_float(atm_data, nlev, ATMFILE_CO2);
    else 
      dens[MOL_CO2]  = (float *) calloc (nlev, sizeof(float));
    
    
    if (min_columns>8)  /* nitrogen dioxide */
      dens[MOL_NO2] = ASCII_column_float(atm_data, nlev, ATMFILE_NO2);
    else 
      dens[MOL_NO2]  = (float *) calloc (nlev, sizeof(float));

    /* Allocate memory for species not in atmosphere file */
    dens[MOL_BRO]  = (float *) calloc (nlev, sizeof(float));
    dens[MOL_OCLO] = (float *) calloc (nlev, sizeof(float));
    dens[MOL_HCHO] = (float *) calloc (nlev, sizeof(float));

    ASCII_free_float(atm_data, rows);
  }
  else {

#if HAVE_LIBNETCDF

    /* open netcdf file */
    ncid = ncopen (filename, NC_NOWRITE);
    if (ncid==-1) {
      fprintf (stderr, "Error %d opening netCDF file %s\n", -1, filename);
      return -1;
    }

    /* get dimension id for "nlev" */
    idd_nlev = ncdimid (ncid, "nlev");
  
    /* get dimension length for "nwvn" */
    ncdiminq (ncid, idd_nlev, NULL, &dimlen);
    nlev = dimlen;
    nlyr = dimlen-1;


    start[0] = 0;
    count[0] = nlev;

    /* allocate memory for profiles */
    zd     = calloc (nlev, sizeof (float));
    press  = calloc (nlev, sizeof (float));
    temper = calloc (nlev, sizeof (float));


    /* get variable id for "z" */
    id_z = ncvarid (ncid, "z");
    
    /* read altitude "z" */
    status = ncvarget (ncid, id_z, start, count, (void *) zd);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading z from %s\n", status, filename);
      return status;
    }


    /* get variable id for "pressure" */
    id_press = ncvarid (ncid, "pressure");
    
    /* read "pressure" */
    status = ncvarget (ncid, id_press, start, count, (void *) press);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading pressure from %s\n", status, filename);
      return status;
    }


    /* get variable id for "temperature" */
    id_temper = ncvarid (ncid, "temperature");
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading temperature from %s\n", status, filename);
      return status;
    }
    
    /* read "temperature" */
    status = ncvarget (ncid, id_temper, start, count, (void *) temper);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading temperature from %s\n", status, filename);
      return status;
    }


    /* allocate memory for density array */
    dens = calloc(MOL_NN, sizeof(float *));
    if (dens == NULL) {
      fprintf (stderr, "Error allocating memory for dens\n");
      return -1;
    }

    for (i=0; i<MOL_NN; i++) {
      dens[i] = calloc(nlev, sizeof(float));
      if (dens[i] == NULL) {
	fprintf (stderr, "Error allocating memory for dens[%d]\n", i);
	return -1;
      }
    }


    /* get variable id for "density" */
    id_air = ncvarid (ncid, "density");
    
    /* read "density" */
    status = ncvarget (ncid, id_air, start, count, (void *) dens[MOL_AIR]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading density from %s\n", status, filename);
      return status;
    }


    /* get variable id for "o3" */
    id_o3 = ncvarid (ncid, "o3");
    
    /* read "o3" */
    status = ncvarget (ncid, id_o3, start, count, (void *) dens[MOL_O3]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading o3 from %s\n", status, filename);
      return status;
    }


    /* get variable id for "o2" */
    id_o2 = ncvarid (ncid, "o2");
    
    /* read "o2" */
    status = ncvarget (ncid, id_o2, start, count, (void *) dens[MOL_O2]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading o2 from %s\n", status, filename);
      return status;
    }


    /* get variable id for "h2o" */
    id_h2o = ncvarid (ncid, "h2o");
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading h2o from %s\n", status, filename);
      return status;
    }
    
    /* read "h2o" */
    status = ncvarget (ncid, id_h2o, start, count, (void *) dens[MOL_H2O]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading h2o from %s\n", status, filename);
      return status;
    }


    /* get variable id for "co2" */
    id_co2 = ncvarid (ncid, "co2");
    
    /* read "co2" */
    status = ncvarget (ncid, id_co2, start, count, (void *) dens[MOL_CO2]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading co2 from %s\n", status, filename);
      return status;
    }


    /* we don't have no2 in the netcdf file, hence we keep it 0;        */
    /* also we don't read n2o, co, and ch4 because we never used        */
    /* those before (should change that by introducing all relevant     */
    /* gases into dens; should also get rid of ndens  */
    /* and replace that by MOL_NN)                                      */

    ncclose(ncid);
    
#else
  fprintf (stderr, " ***********************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot            *\n");
  fprintf (stderr, " * use the generic correlated_k option. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ***********************************************************************\n");
  return -1;
#endif

  }

  /**************************/
  /* interpolate atmosphere */
  /**************************/

  /* pressure, logarithmic */
  out->microphys.press  = calloc (out->nlev, sizeof (float));
  status += arb_wvn (nlev, zd, press, out->nlev, out->zd, out->microphys.press, 2, 1);

  /* temperature, linear */
  out->microphys.temper = calloc (out->nlev, sizeof (float));
  status += arb_wvn (nlev, zd, temper, out->nlev, out->zd, out->microphys.temper, 1, 1);

  /* densities, logarithmic */
  out->microphys.dens = calloc (MOL_NN, sizeof(float *));
  for (i=0; i<MOL_NN; i++) {
    (out->microphys.dens)[i] = calloc (out->nlev, sizeof(float));
    status += arb_wvn (nlev, zd, dens[i], out->nlev, out->zd, (out->microphys.dens)[i], 2, 1);
  }

  if (status!=0) {
    fprintf (stderr, "Error interpolating atmospheric profiles\n");
    return status;
  }


  /* free memory */
  free (zd); free (press); free (temper);
  for (i=0; i<MOL_NN; i++)
    free (dens[i]);
  free (dens);
  
  return 0;
}

/**************************************************************/
/* Read atmosphere file and write to out_struct           */
/**************************************************************/

static int read_density (char *filename,
			 atm_out_struct *out, int dens_id, 
			 int quiet, int verbose) 
{
  int rows=0, min_columns=0, max_columns=0, nlev=0;
  int status=0, i=0, j=0;
  float **atm_data=NULL;
  float *zd=NULL, *dens=NULL;
  float ***denstab=NULL, **denstab_amf=NULL;
  static int first_dens = 1;
  
  status = ASCII_file2float (filename, 
			     &rows, &max_columns, &min_columns, 
			     &atm_data);

  if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
  }
    

  if (max_columns > 2 ) {   /* density matrix for airmass calculations */
    if (max_columns!=min_columns) {
      fprintf (stderr, "Error: Inconsistent number of columns in %s\n", filename);
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
      return -1;
    }


    /* first row/column contains sza/altitude */
    out->microphys.nsza_denstab = max_columns-1;
    nlev = rows-1;

    zd  = (float *) calloc (nlev, sizeof (float));

    if (out->microphys.denstab_id > 0 ) {
      fprintf (stderr, "Error: dens_file with matrix instead of profile\n");
      fprintf (stderr, "       already specified. The matrix may be used\n");
      fprintf (stderr, "       only for one specie. Profiles for as many as you like\n");
      return -1;
    }

    if (first_dens) {
      out->microphys.denstab_id  = dens_id;
      out->microphys.sza_denstab = (float *) calloc (out->microphys.nsza_denstab, sizeof (float));

      status  = ASCII_calloc_float_3D (&out->microphys.denstab, MOL_NN, out->microphys.nsza_denstab, out->nlev);
      if (status!=0) { 
	fprintf (stderr, "Error %d allocating memory for out->denstab\n", status); 
	return status; 
      } 
      status  = ASCII_calloc_float_3D (&denstab, MOL_NN, out->microphys.nsza_denstab, nlev);
      if (status!=0) { 
	fprintf (stderr, "Error %d allocating memory for denstab\n", status); 
	return status; 
      } 

      status  = ASCII_calloc_float(&out->microphys.denstab_amf, 
				   out->microphys.nsza_denstab, out->nlev);
      if (status!=0) { 
	fprintf (stderr, "Error %d allocating memory for out->denstab_amf\n", status); 
	return status; 
      } 
      status  = ASCII_calloc_float(&denstab_amf, 
				   out->microphys.nsza_denstab, nlev);
      if (status!=0) { 
	fprintf (stderr, "Error %d allocating memory for denstab_amf\n", status); 
	return status; 
      } 

    }

    /* copy first row to solar zenith angle array  */
    for (i=1; i<=out->microphys.nsza_denstab; i++)  
      out->microphys.sza_denstab[i-1] = atm_data[0][i];
    
    /* copy first column to altitude array zd */
    for (j=1; j<=nlev; j++)  
      zd[j-1] = atm_data[j][0];

    /* copy the remainder to 2dim double array denstab */
    for (i=1; i<=out->microphys.nsza_denstab; i++) 
      for (j=1; j<=nlev; j++) {
	denstab     [dens_id][i-1][j-1] = atm_data[j][i];
	denstab_amf          [i-1][j-1] = atm_data[j][i];
      }
    
    /* interpolate to atmospheric grid, logarithmic */
    for (i=0; i<out->microphys.nsza_denstab; i++) {
      status = arb_wvn (nlev, zd, denstab[dens_id][i], out->nlev, 
			out->zd, out->microphys.denstab[dens_id][i], 2, 1);
      if (status!=0) {
	fprintf (stderr, "Error %d interpolating denstab to common atmospheric grid\n", status);
	return status;
      }
      status = arb_wvn (nlev, zd, denstab_amf[i], out->nlev, 
			out->zd, out->microphys.denstab_amf[i], 2, 1);
      if (status!=0) {
	fprintf (stderr, "Error %d interpolating denstab_amf to common atmospheric grid\n", status);
	return status;
      }
    }

    status = ASCII_free_float_3D(denstab, MOL_NN, out->microphys.nsza_denstab);
    status = ASCII_free_float(denstab_amf, out->microphys.nsza_denstab);
    first_dens = 0;

  }
  else if ( max_columns==2) {   /* 2-columns: 1 altitude, 2 density */

    if (max_columns!=min_columns ) {
      fprintf (stderr, "Error: Inconsistent number of columns in %s\n", filename);
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
      fprintf (stderr, "Should have been %d. First altitude, second density.\n", 2);
      return -1;
    }

    nlev  = rows;
    zd   = ASCII_column_float (atm_data, nlev, 0 );
    dens = ASCII_column_float (atm_data, nlev, 1 );

    /* interpolate to atmospheric grid, logarithmic */
    out->microphys.dens[dens_id] = calloc (out->nlev, sizeof (float));
    status = arb_wvn (nlev, zd, dens, out->nlev, out->zd, out->microphys.dens[dens_id], 2, 1);
    
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating density to common atmospheric grid\n", status);
      return status;
    }
  }
  
  
  ASCII_free_float(atm_data, rows);
  free(zd); free (dens);

  return status;
}


/***********************************************************************************/
/*  Read a two-column file containing altitude [km] and optical thickness          */
/*  of each layer. Check if the levels equal those defined                         */
/*  in zd[]. Memory for dtau[] is allocated automatically.                         */ 
/***********************************************************************************/

static int read_tau (char* filename, int nlyr,
		     float* zd, float** dtau) 
{
  int rows=0;
  int lev=0, lc=0;
  int status = 0;
  double *rayleigh_zd=NULL, *rayleigh_dtau=NULL;
  

  /* read altitude and optical depth from file */
  status = read_2c_file (filename, 
			 &rayleigh_zd, &rayleigh_dtau, &rows);

    
  if (status!=0) {
    switch(status)  {
    case ASCIIFILE_NOT_FOUND: 
      fprintf (stderr, "Error, %s not found\n", filename);
      return status;
      break;
    case ASCII_NO_MEMORY: 
      fprintf (stderr, "Error, not enough memory for reading %s\n",
	       filename);
      return status;
      break;
    case LESS_THAN_TWO_COLUMNS: 
      fprintf (stderr, "Error, %s has less than two columns\n",  
	       filename);
      return status;
      break;
    default:
      fprintf (stderr, "Error %d reading %s\n",  
	       status, filename);
      return status;
      break;
    }
  }

  /* check for consistency with atmospheric levels */

  if (rows != nlyr+1) {
    fprintf (stderr, "Error in %s: layering differs from atmospheric file:\n", filename); 
    fprintf (stderr, "atmosphere has %d levels, found %d in %s.\n", nlyr, rows, filename);
    status = 3;
    return status;
  }
  
  for (lev=0; lev<=nlyr; lev++) {
    if (zd[lev] != (float) rayleigh_zd[lev]) {
      fprintf (stderr, "Error, layers in %s differ from atmosphere layers: %d %f %f\n", 
	       filename, lev, rayleigh_zd[lev], zd[lev]);
      return -1;
    }
  }

  /* copy optical thickness to final destination */
  *dtau = calloc (nlyr, sizeof (float));
  for (lc=0; lc<nlyr; lc++)
    (*dtau)[lc] = rayleigh_dtau[lc+1];
    


  /* free memory */
  free (rayleigh_zd);
  free (rayleigh_dtau);

  return 0;
}




/***********************************************************************************/
/*  Read the molecular absorption file which can either be a two-column file       */
/*  containing altitude [km] and optical thickness of each layer or a matrix       */
/*  tau[wavelength, z]. Check if the levels equal those defined in zd[]. Memory    */
/*  for dtau[] is allocated automatically.                                         */ 
/***********************************************************************************/

static int read_molecular_absorption (char* filename, int nlyr,
				      float *zd, int quiet,
				      float **dtau_mono, 
				      float ***dtau_spec, float **lambda, int *nlambda,
				      int *monochromatic) 
{
  int rows=0, min_columns=0, max_columns=0, status=0;
  int lu=0, iv=0;


  float **data=NULL;
  

#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  char cdf_filename[FILENAME_MAX]="";
  int ncid=0;

  int idd_nlyr=0, idd_nwvl=0;
  int id_wvl=0, id_z=0, id_tau=0;

  long start[1] = {0};
  long count[1] = {0};

  long taustart[2] = {0,0};
  long taucount[2] = {0,0};

  int cdf_nlyr=0;
  float *z=NULL;

  long dimlen=0;

  strcpy (cdf_filename, filename);
  strcat (cdf_filename, ".cdf");


  /* open netcdf file */
  ncid = ncopen (filename, NC_NOWRITE);
  if (ncid==-1) {
    if (!quiet) {
      fprintf (stderr, "*** %s is not a netcdf file, trying to open as ASCII instead.\n", filename);
      fprintf (stderr, "*** It is recommended to use the netCDF format because access is much faster.\n");
    }
  }
  else {
    if (!quiet)
      fprintf (stderr, " ... reading molecular absorption from netcdf file %s\n", filename);

    /* get dimension id for "nlyr" */
    idd_nlyr = ncdimid (ncid, "nlyr");
    
    /* get dimension length for "nlyr" */
    ncdiminq (ncid, idd_nlyr, NULL, &dimlen);
    cdf_nlyr = dimlen;

    if (cdf_nlyr!=nlyr) {
      fprintf (stderr, "Error, number of layers in %s differ from atmospheric layers: %d vs. %d\n", 
	       filename, cdf_nlyr, nlyr);
      return -1;
    }

    /* get dimension id for "nwvl" */
    idd_nwvl = ncdimid (ncid, "nwvl");
    
    /* get dimension length for "nwvl" */
    ncdiminq (ncid, idd_nwvl, NULL, &dimlen);
    *nlambda = dimlen;

    /* read levels */
    z = calloc (cdf_nlyr+1, sizeof(float));

    /* get variable id for "z" */
    id_z = ncvarid (ncid, "z");

    count[0] = cdf_nlyr+1;

    /* read "z" */
    status = ncvarget (ncid, id_z, start, count, (void *) z);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    
    /* compare levels */
    for (lu=0; lu<=nlyr; lu++) {
      if (zd[lu] != z[lu]) {
	fprintf (stderr, "Error, levels in %s differ from atmospheric levels: %f vs. %f\n", 
		 filename, zd[lu], z[lu]);
	return -1;
      }
    }
    
    /* read wavelengths */
    *lambda = calloc (*nlambda, sizeof(float));

    /* get variable id for "wvl" */
    id_wvl = ncvarid (ncid, "wvl");
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    
    free(z);


    /* read "wvl" */
    count[0] = *nlambda;

    status = ncvarget (ncid, id_wvl, start, count, (void *) *lambda);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }


    /* get variable id for "tau" */
    id_tau = ncvarid (ncid, "tau");

    /* read optical thickness */
    taustart[1] = 0;  /* start with first optical thickness */
    taucount[0] = 1;  /* read one layer at a time           */
    
    *dtau_spec = calloc (nlyr, sizeof(float *));
    for (lu=0; lu<nlyr; lu++) {
      (*dtau_spec)[lu] = calloc (*nlambda, sizeof(float));

      taustart[0] = lu;
      taucount[1] = *nlambda;

      status = ncvarget (ncid, id_tau, taustart, taucount, (void *) (*dtau_spec)[lu]);
      if (status!=NC_NOERR) {
	fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
	return status;
      }
    }

    return 0;
  }

#endif

  if (!quiet)
    fprintf (stderr, " ... reading molecular absorption from ASCII file %s\n", filename);

  /* read altitude and optical depth from file */
  status = ASCII_file2float (filename,   
			     &rows,        
			     &max_columns, 
			     &min_columns, 
			     &data);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }


  if (min_columns<2) {
    fprintf (stderr, "Error, found less than 2 columns in %s\n", filename);
    return -1;
  }

  *monochromatic = 0;

  if (min_columns==2 && rows == nlyr+1) {
    
    *monochromatic = 1;

    /* here we assume that the file comes in the old format,  */
    /* two columns z and dtau; in this case, the first column */
    /* must equal the atmospheric profile                     */
    
    for (lu=0; lu<=nlyr; lu++) {
      if (zd[lu] != (float) data[lu][0])
	*monochromatic = 0;
    }

    if (!(*monochromatic)) {
      fprintf (stderr, "%s has two columns; nevertheless it is not interpreted as (z, dtau)\n", filename);
      fprintf (stderr, "format because values in the 1st column differ from atmospheric profile.\n");
    }
    else {

      /* copy optical thickness to final destination */
      *dtau_mono = calloc (nlyr, sizeof (float));
      for (lu=0; lu<nlyr; lu++)
	(*dtau_mono)[lu] = data[lu+1][1];
    }
  }
  
  
  if (!(*monochromatic)) {

    /* now we need to assume that we deal with a new format */
    /* where the first row must be the altitude levels      */
    
    fprintf (stderr, " ... now testing as wavelength-dependent absorption file\n");
    fflush (stderr);

    if (min_columns!=max_columns) {
      fprintf (stderr, "Error, inconsistent number of columns in %s, %d - %d\n",
	       filename, min_columns, max_columns);
      return -1;
    }


    if (min_columns<nlyr+1) {
      fprintf (stderr, "Error, expecting at least %d columns in %s\n", nlyr, filename);
      fprintf (stderr, "if it is supposed to be a wavelength-dependent file\n");

      return -1;
    }
    
    /* compare levels */
    for (lu=0; lu<=nlyr; lu++) {
      if (zd[lu] != (float) data[0][lu]) {
	fprintf (stderr, "Error, levels in %s differ from atmospheric levels: %f vs. %f\n", 
		 filename, zd[lu], data[0][lu]);
	return -1;
      }
    }


    /* now we know that the format is correct, and the only thing   */
    /* we need to do is to copy the data to their final destination */
    
    *nlambda = rows-1;
    *lambda = calloc (*nlambda, sizeof(float));
    *dtau_spec = calloc (nlyr, sizeof (float *));
    for (lu=0; lu<nlyr; lu++)
      (*dtau_spec)[lu] = calloc (*nlambda, sizeof(float));
    


    for (iv=0; iv<*nlambda; iv++) {
      (*lambda)[iv] = data[iv+1][0];
      
      for (lu=0; lu<nlyr; lu++)
	(*dtau_spec)[lu][iv] = data[iv+1][lu+1];
    }
  }

  return 0;
}


/**************************************************************/
/* Calculate atmospheric column in DU.                        */
/**************************************************************/

static int column (float *dens, float *zd, int nlev, float altitude, float *column)
{

  float deltaz=0;
  int lc=0, ialt=0;
  
  for (ialt=0; ialt<nlev; ialt++)
    if (altitude == zd[ialt])
      break;
    
  if (ialt == nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid;\n", altitude);
    fprintf (stderr, "exiting column()!\n");
    return -1;
  }
  

  *column=0;
  for (lc=1; lc<=ialt; ++lc) {
    deltaz = (zd[lc-1] - zd[lc]) * 1.0E5;
    *column += 0.5 * (dens[lc - 1] + dens[lc]) * deltaz;
  }
  
  *column /= 2.687E16;
  
  return 0;
}



/**************************************************************/
/* Scale ozone column to a given value.                       */
/**************************************************************/

static int scale_profile (float *dens, float *zd, int nlev, float newcol, float altitude, float *factor)
{
  int lc=0, status=0;
  float oldcol=0;

  /* determine actual ozone column */
  status = column (dens, zd, nlev, altitude, &oldcol);
  if (status!=0) {
    fprintf (stderr, "Error %d calculating old column in scale_profile(), exiting\n", status);
    return status;
  }

  /* determine scale factor */
  if ( oldcol > 0.0 )
    *factor = newcol / oldcol;
  else
    *factor = 0;

  /* scale profile */
  for (lc = 0; lc<nlev; ++lc)
    dens[lc] *= (*factor);
  
  return 0;
}



/**************************************************************/
/* Calculate preciptable water [kg / m2].                     */
/**************************************************************/

static int precip_water (float *dens, float *zd, int nlev, float altitude, float *precip)
{

  float deltaz=0, tst1=0, tst2=0;
  int lc=0, ialt=0;
  
  for (ialt=0; ialt<nlev; ialt++)
    if (altitude == zd[ialt])
      break;
    
  if (ialt == nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid;\n", altitude);
    fprintf (stderr, "exiting column()!\n");
    return -1;
  }
  

  /* logarithmic interpolation */
  *precip=0;
  for (lc=1; lc<=ialt; ++lc) {
    deltaz = (zd[lc-1] - zd[lc]) * 1E3;
    tst1 = fabs(dens[lc-1]-dens[lc]);
    tst2 = (dens[lc-1]<dens[lc] ? dens[lc-1] : dens[lc]);
    if (tst1<=0.001*dens[lc-1] || tst2<=0)
      *precip += 0.5 * deltaz * (dens[lc-1] + dens[lc]);
    else 
      *precip += deltaz * (dens[lc-1]-dens[lc]) / log(dens[lc-1]/dens[lc]);
  }

  *precip *= (M_H2O / N_MOL_CKDFU * 1e3);
  
  return 0;
}


/**************************************************************/
/* Scale precipitable water to a given value.                 */
/**************************************************************/

static int scale_precip (float *dens, float *zd, int nlev, float precip, float altitude, float *factor)
{
  int lc=0, status=0;
  float precipwat=0;

  /* determine actual precipitable water */
  status = precip_water (dens, zd, nlev, altitude, &precipwat);
  if (status!=0) {
    fprintf (stderr, "Error %d returned by precip_water(), exiting scale_precip()\n", status);
    return status;
  }

  /* determine scale factor */
  *factor = precip / precipwat;

  /* scale profile */
  for (lc = 0; lc<nlev; ++lc)
    dens[lc] *= (*factor);
  
  return 0;
}



/**************************************************************/
/* Convert relative humidity [%] to concentration [1/cm3]     */
/**************************************************************/

static double vapor_pressure (double t)
{
  /*
  double c1=-6096.9385;
  double c2=16.635794;
  double c3=-2.711193e-2;
  double c4=1.673952e-5;
  double c5=2.433502;
  double boltzmann = 1.38065e-23;
  */

  double avogadro  = 6.022142e23;
  
  double tzero = 273.15;
  double a = tzero/t;
  double mh2o = M_H2O/avogadro;


  /* Sonntag, Meteorol. Zeitschr, n.f. 3, 51-66, 1994                                 */
  /* return 100.0 / boltzmann / t * exp (c1/t +c2 +c3*t +c4*t*t + c5*log(t)) / 1.0e6; */
  
  /* SBDART, -> CRC handbook */
  return a * exp (18.916758 - a * (14.845878 + a*2.4918766)) / mh2o / 1.0e6;
}



/**************************************************************/
/* Convert relative humidity [%] to concentration [1/cm3]     */
/**************************************************************/

static int read_and_convert_rh (char *filename,
				float *zd, float *t, int nlev,
				float *rh,
				float *h2o)
{
  int status=0, rows=0, lu=0;

  float *z=NULL, *rh_temp=NULL;
  double conc=0;
  

  /* read relative humidity from file */
  if (strlen(filename)>0) {

    fprintf (stderr, " ... reading relative humidity from %s\n", filename);

    status = read_2c_file_float (filename, 
				 &z, &rh_temp, &rows);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }


    /* interpolate profile to common atmospheric grid, linear */
    status = arb_wvn_zero (rows, z, rh_temp, nlev, zd, rh, 1, 1);
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating relative humidity to common atmospheric grid\n", status);
      return status;
    }
  }
  else {  /* create dummy arrays */
    for (lu=0; lu<nlev; lu++)
      rh[lu] = -1.0;
  }
  

  for (lu=0; lu<nlev; lu++) {

    conc = vapor_pressure (t[lu]);

    /* if a RH is defined for this level, replace the water vapor concentration */
    if (rh[lu]>=0) {
      h2o[lu] = conc * rh[lu]/100.0;
    }
    /* if no RH is defined for this level, use the original water vapor concentration */
    else {
      rh[lu] = 100.0 * h2o[lu] / conc;
    }
  }

  

  return 0;   /* o.k. */
}


/*******************************************/
/* read atmospheric grid from generic ck   */
/* netcdf file                             */
/*******************************************/

static int read_nc_atmosphere_grid (float **zd, int *nlev, char *filename)
{
#if HAVE_LIBNETCDF
  int status=0;

  long dimlen=0;

  long start[1] = {0};
  long count[1] = {0};

  int ncid=0;
  int idd_nlev=0, id_z=0;

  /* determine number of atmospheric levels */
  ncid = ncopen (filename, NC_NOWRITE);
  if (ncid==-1) {
    fprintf (stderr, "Error %d opening netCDF file %s\n", -1, filename);
    return -1;
  }
  
  /* get dimension id for "nlev" */
  idd_nlev = ncdimid (ncid, "nlev");
  
  /* get dimension length for "nwvn" */
  ncdiminq (ncid, idd_nlev, NULL, &dimlen);
  *nlev = dimlen;
  *zd   = calloc (*nlev, sizeof (float));

  /* get variable id for "z" */
  id_z = ncvarid (ncid, "z");
  
  /* read altitude "z" */
  status = ncvarget (ncid, id_z, start, count, (void *) (*zd));
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading z from %s\n", status, filename);
    return status;
  }

  ncclose (ncid);
  return 0;

#else
  fprintf (stderr, "Error, netcdf required by read_nc_atmosphere_grid();\n");
  fprintf (stderr, "please install netcdf and recompile libRadtran!\n");
  return -1;
#endif
}
