/*--------------------------------------------------------------------
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>

#include "uvspec.h"
#include "../src_c/ascii.h"

/************************************/
/* prototypes of internal functions */
/************************************/

static int redistribute_1D (float **data, 
			    float *z_old, int nlyr_old, 
			    float *z_new, int nlyr_new,
			    int scale);

static int redistribute_2D (float ***data, int nlambda, 
			    int nlambda_lower, int nlambda_upper,
			    float *z_old, int nlyr_old, 
			    float *z_new, int nlyr_new,
			    int scale);

static int redistribute_moment (float ****moment, int ***nmom, int nlambda, 
				int nlambda_lower, int nlambda_upper,
				float *z_old, int nlyr_old, 
				float *z_new, int nlyr_new);


int redistribute_molecular (atm_optprop_struct *optprop, int nlambda, 
			    int nlambda_lower, int nlambda_upper,
			    int *nq, 
			    float *zd, int nlyr, 
			    float *zd_common, int nlyr_common);

int redistribute_amf (atm_microphys_struct *microphys, 
		      crs_out_struct *crs,
		      int nlambda, 
		      int nlambda_lower, int nlambda_upper,
		      float *zd, int nlyr, 
		      float *zd_common, int nlyr_common);

/**************************************************************/
/* Setup redistribute irradiance.                         */
/**************************************************************/

int setup_redistribute (input_struct input, output_struct *output)
{
  int lc=0, ialt=0, i=0, status=0;

  /************************ Redistribute molecular optical properties *******************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing molecular optical properties\n");
    fflush (stderr);
  }
  
  status = redistribute_molecular (&(output->atm.optprop), output->wl.nlambda_t, 
				   output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				   output->atm.nq_t,
				   output->atm.zd, output->atm.nlev-1, 
				   output->atm.zd_common, output->atm.nlev_common-1);

  if (status!=0) {
    fprintf (stderr, "Error regridding molecular scattering and absorption\n");
    return status;
  }

  /******************** Redistribute aerosol stuff ****************************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing aerosol optical properties\n");
    fflush (stderr);
  }

  status = redistribute_optprop (&(output->aer.optprop), output->wl.nlambda_t, 
				 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				 output->aer.zd, output->aer.nlev-1, 
				 output->atm.zd_common, output->atm.nlev_common-1);
  
  if (status!=0) {
    fprintf (stderr, "Error regridding aerosol parameters\n");
    return status;
  }

  /**************************** Redistribute water cloud stuff *************************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing water cloud optical properties\n");
    fflush (stderr);
  }

  status = redistribute_optprop (&(output->wc.optprop), output->wl.nlambda_t, 
				 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				 output->wc.zd, output->wc.nlev-1, 
				 output->atm.zd_common, output->atm.nlev_common-1);

  if (status!=0) {
    fprintf (stderr, "Error regridding water cloud parameters\n");
    return status;
  }

  /* also redistribute cloud microphysics (required by MYSTIC) */
  if (output->wc.microphys.alloc) {
    status = redistribute_cld_microphys (&(output->wc.microphys),
					 output->wc.zd, output->wc.nlev-1, 
					 output->atm.zd_common, output->atm.nlev_common-1);
    
    if (status!=0) {
      fprintf (stderr, "Error regridding water cloud microphysical parameters\n");
      return status;
    }
  }

  /**************************** Redistribute ice cloud stuff ***************************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing ice cloud optical properties\n");
    fflush (stderr);
  }

  status = redistribute_optprop (&(output->ic.optprop), output->wl.nlambda_t, 
				 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				 output->ic.zd, output->ic.nlev-1, 
				 output->atm.zd_common, output->atm.nlev_common-1);

  if (status!=0) {
    fprintf (stderr, "Error regridding ice cloud parameters\n");
    return status;
  }


  /************************* Interpolate pressure *************************/

  if (input.verbose) {
    fprintf (stderr, " ...  interpolating pressure\n");
    fflush (stderr);
  }

  status = interpolate_profile (output->atm.zd, &output->atm.microphys.press, output->atm.nlev,
				output->atm.zd_common, output->atm.nlev_common, 2);

  if (status!=0) {
    fprintf (stderr, "Error interpolating pressure\n");
    return status;
  }

  /************************* Interpolate densities ************************/

  if (input.verbose) {
    fprintf (stderr, " ...  interpolating densities\n");
    fflush (stderr);
  }

  status=0;
  for (i=0; i<MOL_NN; i++)
    status += interpolate_profile (output->atm.zd, &output->atm.microphys.dens[i], output->atm.nlev,
				   output->atm.zd_common, output->atm.nlev_common, 2);

  if (status!=0) {
    fprintf (stderr, "Error interpolating densities\n");
    return status;
  }

  /************************ Interpolate temperature ************************/

  if (input.verbose) {
    fprintf (stderr, " ...  interpolating temperature\n");
    fflush (stderr);
  }

  status = interpolate_profile (output->atm.zd, &output->atm.microphys.temper, output->atm.nlev,
				output->atm.zd_common, output->atm.nlev_common, 1);

  if (status!=0) {
    fprintf (stderr, "Error interpolating temperature\n");
    return status;
  }

  /************************ Interpolate index of refraction ************************/

  if (input.verbose) {
    fprintf (stderr, " ...  interpolating index of refraction\n");
    fflush (stderr);
  }

  status = interpolate_profile (output->atm.zd, &output->atm.microphys.refind, output->atm.nlev,
				output->atm.zd_common, output->atm.nlev_common, 1);

  if (status!=0) {
    fprintf (stderr, "Error regridding refractive index\n");
    return status;
  }

  /************************ Redistribute possible amf stuff ************************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing airmass stuff\n");
    fflush (stderr);
  }

  status = redistribute_amf (&(output->atm.microphys), 
			     &(output->crs),
			     output->wl.nlambda_t, 
			     output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
			     output->atm.zd, output->atm.nlev-1, 
			     output->atm.zd_common, output->atm.nlev_common-1);

  if (status!=0) {
    fprintf (stderr, "Error regridding airmass stuff\n");
    return status;
  }

  /* Finally set new number of output level and layers, and set new zd */
  output->atm.nlev = output->atm.nlev_common;
  output->atm.nlyr = output->atm.nlev_common-1;
  free(output->atm.zd);
  output->atm.zd = calloc (output->atm.nlev, sizeof (float));
  for (lc=0; lc<output->atm.nlev_common; lc++) 
    output->atm.zd[lc] = output->atm.zd_common[lc];
  

  /* last but not least consider altitude */
  if (input.atm.altitude > 0) {
    for (ialt=0; ialt<output->atm.nlev_common; ialt++)
      if (input.atm.altitude == output->atm.zd_common[ialt])
	break;
    
    if (ialt == output->atm.nlev_common) {
      fprintf (stderr, "Fatal error, user-defined altitude %f not found in common grid\n", input.atm.altitude);
      return -1;
    }
	
    for (lc=0; lc<output->atm.nlev_common; lc++) {
      output->atm.zd_common[lc] -= input.atm.altitude;
      output->atm.zd[lc]        -= input.atm.altitude;
    }

    output->atm.nlev_common = ialt+1;
    output->atm.nlyr_common = ialt;
    output->atm.nlev        = ialt+1;
    output->atm.nlyr        = ialt;
  }

  return status;
}

int compare_float (void *ap, void *bp) { /* Sort such that altitude is in descending order */
  float *a = (float *) ap;
  float *b = (float *) bp;
  
  if (*a == *b) 
    return 0;

  if (*a > *b) 
    return -1;
  else 
    return 1;
}

/****************************************************************/
/* Regrid data from an input grid to an output grid. All levels */
/* of the input grid must also be available in the output grid; */
/* the scale flag toggles between distributing a quantity over  */
/* the output levels (e.g. an optical thickness, scale==1)      */
/* or simply copying it (e.g. an asymmetry parameter, scale==0) */
/****************************************************************/

int regrid_float (float *xin, float  *yin,  int nin,
		  float *xout, float *yout, int nout,
		  int scale)
{
  int iin=0, iout=0;

  /* first check if both grids are sorted in descending order */
  for (iin=0; iin<nin-1; iin++)
    if (xin[iin] <= xin[iin+1]) {
      fprintf (stderr, "Error, input grid not sorted in descending order!\n");
      fprintf (stderr, "xin[%d] = %f, xin[%d] = %f\n", iin, xin[iin], iin+1, xin[iin+1]);
      return -1;
    }

  for (iout=0; iout<nout-1; iout++)
    if (xout[iout] <= xout[iout+1]) {
      fprintf (stderr, "Error, output grid not sorted in descending order!\n");
      fprintf (stderr, "xout[%d] = %f, xout[%d] = %f\n", iout, xout[iout], iout+1, xout[iout+1]);
      return -1;
    }

  /* first check if all grid points of the input grid */
  /* are also available in the output grid            */

  iout=nout-1;
  for (iin=nin-1; iin>=0; iin--) {
    while (xout[iout]<xin[iin]) 
      if (--iout<0)
	break;
      
    if (xin[iin]!=xout[iout]) {
      fprintf (stderr, "Error, level %f from input not contained in output\n",
	       xin[iin]);
      return -1;
    }
  }
  
  iin  = nin -1;
  iout = nout-1;

  /* go up to first output level */
  while (xout[iout]<xin[iin]) 
    iout--;

  for (; iout>=0; iout--) {

    while (xout[iout]<xin[iin-1]) {
      if (scale) 
	/*	yout[iout] = yin[iin-1] *  */
	yout[iout] = yin[iin] * 
	  (xout[iout-1]-xout[iout]) / (xin[iin-1] - xin[iin]);
      else 
	yout[iout] = yin[iin];
      iout--;
    }

    iout++;
    iin--;
    
    if (iin<1)
      break;
  }

  return 0;
}


/*****************************************************************/
/* Regrid data from an input grid to an output grid. All levels  */
/* of the input grid must also be available in the output grid;  */
/* the scale flag toggles between distributing a quantity over   */
/* the output levels (e.g. an optical thickness, scale==1)       */
/* or simply copying it (e.g. an asymmetry parameter, scale==0); */
/*****************************************************************/

int regrid (double *xin, double  *yin,  int nin,
	    double *xout, double *yout, int nout,
	    int scale)
{
  int iin=0, iout=0;

  /* first check if both grids are sorted in descending order */
  for (iin=0; iin<nin-1; iin++)
    if (xin[iin] <= xin[iin+1]) {
      fprintf (stderr, "Error, input grid not sorted in descending order!\n");
      fprintf (stderr, "xin[%d] = %f, xin[%d] = %f\n", iin, xin[iin], iin+1, xin[iin+1]);
      return -1;
    }

  for (iout=0; iout<nout-1; iout++)
    if (xout[iout] <= xout[iout+1]) {
      fprintf (stderr, "Error, output grid not sorted in descending order!\n");
      fprintf (stderr, "xout[%d] = %f, xout[%d] = %f\n", iout, xout[iout], iout+1, xout[iout+1]);
      return -1;
    }

  /* first check if all grid points of the input grid */
  /* are also available in the output grid            */

  iout=nout-1;
  for (iin=nin-1; iin>=0; iin--) {
    while (xout[iout]<xin[iin]) 
      if (--iout<0)
	break;
      
    if (xin[iin]!=xout[iout]) {
      fprintf (stderr, "Error, level %16.12f from input not contained in output\n",
	       xin[iin]);
      return -1;
    }
  }

  
  iin  = nin -1;
  iout = nout-1;

  /* go up to first output level */
  while (xout[iout]<xin[iin]) 
    iout--;

  for (; iout>=0; iout--) {

    while (xout[iout]<xin[iin-1]) {
      if (scale) 
	/*	yout[iout] = yin[iin-1] *  */
	yout[iout] = yin[iin] * 
	  (xout[iout-1]-xout[iout]) / (xin[iin-1] - xin[iin]);
      else 
	yout[iout] = yin[iin];
      iout--;
    }

    iout++;
    iin--;
    
    if (iin<1)
      break;
  }

  return 0;
}


/* copy array source to target and round to 6 digits */
void cp_round (float *source, float *target, int nlev)
{
  int lc=0;
  float pa=0;

  for (lc=0; lc<nlev; lc++) {
    if (source[lc]<=0.0)
      target[lc] = 0.0;
    else {
      pa=pow(10.0,floor(log10(source[lc]))-6.0);
      target[lc] = floor(source[lc]/pa)*pa;
    }
  }
}


void set_common_z_grid (float *zd, int nlev, float **zd_common, int *nlev_common) {
  int i=0, k=0, l=0, included=0;
  float *tmp = (float *) calloc (nlev+*nlev_common, sizeof(float));

  /* add the existing common grid to tmp */
  for (k=0; k<*nlev_common; k++)
    tmp[k] = (*zd_common)[k];

  /* cp_round (*zd_common, tmp, *nlev_common); */


  for (i=0; i<nlev; i++) {

    /* Is this level already included? */
    included = 0;
    for (l=0; l<k; l++)
      if (tmp[l] == zd[i]) {
	included=1;
	break;
      }
    
    /* if not already included then add */
    if (!included) 
      tmp[k++] = zd[i];
  }

  *nlev_common = k;

  if (*nlev_common==0) {   /* Special for first time, and no need to sort this */
    (*zd_common) = (float *) calloc (nlev, sizeof(float));

    for (i=0;i<nlev;i++) 
      (*zd_common)[i] = zd[i];      

    *nlev_common = nlev;
  }
  else {
    free(*zd_common);
    (*zd_common) = (float *) calloc (*nlev_common, sizeof(float));
    for (i=0; i<*nlev_common; i++) 
      (*zd_common)[i] = tmp[i];

    qsort((*zd_common), *nlev_common, sizeof(float), compare_float );
  }
  
  free(tmp);
}



/* add z levels from a file to an existing grid; */
/* reallocate memory as needed.                  */
int add_file2grid (float **zd, int *nlev, char *filename)
{
  int status=0;
  float *znew=NULL;
  int nnew=0;

  status = read_1c_file_float (filename, &znew, &nnew);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* add levels to aerosol profile */
  set_common_z_grid (znew, nnew, zd, nlev);
  free (znew);

  return 0;
}




/****************************************************************/
/* Regrid a 1D float array data from a given input grid (z_old) */
/* to an given output grid. All levels of the input grid must   */
/* also be available in the output grid; the scale flag toggles */
/* between distributing a quantity over the output levels       */
/* (e.g. an optical thickness, scale==1) or simply copying it   */
/* (e.g. an asymmetry parameter, scale==0); memory for data is  */
/* reallocated automatically.                                   */
/****************************************************************/

static int redistribute_1D (float **data, 
			    float *z_old, int nlyr_old, 
			    float *z_new, int nlyr_new,
			    int scale) 
{
  int lc=0, status=0;
  double *xin =NULL, *yin =NULL;
  double *xout=NULL, *yout=NULL;

  double *tmp=NULL; 


  xin  = (double *) calloc (nlyr_old+1, sizeof(double));
  yin  = (double *) calloc (nlyr_old+1, sizeof(double));

  xout = (double *) calloc (nlyr_new+2, sizeof(double));
  yout = (double *) calloc (nlyr_new+2, sizeof(double));



  /* Allocate memory for temporary array */
  tmp = (double *) calloc (nlyr_old+1, sizeof(double));

  /* copy original stuff to temporary array */
  for (lc=0; lc<=nlyr_old; lc++) 
    tmp[lc] = (*data)[lc];
  
  /* free original array and reallocate with changed resolution */
  free (*data);
  *data = (float *) calloc (nlyr_new, sizeof(float));

  xin[0] = z_old[0];
  yin[0] = tmp[0];
  for (lc=0; lc<nlyr_old; lc++) {
    xin[lc+1]  = z_old[lc+1];
    yin[lc+1]  = tmp[lc];
  }

  for (lc=0; lc<=nlyr_new; lc++)
    xout[lc] = z_new[lc];
  
  status = regrid (xin, yin, nlyr_old+1,
		   xout, yout, nlyr_new+1, scale);
  
  if ( status != 0 ) {
    fprintf(stderr, "redistribute_1D(): Error %d returned by regrid()\n", status);
    return status;
  }
  
  for (lc=1; lc<=nlyr_new; lc++)
    (*data)[lc-1] = yout[lc];      
    

  free (tmp);
  
  free(xin); free(xout);
  free(yin); free(yout);
  
  return 0;
}



/****************************************************************/
/* Regrid a 2D float array data from a given input grid (z_old) */
/* to an given output grid. All levels of the input grid must   */
/* also be available in the output grid; the scale flag toggles */
/* between distributing a quantity over the output levels       */
/* (e.g. an optical thickness, scale==1) or simply copying it   */
/* (e.g. an asymmetry parameter, scale==0); memory for data is  */
/* reallocated automatically.                                   */
/****************************************************************/

static int redistribute_2D (float ***data, int nlambda, 
			    int nlambda_lower, int nlambda_upper,
			    float *z_old, int nlyr_old, 
			    float *z_new, int nlyr_new,
			    int scale) 
{
  int lc=0, iv=0, status=0;
  double *xin =NULL, *yin =NULL;
  double *xout=NULL, *yout=NULL;

  double **tmp=NULL; 


  xin  = (double *) calloc (nlyr_old+1, sizeof(double));
  yin  = (double *) calloc (nlyr_old+1, sizeof(double));

  xout = (double *) calloc (nlyr_new+1, sizeof(double));
  yout = (double *) calloc (nlyr_new+1, sizeof(double));



  /* Allocate memory for temporary array */
  if ((status = ASCII_calloc_double (&tmp, nlambda, nlyr_old)) != 0) {
    fprintf (stderr, "Error allocating %d x %d doubles for tmp\n",
	     nlambda, nlyr_old);
    return status;
  }

  /* copy original stuff to temporary array */
  for (lc=0; lc<nlyr_old; lc++) 
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      tmp[iv][lc] = (*data)[iv][lc];
  
  /* free original array and reallocate with changed resolution */
  ASCII_free_float (*data, nlambda);
  if ((status = ASCII_calloc_float (data, nlambda, nlyr_new)) != 0) {
    fprintf (stderr, "Error allocating memory for data\n");
    return status;
  }
  
  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    xin[0] = z_old[0];
    yin[0] = tmp[iv][0];
    for (lc=0; lc<nlyr_old; lc++) {
      xin[lc+1]  = z_old[lc+1];
      yin[lc+1]  = tmp[iv][lc];
    }

    for (lc=0; lc<=nlyr_new; lc++)
      xout[lc] = z_new[lc];

    status = regrid (xin, yin, nlyr_old+1,
		     xout, yout, nlyr_new+1, scale);

    if ( status != 0 ) {
      fprintf(stderr, "redistribute_2D(): Error %d returned by regrid()\n", status);
      return status;
    }

    for (lc=1; lc<=nlyr_new; lc++)
      (*data)[iv][lc-1] = yout[lc];      
      
  }

  ASCII_free_double (tmp, nlambda);

  free(xin); free(xout);
  free(yin); free(yout);

  return 0;
}



/*****************************************************************/
/* Regrid phase function moments from a given input grid (z_old) */
/* to an given output grid. All levels of the input grid must    */
/* also be available in the output grid; the scale flag toggles  */
/* between distributing a quantity over the output levels        */
/* (e.g. an optical thickness, scale==1) or simply copying it    */
/* (e.g. an asymmetry parameter, scale==0); memory for the       */
/* moments is reallocated automatically.                         */
/*****************************************************************/

static int redistribute_moment (float ****moment, int ***nmom, int nlambda, 
				int nlambda_lower, int nlambda_upper,
				float *z_old, int nlyr_old, 
				float *z_new, int nlyr_new) 
{
  int lc=0, iv=0, k=0, status=0;
  int max=0;
  double *xin =NULL, *yin =NULL;
  double *xout=NULL, *yout=NULL;

  double ***tmpopt=NULL, ***tmpout=NULL;

  int **tmp=NULL; 


  xin  = (double *) calloc (nlyr_old+1, sizeof(double));
  yin  = (double *) calloc (nlyr_old+1, sizeof(double));

  xout = (double *) calloc (nlyr_new+1, sizeof(double));
  yout = (double *) calloc (nlyr_new+1, sizeof(double));



  /* Allocate memory for temporary array */
  if ((status = ASCII_calloc_int (&tmp, nlambda, nlyr_old)) != 0)
    return status;

  /* copy original stuff to temporary array */
  for (lc=0; lc<nlyr_old; lc++) 
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      tmp[iv][lc] = (*nmom)[iv][lc];
  
  /* free original array and reallocate with changed resolution */
  ASCII_free_int (*nmom, nlambda);
  if ((status = ASCII_calloc_int (nmom, nlambda, nlyr_new)) != 0)
    return status;
  
  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    xin[0] = z_old[0];
    yin[0] = (float) tmp[iv][0];
    for (lc=0; lc<nlyr_old; lc++) {
      xin[lc+1]  = z_old[lc+1];
      yin[lc+1]  = tmp[iv][lc];
    }

    for (lc=0; lc<=nlyr_new; lc++)
      xout[lc] = z_new[lc];

    status = regrid (xin, yin, nlyr_old+1,
		     xout, yout, nlyr_new+1, 0);

    if ( status != 0 ) {
      fprintf(stderr, "redistribute_moment(): Error %d returned by regrid()\n", status);
      return status;
    }

    for (lc=1; lc<=nlyr_new; lc++)
      (*nmom)[iv][lc-1] = (int) (yout[lc] + 0.5);

  }


  max = 0;
  for (iv=0; iv<nlambda; iv++) 
    for (lc=0; lc<nlyr_old; lc++) 
      if ( tmp[iv][lc] > max) 
	max = tmp[iv][lc];

  /* Allocate memory for temporary ary */
  tmpopt =  calloc (nlambda, sizeof (double **));
  for (iv=0; iv<nlambda; iv++) {
    tmpopt[iv] = calloc (nlyr_old, sizeof (double *));
    for (lc=0; lc<nlyr_old; lc++) 
      tmpopt[iv][lc] = calloc (max+1, sizeof (double));
  }

  /* cp original stuff to tmp ary */
  for (lc=0; lc<nlyr_old; lc++) 
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (k=0; k<tmp[iv][lc]; k++) 
	tmpopt[iv][lc][k] = (*moment)[iv][lc][k];
  
  /* allocate tmpary with max moments in all layers */
  tmpout =  calloc (nlambda, sizeof (double **));
  for (iv=0; iv<nlambda; iv++) {
    tmpout[iv] = calloc (nlyr_new, sizeof (double *));
    for (lc=0; lc<nlyr_new; lc++) 
      tmpout[iv][lc] = calloc (max+1, sizeof (double));
  }

  /* free original ary and allocate with changed resolution */
  ASCII_free_float_3D (*moment, nlambda, nlyr_old);
  *moment = calloc (nlambda, sizeof (float **));
  for (iv=0; iv<nlambda; iv++) {
    (*moment)[iv] = calloc (nlyr_new, sizeof (float *));
    for (lc=0; lc<nlyr_new; lc++) 
      (*moment)[iv][lc] = calloc ((*nmom)[iv][lc]+1, sizeof (float));
  }

  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    for (k=0;k<max;k++) {
      xin[0] = z_old[0];
      yin[0] = tmpopt[iv][0][k];
      for (lc=0; lc<nlyr_old; lc++) {
	xin[lc+1]  = z_old[lc+1];
	yin[lc+1]  = tmpopt[iv][lc][k];
      }
      for (lc=0; lc<=nlyr_new; lc++) 
      	xout[lc] = z_new[lc];
      status = regrid (xin, yin, nlyr_old+1,
		       xout, yout, nlyr_new+1, 0);
      if ( status != 0 ) {
	fprintf(stderr, "Redistribute: error calling regrid aer.moment\n");
	return status;
      }
      for (lc=1; lc<=nlyr_new; lc++) 
	tmpout[iv][lc-1][k] = yout[lc];
    }
  }

  for (lc=0; lc<nlyr_new; lc++) 
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (k=0; k<(*nmom)[iv][lc]; k++) 
	(*moment)[iv][lc][k] = (float) tmpout[iv][lc][k];


  ASCII_free_double_3D (tmpopt, nlambda, nlyr_old);
  ASCII_free_double_3D (tmpout, nlambda, nlyr_new);
  ASCII_free_int (tmp, nlambda);

  free(xin); free(xout);
  free(yin); free(yout);

  return 0;
}





/***********************************/
/* Redistribute optical properties */
/***********************************/

int redistribute_optprop (optprop_struct *optprop, int nlambda, 
			  int nlambda_lower, int nlambda_upper,
			  float *zd, int nlyr, 
			  float *zd_common, int nlyr_common)
{
  int status=0;

  /** dtau **/
  status += redistribute_2D (&(optprop->dtau), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     1);

  /** ssa  **/
  status += redistribute_2D (&(optprop->ssa), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0);

  /**  g1  **/
  status += redistribute_2D (&(optprop->g1), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0);

  /**  g2  **/
  status += redistribute_2D (&(optprop->g2), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0);

  /**  ff  **/
  status += redistribute_2D (&(optprop->ff), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0);

  /** moments **/
  status += redistribute_moment (&(optprop->moment), &(optprop->nmom), nlambda, 
				 nlambda_lower, nlambda_upper,
				 zd, nlyr, 
				 zd_common, nlyr_common);
  
  return status;
}


/***********************************/
/* Redistribute optical properties */
/***********************************/

int redistribute_cld_microphys (cld_microphys_struct *microphys,
				float *zd, int nlyr, 
				float *zd_common, int nlyr_common)
{
  int lc=0, status=0;
  
  /** lwc  **/
  status += redistribute_1D (&(microphys->lwc_layer),
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0);

  /** reff **/
  status += redistribute_1D (&(microphys->effr_layer),
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0);

  return status;
}






/**************************************************/
/* Redistribute molecular scattering / absorption */
/**************************************************/

int redistribute_molecular (atm_optprop_struct *optprop, int nlambda, 
			    int nlambda_lower, int nlambda_upper,
			    int *nq, 
			    float *zd, int nlyr, 
			    float *zd_common, int nlyr_common)
{
  int lc=0, iv=0, iq=0, scale=0, status=0;
  double ***tmpopt=NULL;

  double *xin=NULL, *yin=NULL, *xout=NULL, *yout=NULL;

  xin  = (double *) calloc (nlyr_common+1, sizeof(double));
  yin  = (double *) calloc (nlyr_common+1, sizeof(double));

  xout = (double *) calloc (nlyr_common+1, sizeof(double));
  yout = (double *) calloc (nlyr_common+1, sizeof(double));


  /* Allocate memory for temporary ary */
  if ((status = ASCII_calloc_double_3D_arylen (&tmpopt, nlyr, nlambda, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for tmpopt\n");
    return status;
  }

  /* cp original stuff to tmp ary */
  for (lc=0; lc<nlyr; lc++) 
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (iq=0; iq<nq[iv]; iq++) 
	tmpopt[lc][iv][iq] = optprop->tau_molabs_t[lc][iv][iq];
  
  /* free original ary and allocate with changed resolution */
  ASCII_free_double_3D (optprop->tau_molabs_t, nlyr, nlambda);
  if ((status = ASCII_calloc_double_3D_arylen (&optprop->tau_molabs_t,
					       nlyr_common, nlambda, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for optprop->tau_molabs_t\n");
    return status;
  }

  scale = 1;
  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    for (iq=0; iq<nq[iv]; iq++) {
      xin[0] = zd[0];
      yin[0] = tmpopt[0][iv][iq];
      for (lc=0; lc<nlyr; lc++) {
	xin[lc+1] = zd[lc+1];
	yin[lc+1] = tmpopt[lc][iv][iq];
      }

      for (lc=0; lc<=nlyr_common; lc++) 
      	xout[lc] = zd_common[lc];

	status = regrid (xin, yin, nlyr+1,
			 xout, yout, nlyr_common+1, scale);
	if ( status != 0 ) {
	  fprintf(stderr, "Redistribute: error calling regrid for tau_molabs_t\n");
	  return status;
	}
	for (lc=1; lc<=nlyr_common; lc++) 
	  optprop->tau_molabs_t[lc-1][iv][iq] = yout[lc];      
    }
  }
  ASCII_free_double_3D (tmpopt, nlyr, nlambda);
  
  /************************ Redistribute molecular optical properties -md *********************/

  /* Allocate memory for temporary ary */
  if ((status = ASCII_calloc_double_3D_arylen (&tmpopt, nlyr, nlambda, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for tmpopt\n");
    return status;
  }

  /* cp original stuff to tmp ary */
  for (lc=0; lc<nlyr; lc++) 
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (iq=0; iq<nq[iv]; iq++) 
	tmpopt[lc][iv][iq] = optprop->tau_molabs_md_t[lc][iv][iq];
  
  /* free original ary and allocate with changed resolution */
  ASCII_free_double_3D (optprop->tau_molabs_md_t, nlyr, nlambda);
  if ((status = ASCII_calloc_double_3D_arylen (&optprop->tau_molabs_md_t,
					       nlyr_common, nlambda, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for optprop->tau_molabs_md_t\n");
    return status;
  }

  scale = 1;
  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    for (iq=0; iq<nq[iv]; iq++) {
      xin[0] = zd[0];
      yin[0] = tmpopt[0][iv][iq];
      for (lc=0; lc<nlyr; lc++) {
	xin[lc+1]  = zd[lc+1];
	yin[lc+1]  = tmpopt[lc][iv][iq];
      }
      for (lc=0; lc<=nlyr_common; lc++) 
      	xout[lc] = zd_common[lc];

	status = regrid (xin, yin, nlyr+1,
			 xout, yout, nlyr_common+1, scale);
	if ( status != 0 ) {
	  fprintf(stderr, "Redistribute: error calling regrid for tau_molabs_md\n");
	  return status;
	}
	for (lc=1; lc<=nlyr_common; lc++) 
	  optprop->tau_molabs_md_t[lc-1][iv][iq] = yout[lc];      
    }
  }
  ASCII_free_double_3D (tmpopt, nlyr, nlambda);

  /********************************** Redistribute Rayleigh scattering *********************/

  /* Allocate memory for temporary ary */
  if ((status = ASCII_calloc_double_3D_arylen (&tmpopt, nlyr, nlambda, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for tmpopt\n");
    return status;
  }

  /* cp original stuff to tmp ary */
  for (lc=0; lc<nlyr; lc++) 
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (iq=0; iq<nq[iv]; iq++) 
	tmpopt[lc][iv][iq] = optprop->tau_rayleigh_t[lc][iv][iq];
  
  /* free original ary and allocate with changed resolution */
  ASCII_free_double_3D (optprop->tau_rayleigh_t, nlyr, nlambda);
  if ((status = ASCII_calloc_double_3D_arylen (&optprop->tau_rayleigh_t,
					       nlyr_common, nlambda, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for tmpopt\n");
    return status;
  }

  scale = 1;
  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    for (iq=0; iq<nq[iv]; iq++) {
      xin[0] = zd[0];
      yin[0] = tmpopt[0][iv][iq];
      for (lc=0; lc<nlyr; lc++) {
	xin[lc+1]  = zd[lc+1];
	yin[lc+1]  = tmpopt[lc][iv][iq];
      }
      for (lc=0; lc<=nlyr_common; lc++) 
      	xout[lc] = zd_common[lc];

	status = regrid (xin, yin, nlyr+1,
			 xout, yout, nlyr_common+1, scale);
	if ( status != 0 ) {
	  fprintf(stderr, "Redistribute: error calling regrid for tau_rayleigh_t\n");
	  return status;
	}
	for (lc=1; lc<=nlyr_common; lc++) 
	  optprop->tau_rayleigh_t[lc-1][iv][iq] = yout[lc];      
    }
  }
  ASCII_free_double_3D (tmpopt, nlyr, nlambda);

  /* free memory */
  free (xin); free (xout);
  free (yin); free (yout);

  return 0;
}



/******************************/
/* Redistribute airmass stuff */
/******************************/

int redistribute_amf (atm_microphys_struct *microphys, 
		      crs_out_struct *crs,
		      int nlambda, 
		      int nlambda_lower, int nlambda_upper,
		      float *zd, int nlyr, 
		      float *zd_common, int nlyr_common)
{
  int is=0, iv=0, lc=0, scale=0, status=0;
  

  float  *xfin=NULL, *yfin=NULL, *xfout=NULL, *yfout=NULL;
  double *xin=NULL, *yin=NULL, *xout=NULL, *yout=NULL;
  double **tmp2=NULL;


  xin  = (double *) calloc (nlyr_common+1, sizeof(double));
  yin  = (double *) calloc (nlyr_common+1, sizeof(double));
  xout = (double *) calloc (nlyr_common+1, sizeof(double));
  yout = (double *) calloc (nlyr_common+1, sizeof(double));

  xfin  = (float *) calloc (nlyr_common+1, sizeof(float));
  yfin  = (float *) calloc (nlyr_common+1, sizeof(float));
  xfout = (float *) calloc (nlyr_common+1, sizeof(float));
  yfout = (float *) calloc (nlyr_common+1, sizeof(float));



  if (microphys->denstab_id > 0) {

    /** crs_amf **/

    /* Allocate memory for temporary ary */
    if ((status = ASCII_calloc_double (&tmp2, nlambda, nlyr+1)) != 0) {
      fprintf (stderr, "Error allocating memory for tmp2\n");
      return status;
    }

    /* cp original stuff to tmp ary */
    for (lc=0; lc<nlyr+1; lc++) 
      for (iv=nlambda_lower; iv<=nlambda_upper; iv++) 
	tmp2[iv][lc] = crs->crs_amf[iv][lc];

    /* allocate needed memory */
    ASCII_free_float (crs->crs_amf, nlambda);
    if ((status = ASCII_calloc_float (&crs->crs_amf, nlambda,
				      nlyr_common+1)) != 0)
      return status;

    scale = 0;
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
      for (lc=0; lc<nlyr+1; lc++) {
	xin[lc]  = zd[lc];
	yin[lc]  = tmp2[iv][lc];
      }
      for (lc=0; lc<=nlyr_common; lc++) 
	xout[lc] = zd_common[lc];
      
      status = regrid (xin, yin, nlyr+1,
		       xout, yout, nlyr_common+1, scale);
      if ( status != 0 ) {
	fprintf(stderr, "Redistribute: error calling regrid for crs.crs_amf\n");
	return status;
      }
      for (lc=0; lc<=nlyr_common; lc++)
	crs->crs_amf[iv][lc] = yout[lc];      
    }
    ASCII_free_double (tmp2, nlambda);

    /** nsza_denstab **/

    /* Allocate memory for temporary ary */
    if ((status = ASCII_calloc_double (&tmp2, 
      				       microphys->nsza_denstab, 
      				       nlyr+1)) != 0)
      return status;

    /* cp original stuff to tmp ary */
    for (lc=0; lc<=nlyr; lc++) 
      for (is=0; is<microphys->nsza_denstab; is++) 
    	tmp2[is][lc] = microphys->denstab_amf[is][lc];

    /* free original ary and allocate with changed resolution */
    ASCII_free_float (microphys->denstab_amf, microphys->nsza_denstab);
    if ((status = ASCII_calloc_float (&microphys->denstab_amf, 
				      microphys->nsza_denstab, 
				      nlyr_common+1)) != 0)
      return status;
    
    for (lc=0; lc<=nlyr; lc++) 
      xfin[nlyr-lc] = zd[lc];
    for (lc=0; lc<=nlyr_common; lc++) 
      xfout[lc]  = zd_common[lc];

    for (is=0; is<microphys->nsza_denstab; is++) {
      for (lc=0; lc<=nlyr; lc++) 
      	yfin[nlyr-lc] = (float) tmp2[is][lc];

      status = arb_wvn (nlyr+1, xfin, yfin,
			nlyr_common+1, xfout, yfout, 1, 0); 
      if (status!=0) {
	fprintf (stderr, "Redistribute: Error %d interpolating denstab_amf\n", status);
	return status;
      }
      for (lc=0; lc<=nlyr_common; lc++) 
	microphys->denstab_amf[is][lc] = yfout[lc];
    }
    
    ASCII_free_double (tmp2, microphys->nsza_denstab);

  }

  /* free memory */
  free (xin); free (xout);
  free (yin); free (yout);

  free (xfin); free (xfout);
  free (yfin); free (yfout);



  return 0;
}
