/*--------------------------------------------------------------------
 * $Id: zenith.c,v 1.4 2002/03/13 16:10:51 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h> 
#include "sun.h"


#define PROGRAM "ZENITH"
#define VERSION "1.99c"
#define DATE    "March 13, 2002"


#define IFU_LATITUDE   47.48
#define IFU_LONGITUDE -11.07
#define IFU_LONG_STD  -15.00


/*************************************************************/
/* print usage information                                   */
/*************************************************************/

static void print_usage (char *filename)
{
  fprintf (stderr, "%s %s - calculate solar zenith and azimuth\n\n", PROGRAM, VERSION);
  fprintf (stderr, "written by  Bernhard Mayer,\n");
  fprintf (stderr, "            DLR, eMail bernhard.mayer@dlr.de\n");
  fprintf (stderr, "Version %s finished %s\n\n", VERSION, DATE);
  fprintf (stderr, "Be aware that this program is a beta version! Please report any\n");
  fprintf (stderr, "kind of error to the author, including the error message and the\n");
  fprintf (stderr, "database being processed when the error occured. Thanks!\n\n");
  fprintf (stderr, "USAGE: %s [options] <day> <month> <hour> <min> [sec]\n\n", filename);
  fprintf (stderr, "Optional arguments:\n");
  fprintf (stderr, "  -h             display help message\n");
  fprintf (stderr, "  -a             Latitude  (North positive)\n");
  fprintf (stderr, "  -o             Longitude (West positive !!!)\n");
  fprintf (stderr, "  -s             Standard Longitude (West positive !!!)\n");
  fprintf (stderr, "  -l <location>  Instead of \"-a -o -s\" define a location;\n");
  fprintf (stderr, "                 possible locations are ifu, dlrop.\n");
  fprintf (stderr, "  -e             Calculate eccentricity\n\n");
  fprintf (stderr, "If the coordinates are omitted, the location of IFU is used for\n");
  fprintf (stderr, "the calculation:\n");
  fprintf (stderr, "  Latitude 47.48, Longitude -11.07, Standard Longitude.\n");
}



/*************************************************************/
/* parse command line options                                */
/*************************************************************/

static int get_options (int argc, char **argv,
			char *programname,
			int *doy,
			int *print,
			int *time_std,
			double *latitude,
			double *longitude,
			double *long_std,
			int *ecc,
			char *locstr)
{
  int c=0;
  char *dummy=NULL;

  int day   = 0;
  int month = 0;
  int hour  = 0;
  int min   = 0;
  int sec   = 0;

  char lat=0, lon=0, std=0;


  /* reset parameters to default values */
  *latitude  = IFU_LATITUDE;
  *longitude = IFU_LONGITUDE;
  *long_std  = IFU_LONG_STD;
  strcpy (locstr, "");
  *ecc       = 0;

  /* save name of program */
  strncpy (programname, argv[0], FILENAME_MAX);
  
  /* get command line arguments */
  /* The following line is not needed with the getopt.{c,h} I have */
  /* Arve Kylling, Jul 25, 1997 */
  /* optmode = GETOPT_ANY; */

  while ((c=getopt (argc, argv, "ha:o:ps:l:e")) != EOF)  {
    switch(c)  {
    case 'h':  /* help */
      print_usage (programname);
      return (-1);
      break;
    case 'a':  /* latitude  */  
      *latitude  = (double) strtod (optarg, &dummy);
      lat=1;
      break;   
    case 'o':  /* longitude */  
      *longitude = (double) strtod (optarg, &dummy);
      lon=1;
      break;   
    case 'p':  /* no print messages */  
      *print = 0;
      break;   
    case 's':  /* standard longitude */  
      *long_std = (double) strtod (optarg, &dummy);
      std=1;
      break;   
    case 'l':
      strcpy (locstr, optarg);
      break;
    case 'e':  /* eccentricity */  
      *ecc = 1;
      break;   
    case '?':
      print_usage (programname);
      return (-1);
      break;
    default:
      print_usage (programname);
      return (-1);
    }
  }
      

  /* check number of remaining command line arguments */
  if (argc - optind != 4 && argc - optind != 5)  {
    print_usage (programname);
    return -1;
  }
  
  /* check arguments */
  if (lat&!lon || lat&!std || lon&!lat || lon&!std || std&!lat || std&!lon) {
    print_usage (programname);
    fprintf (stderr, "\n** Specify all three parameters, latitude, longitude and **\n");
    fprintf (stderr,   "** standard longitude!!                                  **\n");
    
    *latitude  = IFU_LATITUDE;
    *longitude = IFU_LONGITUDE;
    *long_std  = IFU_LONG_STD;

    return -1;
  }

  if ((lat||lon||std) && strlen(locstr)>0) {
    print_usage (programname);
    fprintf (stderr, "\n** It does not make sense to define both \"-l\" and one of **\n");
    fprintf (stderr,   "** \"-a\", \"-o\", \"-l\"!!                                    **\n");

    return -1;
  }

  /* check coordinates range */
  if (*latitude<-90 || *latitude>90)  {
    fprintf (stderr, "\n** Latitude out of range **\n");
    return -1;
  }

  if (*longitude<-180 || *longitude>180)  {
    fprintf (stderr, "\n** Longitude out of range **\n");
    return -1;
  }

  if (*long_std<-180 || *long_std>180)  {
    fprintf (stderr, "\n** Standard Longitude out of range **\n");
    return -1;
  }


  /* get command line arguments */
  day   = strtod (argv[optind+0], &dummy);
  month = strtod (argv[optind+1], &dummy);
  hour  = strtod (argv[optind+2], &dummy);
  min   = strtod (argv[optind+3], &dummy);

  if (argc - optind == 5)
    sec = strtod (argv[optind+4], &dummy);
  else
    sec = 0; 


  /* calculate standard time [sec] and check range */
  if ((*time_std = hour*3600 + min*60 + sec) > 86400)  {
    print_usage (programname);
    fprintf (stderr, "\n** Time out of range **\n");
    return -1;
  }

  /* calculate day of year */
  if ((*doy = day_of_year(day, month)) < 0)  {
    print_usage (programname);
    fprintf (stderr, "\n** Date out of range **\n");
    return -1;
  }


  return 0;  /* if o.k. */
}






int main(int argc, char ** argv)
{
  char programname[FILENAME_MAX]="";
  char locstr[255]="";

  int day=0;

  double zenith    =  0.0;
  double azimuth   =  0.0;
  
  double latitude=0, longitude=0, long_std=0;

  int print =1;
  int status=0;

  int time=0;
  int time_std=0;

  int hour_lat  = 0;
  int min_lat   = 0;
  int sec_lat   = 0;

  int hour  = 0;
  int min   = 0;
  int sec   = 0;

  char dummy[10] = "";

  int ecc=0;
  double eccent=0;

  /* get command line options */
  status = get_options (argc, argv,
			programname,
			&day, &print, &time_std,
			&latitude, &longitude, &long_std, &ecc, locstr);

  if (status!=0)
    return status;

  if (strlen(locstr)>0) {
    fprintf (stderr, " ... using location %s\n", locstr);

    status = location (locstr, &latitude, &longitude, &long_std);
    
    if (status!=0) {
      fprintf (stderr, "Error %d calculating coordinates if location %s\n", 
	       status, locstr);
      return status;
    }
  }

  /* calculate solar zenith angle */
  zenith  = solar_zenith  (time_std, day, latitude, longitude, long_std);
  azimuth = solar_azimuth (time_std, day, latitude, longitude, long_std);

  /* calculate LOCAL APPARENT TIME */
  time = LAT (time_std, day, longitude, long_std);

  hour_lat = time / 3600;
  min_lat  = (time - hour_lat*3600) / 60;
  sec_lat  = (time - hour_lat*3600 - min_lat*60);
    
  /* calculate hour, min, sec of standard time */
  hour = time_std / 3600;
  min  = (time_std - hour*3600) / 60;
  sec  = (time_std - hour*3600 - min*60);
    
  if (ecc)
    eccent = eccentricity (day);

  if ( print ) {
    fprintf (stderr, "Coordinates: Lat. %7.2f deg., Long. %7.2f deg., Std. Long. %7.2f deg.\n",
	     latitude, longitude, long_std);
    fprintf (stderr, "  Standard Time:        %s CET\n", time2str(dummy, hour,min,sec));
    fprintf (stderr, "  Local Apparent Time:  %s\n", time2str(dummy, hour_lat,min_lat,sec_lat));
    fprintf (stderr, "  Solar Zenith  Angle:  %8.3f degree\n", zenith);
    fprintf (stderr, "  Solar Azimuth Angle:  %8.3f degree\n", azimuth);
    if (ecc)
      fprintf (stderr, "  Eccentricity:           %6.4f\n", eccent);

    fprintf (stderr, "\noutput to stdout: time [CET], zenith, azimuth");
    if (ecc)
      fprintf (stderr, ", eccentricity\n");
    else 
      fprintf (stderr, "\n");

  }

  fprintf (stdout, "%s, %g  %g", time2str(dummy, hour, min, sec), zenith, azimuth);

  if (ecc)
    fprintf (stdout, "  %g\n", eccent);
  else 
    fprintf (stdout, "\n");
    

  return 0;   /* if o.k. */
}
